import math
#from mojo.drawingTools import *
import mojo.drawingTools as ctx
import AppKit
from mojo.UI import UpdateCurrentGlyphView, MenuBuilder
import mojo.roboFont
from mojo.events import installTool, EditingTool, BaseEventTool, setActiveEventTool, publishEvent
from mojo.UI import getDefault
"""
    This is a tool provides an interface for the current designspace location
    Send events that other tools can respond to.
"""
import os
DEVELOP = False

if DEVELOP:
    pathForBundle = os.path.dirname(__file__)
    resourcePathForBundle = os.path.join(os.path.dirname(pathForBundle), "resources")
    skateboardBundle = mojo.extensions.ExtensionBundle(path=pathForBundle, resourcesName=resourcePathForBundle)
else:
    skateboardBundle = mojo.extensions.ExtensionBundle("SkateBoard")

toolbarIconPath = os.path.join(skateboardBundle.resourcesPath(), "navigatorButton.pdf")
toolbarIcon = AppKit.NSImage.alloc().initWithContentsOfFile_(toolbarIconPath)

#textWindowImagePath = os.path.join(skateboardBundle.resourcesPath(), "toolbar_explore.pdf")
#textWindowImage = AppKit.NSImage.alloc().initWithContentsOfFile_(textWindowImagePath)
#spaceWindowImagePath = os.path.join(skateboardBundle.resourcesPath(), "toolbar_space.pdf")
#spaceWindowImage = AppKit.NSImage.alloc().initWithContentsOfFile_(spaceWindowImagePath)

#cursorIconPath = "navigatorCursor.pdf"
#cursorIcon = AppKit.NSImage.alloc().initWithContentsOfFile_(cursorIconPath)
cursorIcon = AppKit.NSImage.alloc().initWithSize_((20, 20))
navigatorCursor = mojo.roboFont.CreateCursor(cursorIcon, (5,5))

class NavigatorTool(BaseEventTool):
    def setup(self):
        self.start = None
        self.end = None
        self.snapped = []
        self._xOptions =[]    # interesting things for the x axis cursor
        self._yOptions = []     # interesting things for the y axis cursor
        self._axisRange = 800    # in pixels the length a full axis is scaled to
        self.lastLocation = None
        self.draggingStartedLocation = None
        self.drawPreviewWithNavigator = None
        self.didDrag = False
        self.snapValues = None
        self.snapped = []
        self.interactionSources = {}
        self.previewDrawPref = None
        self.previewLocationPref = None
        self.previewMathPref = None
        self.previewRoundPref = False
        self.showSparksPref = False
        self.thick = 1
        self.diameter = 5
        self._maxSpaceWidth = 50
        self.caption = None
        self._captionTextSize = 8
        self._captionTextFont = "Menlo-Regular"
        self.isMouseDown = False
        self._cursorColor = (1,.5,0)    #(.7,.7,.7)

        if not hasattr(self, "RFPreviewCallback"):
            self.RFPreviewCallback = None
        if not hasattr(self, "axesContext"):
            self.axesContext = {}

    def checkColors(self):
        # check if we need to draw in bright or dark colors
        glyphBackgroundColor = getDefault("glyphViewBackgroundColor")
        r,g,b,a = glyphBackgroundColor
        if (r + g + b) * 0.33 < 0.2:
            self._cursorColor = (1,.6,0)    #(.7,.7,.7)
            #print('\tbackground is dark')
        elif (r + g + b) * 0.33 > 0.8:
            #print('\tbackground is bright')
            self._cursorColor = (1,.4,0)    #(.7,.7,.7)

    def becomeActive(self):
        self.checkColors()
        publishEvent("navigatorToolEvent.active")
        #publishEvent("navigatorToolEvent", data={}, callback=self.callbackForSkateboard, snapped=self.snapped)

    def becomeInactive(self):
        publishEvent("navigatorToolEvent.inactive")

    def getDefaultCursor(self):
        if self.isDragging() or self.isMouseDown:
            return navigatorCursor
        return super().getDefaultCursor()
        
    def mouseDown(self, point, event):
        # this is where we start dragging. 
        self.draggingStartedLocation = self.start = point.x, point.y
        self.didDrag = False
        self.isMouseDown = True
        self.mouseDragged(point, delta=(0,0))
        publishEvent("navigatorToolEvent.mousedown")

    def mouseUp(self, event):
        self.isMouseDown = False
        if self.didDrag:
            publishEvent("navigatorToolEvent.mouseup") #dragged?
        else:
            data = dict(point = self.start)
            publishEvent("navigatorToolEvent.click", data=data, callback=None)

        self.start = None
        self.end = None
        self.lastLocation = None
        #self.locationContext = None
        self.caption = None
        self.isDragged = False
    
    def _regularizeAxisValue(self, axisMin, axisCurrent, axisMax):
        return int(axisCurrent)
        
    def mouseDragged(self, point=None, delta=None):
        self.didDrag = True
        self.snapped = []
        if point is None:
            proposedEnd = self.end
        else:
            proposedEnd = point.x, point.y
        proposedDelta = [proposedEnd[0]-self.start[0], proposedEnd[1]-self.start[1]]
        new = {}
        caption = []
        horizontalAxes = self.interactionSources.get('horizontal', [])
        if horizontalAxes is None:
            horizontalAxes = []
        verticalAxes = self.interactionSources.get("vertical", [])
        if verticalAxes is None:
            verticalAxes = []
        new = {}
        self._xOptions = []
        self._yOptions = []
        if self.previewMathPref != "previewMutatorMath":
            clipping = True
        else:
            clipping = False
        for axisName in horizontalAxes:
            aMin, aDef, aCur, aMax = self.axesContext[axisName]
            axisScale = self._axisRange / (aMax - aMin)            
            pDScaled = proposedDelta[0]/axisScale
            if not clipping:
                new[axisName] = pDScaled+aCur
            else:
                if pDScaled+aCur > aMax:
                    new[axisName] = aMax
                elif pDScaled+aCur < aMin:
                    new[axisName] = aMin
                else:
                    new[axisName] = pDScaled+aCur
            if self.snapValues and self.shiftDown:
                for z, v in self.snapValues[axisName]:
                    if z[0] <= new[axisName] <= z[1]:
                        if not axisName in self.snapped:
                            self.snapped.append(axisName)
                        new[axisName] = v
            if self.shiftDown:
                new[axisName] = self._regularizeAxisValue(aMin, new[axisName], aMax)
            self._xOptions.append((aMin, new[axisName], aMax, axisName))
        for axisName in verticalAxes:
            aMin, aDef, aCur, aMax = self.axesContext[axisName]
            axisScale = self._axisRange / (aMax - aMin)            
            pDScaled = proposedDelta[1]/axisScale
            if not clipping:
                new[axisName] = pDScaled+aCur
            else:
                if pDScaled+aCur > aMax:
                    new[axisName] = aMax
                elif pDScaled+aCur < aMin:
                    new[axisName] = aMin
                else:
                    new[axisName] = pDScaled+aCur
            if self.snapValues and self.shiftDown:
                for z, v in self.snapValues[axisName]:
                    if z[0] <= new[axisName] <= z[1]:
                        if not axisName in self.snapped:
                            self.snapped.append(axisName)
                        new[axisName] = v
            if self.shiftDown:
                new[axisName] = self._regularizeAxisValue(aMin, new[axisName], aMax)
            self._yOptions.append((aMin, new[axisName], aMax, axisName))
        self.start = proposedEnd
        self.end = proposedEnd
        publishEvent("navigatorToolEvent", data=new, callback=self.callbackForSkateboard, snapped=self.snapped)
    
    def clear(self):
        # about to be uninstalled. Maybe the object is cached? to clear.
        #print("NavigatorTool.clear")
        self.axesContext = {}
        self._xOptions =[]    # interesting things for the x axis cursor
        self._yOptions = []     # interesting things for the y axis cursor
        self.snapValues = None
        self.lastLocation = None

    def callbackForSkateboard(self, data=None):
        # we're going to send this to skateboard and see what happens
        # skateboard navigatorToolLocationContextCallback
        loc = data.get('location')
        self.axesContext = data.get('context')
        self.interactionSources = data.get('interactionSources')
        self.previewDrawPref = data.get('previewDrawPref', '--')
        self.previewLocationPref = data.get('previewLocationPref', '--')
        self.previewMathPref = data.get("previewMathPref")
        self.showSparksPref = data.get("showSparksPref")
        self.previewRoundPref = data.get('previewRoundPref')
        self.drawPreviewWithNavigator = data.get('previewWithNavigator')
        self.drawPreviewElsewhere = data.get('previewElsewhere')
        self.useAxisMapPref = data.get('useAxisMaps')
        self.RFPreviewCallback = data.get('previewCallback')
        self.snapValues = data.get('snapValues')
        self.lastLocation = loc
        
    def drawPreview(self, thisScale):
        # weird callback:  the drawing will be done by skateboard.drawPreviewCallback
        if thisScale > 0:
            if self.RFPreviewCallback is not None:
                self.RFPreviewCallback(thisScale)
        
    def draw(self, thisScale):
        if not self.start: return
        if not self.end: return
        ctx.save()
        ctx.strokeWidth(self.thick*thisScale)
        r = self.diameter*thisScale*.5
        dx = self.end[0]-self.start[0]
        dy = self.end[1]-self.start[1]
        d = 1.5*thisScale
        ctx.translate(self.end[0], self.end[1])
        ctx.stroke(*self._cursorColor)
        ctx.oval(-d, -d, 2*d, 2*d)
        left = right = top = bottom = None
        dy = dx = 0        
        verticalNames = self.interactionSources.get('vertical', [])
        horizontalNames = self.interactionSources.get('horizontal', [])
        snapColor = (0,.4,.8)
        for xOptions in self._xOptions[:1]:
            left = xOptions[0]-xOptions[1]
            right = xOptions[2]-xOptions[1]
            ctx.save()
            ctx.translate(0,dy)
            scaleX = (self._maxSpaceWidth / (right-left))*thisScale
            if scaleX == 0:
                continue
            ctx.scale(scaleX)
            ctx.fill(*self._cursorColor)
            ctx.stroke(None)
            ts = self._captionTextSize*thisScale/scaleX
            ctx.font(self._captionTextFont)
            ctx.fontSize(ts)
            snap = False
            if verticalNames:
                isSnap = verticalNames[0] in self.snapped
                aMin, aDef, aCur, aMax = self.axesContext[verticalNames[0]]
                if len(verticalNames) == 1:
                    caption = "%s %3.3f" % (verticalNames[0], aCur)
                else:
                    caption = "%s %3.3f +" % (verticalNames[0], aCur)
                if not isSnap:
                    ctx.fill(*self._cursorColor)
                else:
                    ctx.fill(*snapColor)
                ctx.text(caption, (right+10*thisScale/scaleX,-1))
            else:
                isSnap = horizontalNames[0] in self.snapped
                aMin, aDef, aCur, aMax = self.axesContext[horizontalNames[0]]
                if len(horizontalNames) == 1:
                    caption = horizontalNames[0]
                else:
                    caption = horizontalNames[0]+" +"
                if not isSnap:
                    ctx.fill(*self._cursorColor)
                else:
                    ctx.fill(*snapColor)
                ctx.text("%s %3.3f" % (caption, aCur), (right+10*thisScale/scaleX,-1))
            ctx.strokeWidth(.7*thisScale/scaleX)
            if isSnap:
                ctx.stroke(*snapColor)
            else:
                ctx.stroke(*self._cursorColor)
            ctx.fill(None)
            ctx.line((left,0), (right, 0))
            dy += (10*thisScale)
            ctx.restore()
        for yOptions in self._yOptions[:1]:
            bottom = yOptions[0]-yOptions[1]
            top = yOptions[2]-yOptions[1]
            ctx.save()
            ctx.translate(dx,0)
            scaleY = (self._maxSpaceWidth / (top-bottom))*thisScale
            if scaleY == 0:
                continue
            ctx.scale(scaleY)
            ctx.fill(*self._cursorColor)
            ctx.stroke(None)
            ts = self._captionTextSize*thisScale/scaleY
            ctx.font(self._captionTextFont)
            ctx.fontSize(ts)
            isSnap = False
            if horizontalNames:
                isSnap = horizontalNames[0] in self.snapped
                aMin, aDef, aCur, aMax = self.axesContext[horizontalNames[0]]
                if len(horizontalNames) == 1:
                    caption = horizontalNames[0]
                else:
                    caption = horizontalNames[0]+" +"
                if not isSnap:
                    ctx.fill(*self._cursorColor)
                else:
                    ctx.fill(*snapColor)
                ctx.text("%s %3.3f" % (caption, aCur), (0, -1+bottom-18*thisScale/scaleY))
            else:
                isSnap = verticalNames[0] in self.snapped
                aMin, aDef, aCur, aMax = self.axesContext[verticalNames[0]]
                if len(verticalNames) == 1:
                    caption = verticalNames[0]
                else:
                    caption = verticalNames[0]+" +"
                if not isSnap:
                    ctx.fill(*self._cursorColor)
                else:
                    ctx.fill(*snapColor)
                ctx.text("%s %3.3f" % (caption, aCur), (0, -1+bottom-18*thisScale/scaleY))
            ctx.strokeWidth(.7*thisScale/scaleY)
            if isSnap:
                ctx.stroke(*snapColor)
            else:
                ctx.stroke(*self._cursorColor)
            #ctx.fill(None)
            ctx.line((0,bottom), (0, top))
            dx += (10*thisScale)/scaleY
            ctx.restore()
        if left is not None and right is not None and top is not None and bottom is not None:
            ctx.save()
            ctx.scale(scaleX, scaleY)
            if self.shiftDown:
                ctx.stroke(*snapColor)
            else:
                ctx.stroke(*self._cursorColor)
                #fill(0,0,0,0.025)
            ctx.fill(None)
            ctx.strokeWidth(1/scaleX)
            ctx.lineDash(3/scaleX, 3/scaleX)
            ctx.rect(left, bottom, right-left, top-bottom)
            ctx.restore()
        ctx.restore()

    def getToolbarTip(self):
        return 'Skateboard Navigator'

    def getToolbarIcon(self):
        ## return the toolbar icon
        return toolbarIcon
    
    def callback(self, item):
        pass
        #print(item)

    def additionContextualMenuItems(self):
        # get the states of the menus before we build the menu.
        # data = None lets skateboard know what we want.
        publishEvent("navigatorToolEvent", data=None, callback=self.callbackForSkateboard, snapped=self.snapped)
        prefRight = prefCentered = prefOrigin = 0
        prefWireframe = prefSolid = 0
        prefMutator = prefVarlib = 0
        prefSparks = 0
        prefRound = 0
        if self.previewLocationPref == "previewCentered":
            prefCentered = 1
        elif self.previewLocationPref == "previewOrigin":
            prefOrigin = 1
        elif self.previewLocationPref == "previewRight":
            prefRight = 1
        if self.previewDrawPref == "previewSolid":
            prefSolid = 1
        elif self.previewDrawPref == "previewWireframe":
            prefWireframe = 1
        if self.previewMathPref == "previewMutatorMath":
            prefMutator = 1
        elif self.previewMathPref == "previewVarLib":
            prefVarlib = 1
        if self.showSparksPref:
            prefSparks = 1
        if self.previewRoundPref:
            prefRound = 1
        else:
            prefRound = 0
        if self.drawPreviewWithNavigator:
            prefShowUsingNavigator = 1
        else:
            prefShowUsingNavigator = 0
        if self.drawPreviewElsewhere:
            prefShowUsingOtherTools = 1
        else:
            prefShowUsingOtherTools = 0
        if self.useAxisMapPref:
            prefUseAxisMaps = 1
        else:
            prefUseAxisMaps = 0
        m = [
                dict(title="Copy Preview to Clipboard", callback=self.navToolMenuCallback, state=0, ),
                #dict(title="Copy Preview to Current Layer", callback=self.navToolMenuCallback, state=0, ),
                AppKit.NSMenuItem.separatorItem(),
                #dict(title="Track Selection", callback=self.navToolMenuCallback, state=0, ),
                #AppKit.NSMenuItem.separatorItem(),
                dict(title="Current Layer to Support", callback=self.navToolMenuCallback, state=0),
                dict(title="New Instance At Location", callback=self.navToolMenuCallback, state=0, ),
                dict(title="New Support At Location", callback=self.navToolMenuCallback, state=0),
                dict(title="Mark Interesting Location", callback=self.navToolMenuCallback, state=0, ),
                #dict(title="Mute Supports", callback=self.navToolMenuCallback, state=0, ),
                AppKit.NSMenuItem.separatorItem(),
                dict(title="Copy Glyph to All Masters", callback=self.navToolMenuCallback ),
                AppKit.NSMenuItem.separatorItem(),
                #dict(title="Make Preview UFO", callback=self.navToolMenuCallback, state=0, ),
                #AppKit.NSMenuItem.separatorItem(),
                dict(title="Show Preview", items=[
                    dict(title="Using Other Tools", callback=self.navToolMenuCallback, state=prefShowUsingOtherTools, ),
                    dict(title="Using Navigator", callback=self.navToolMenuCallback, state=prefShowUsingNavigator, ),
                    ]),
                dict(title="Draw", items=[
                    dict(title="Origin", callback=self.navToolMenuCallback, state=prefOrigin, ),
                    dict(title="Centered", callback=self.navToolMenuCallback, state=prefCentered, ),
                    dict(title="Right", callback=self.navToolMenuCallback, state=prefRight, ),
                    AppKit.NSMenuItem.separatorItem(),
                    dict(title="Wireframe", callback=self.navToolMenuCallback, state=prefWireframe, ),
                    dict(title="Solid", callback=self.navToolMenuCallback, state=prefSolid, ),
                    dict(title="Sparks", callback=self.navToolMenuCallback, state=prefSparks, ),
                    ] ),
                dict(title="Math", items=[
                    dict(title="MutatorMath", callback=self.navToolMenuCallback, state=prefMutator, ),
                    dict(title="VarLib", callback=self.navToolMenuCallback, state=prefVarlib, ),
                    AppKit.NSMenuItem.separatorItem(),
                    dict(title="Round Geometry to Integers", callback=self.navToolMenuCallback, state=prefRound, ),
                    dict(title="Use Axis Maps", callback=self.navToolMenuCallback, state=prefUseAxisMaps, ),
                    ]),
                AppKit.NSMenuItem.separatorItem(),
                dict(title="Skateboard", callback=self.navToolMenuCallback, state=0),
                dict(title="Text Window", callback=self.navToolMenuCallback, state=0),
                dict(title="Space Window", callback=self.navToolMenuCallback, state=0),
                AppKit.NSMenuItem.separatorItem(),
                dict(title="Save PDF", callback=self.navToolMenuCallback, state=0, ),
            ]
        return m
            
    def navToolMenuCallback(self, sender):
        # XXX restructure
        
        if "Current Layer to Support" in sender.title():
            publishEvent("navigatorToolEvent.menu", item="currentLayerToSupport")
        elif "New Instance" in sender.title():
            publishEvent("navigatorToolEvent.menu", item="newInstance")
        elif "Mark Interesting" in sender.title():
            publishEvent("navigatorToolEvent.menu", item="storeLocation")
        elif "New Support" in sender.title():
            publishEvent("navigatorToolEvent.menu", item="newSupport")
        elif "Preview UFO" in sender.title():
            publishEvent("navigatorToolEvent.menu", item="makeUFO")
        elif "Copy Preview to Clipboard" in sender.title():
            publishEvent("navigatorToolEvent.menu", item="copyPreview")
        elif "Copy Preview to Current Layer" in sender.title():
            publishEvent("navigatorToolEvent.menu", item="copyPreviewToLayer")
        elif "PDF" in sender.title():
            publishEvent("navigatorToolEvent.menu", item="savePDF")
        #elif "Track" in sender.title():
        #    publishEvent("navigatorToolEvent.menu", item="trackSelection")
        elif "Next Master" in sender.title():
            publishEvent("navigatorToolEvent.menu", item="nextMaster")
        elif "Previous Master" in sender.title():
            publishEvent("navigatorToolEvent.menu", item="previousMaster")
        elif "Text" in sender.title():
            publishEvent("navigatorToolEvent.menu", item="openText")
        elif "Space" in sender.title():
            publishEvent("navigatorToolEvent.menu", item="openSpace")
        elif "Skateboard" in sender.title():
            publishEvent("navigatorToolEvent.menu", item="showSkateboard")
        elif "Left" in sender.title():
            publishEvent("navigatorToolEvent.menu", item="previewLeft")
        elif "Origin" in sender.title():
            publishEvent("navigatorToolEvent.menu", item="previewOrigin")
        elif "Centered" in sender.title():
            publishEvent("navigatorToolEvent.menu", item="previewCentered")
        elif "Right" in sender.title():
            publishEvent("navigatorToolEvent.menu", item="previewRight")
        elif "Wireframe" in sender.title():
            publishEvent("navigatorToolEvent.menu", item="previewWireframe")
        elif "Solid" in sender.title():
            publishEvent("navigatorToolEvent.menu", item="previewSolid")
        elif "MutatorMath" in sender.title():
            publishEvent("navigatorToolEvent.menu", item="previewMutatorMath")
        elif "VarLib" in sender.title():
            publishEvent("navigatorToolEvent.menu", item="previewVarLib")
        elif "Round Geometry" in sender.title():
            publishEvent("navigatorToolEvent.menu", item="previewRound")
        elif "Using Other Tools" in sender.title():
            publishEvent("navigatorToolEvent.menu", item="previewElsewhere")
        elif "Using Navigator" in sender.title():
            publishEvent("navigatorToolEvent.menu", item="previewWithNavigator")
        elif "Use Axis Maps" in sender.title():
            publishEvent("navigatorToolEvent.menu", item="useAxisMaps")
        elif "Sparks" in sender.title():
            self.showSparksPref = not self.showSparksPref
            publishEvent("navigatorToolEvent.menu", item="showSparks")
        elif "Copy Glyph to All Masters" in sender.title():
            publishEvent("navigatorToolEvent.menu", item="copyCurrentToAllMasters")
        else:
            print("ignored menu", sender.title(), sender.tag())
            
if __name__ == "__main__":
    import time
    nt = NavigatorTool()
    installTool(nt)
    print(time.time(),"refreshed navigator")
