from getDesignSpaceWindow import getSkateboardDesignSpace, ShowFontWindow
import designspaceProblems
from mojo.canvas import CanvasGroup
from mojo.UI import ProgressBar, OpenGlyphWindow
import ufoProcessor
#from mojo.drawingTools import *
import vanilla
import os

class ProblemWindow(object):
    def __init__(self, ds):
        # get our own copy of designspace
        self.path = ds.path
        #self.ds = ds
        self.currentProblem = None
        self._selectedGlyphs= []
        self.w = vanilla.Window((800, 500),
            "Skateboard: Designspace Problems",
            minSize=((500,300))
            )
        self.pane_report = vanilla.Group((0,0,0,0))
        fileIconWidth = 30
        toolbarHeight = 24
        reportColumns = [
                {   'title': '',
                    'key':'problemIcon',
                    'width':fileIconWidth,
                    'editable':False,
                },
                {   'title': '',
                    'key':'problemCode',
                    'width':70,
                    'editable':False,
                },
                {   'title': 'Category',
                    'key':'problemClass',
                    'width':70,
                    'editable':False,
                },
                {   'title': 'Description',
                    'key':'problemDescription',
                    'width':300,
                    'editable':False,
                },
                {   'title': 'Data',
                    'key':'problemData',
                    'width':500,
                    'editable':False,
                },
            ]

        self.pane_report.report = vanilla.List((0,toolbarHeight,-0,0),
            columnDescriptions=reportColumns,
            items=[],
            selectionCallback=self.selectProblemCallback
            )
        self.pane_report.validateButton = vanilla.Button((10, 0, 100, 20), "Validate", callback=self.validate)
        self.pane_report.openGlyphButton = vanilla.Button((120, 0, 100, 20), "Open glyphs", callback=self.openGlyphsCallback)
        self.pane_report.openGlyphButton.enable(False)
        self.pane_report.markGlyphButton = vanilla.Button((230, 0, 100, 20), "Mark glyphs", callback=self.markGlyphsCallback)
        self.pane_report.markGlyphButton.enable(False)
        self.pane_report.selectGlyphButton = vanilla.Button((340, 0, 100, 20), "Select glyphs", callback=self.selectGlyphsCallback)
        self.pane_report.selectGlyphButton.enable(False)
        
        textCaptionPos = (20, 10, 0, 20)
        self.pane_tools = vanilla.Group((0,0,0,0))
        tabNames = ["File", "Geometry", "Sources", "Instances", "Glyphs", "Kerning", "FontInfo", "Rules"]
        self.pane_tools.tabs = vanilla.Tabs((10, 10, -10, -10), tabNames, showTabs=False)
        self.tool_file = self.pane_tools.tabs[0]
        self.tool_file.text = vanilla.TextBox(textCaptionPos, "tool_file")
        self.tool_geometry = self.pane_tools.tabs[1]
        self.tool_geometry.text = vanilla.TextBox(textCaptionPos, "tool_geometry")
        self.tool_sources = self.pane_tools.tabs[2]
        self.tool_sources.text = vanilla.TextBox(textCaptionPos, "tool_sources")
        self.tool_instances = self.pane_tools.tabs[3]
        self.tool_instances.text = vanilla.TextBox(textCaptionPos, "tool_instances")
        self.tool_glyphs = self.pane_tools.tabs[4]
        self.tool_glyphs.text = vanilla.TextBox(textCaptionPos, "tool_glyphs")
        #self.tool_glyphs.canvas = CanvasGroup((0,50,0,0), delegate=self)
        self.tool_glyphs.fixInPrepolatorButton = vanilla.Button((-150, 10, 140, 20), "Open in Prepolator")
        #self.tool_glyphs.canvas.show(False)
        self.tool_kerning = self.pane_tools.tabs[5]
        self.tool_kerning.text = vanilla.TextBox(textCaptionPos, "tool_kerning")
        self.tool_fontinfo = self.pane_tools.tabs[6]
        self.tool_fontinfo.text = vanilla.TextBox(textCaptionPos, "tool_fontinfo")
        self.tool_rules = self.pane_tools.tabs[7]
        self.tool_rules.text = vanilla.TextBox(textCaptionPos, "tool_rules")

        #self.pane_canvas = vanilla.Group((0,0,0,0))
        #self.pane_canvas.show(False)
        #self.pane_canvas.canvas = CanvasGroup((0,0,0,0), delegate=self)
        
        
        h = 100
        paneDescriptors = [
            dict(   view=self.pane_report,
                    identifier="pane_report",
                    size=h * 2,
                    minSize=100,
                    canCollapse=False),
            # dict(   view=self.pane_tools,
            #         identifier="pane_tools",
            #         size=h * 2,
            #         minSize=100,
            #         canCollapse=False),
        ]
        self.w.panes = vanilla.SplitView((0, 30, -0, -30),
            paneDescriptors,
            isVertical=False,
            dividerThickness=6,
            dividerStyle="thin",
            autosaveName = "com.letterror.skateboard.problems",
            )
        #self.w.bind("close", self.mainWindow_closing)
        #self.w.bind("became main", self.validate)
        self.w.open()
        self.validate()
    
    def getAllSources(self):
        paths = []
        for f in self.ds.fonts.values():
            if f.path not in paths:
                paths.append(f.path)
        fonts = []
        for f in AllFonts():
            if f.path in paths:
                fonts.append(f)
        return fonts
        
    def openGlyphsCallback(self, sender):
        print('open glyphs', self._selectedGlyphs)
        fonts = self.getAllSources()
        for name in self._selectedGlyphs:
            for f in fonts:
                if name in f:
                    OpenGlyphWindow(f[name], newWindow=True)
        
    def markGlyphsCallback(self, sender):
        print('mark glyphs', self._selectedGlyphs)
        self.getAllSources()
        fonts = self.getAllSources()
        for f in fonts:
            for name in self._selectedGlyphs:
                if name in f:
                    f[name].markColor = (1,0,0,0.4)
            f.update()
            ShowFontWindow(f)
        
    def selectGlyphsCallback(self, sender):
        print('select glyphs', self._selectedGlyphs)
        self.getAllSources()
        fonts = self.getAllSources()
        for f in fonts:
            f.selectedGlyphNames = self._selectedGlyphs
            ShowFontWindow(f)
        
    def selectProblemCallback(self, sender):
        self.pane_report.markGlyphButton.enable(False)
        self.pane_report.openGlyphButton.enable(False)
        self.pane_report.selectGlyphButton.enable(False)
        sel = sender.getSelection()
        self._selectedGlyphs= []
        if len(sel) == 1:
            p = self.pane_report.report[sel[0]]['problem']
            self.currentProblem = p
        if len(sel) > 0:
            # mark selected problems
            # open selected glyphs
            pass
        print(self.currentProblem)
        for s in sel:
            p = self.pane_report.report[s]['problem']
            if 'glyphName' in p.data.keys():
                name = p.data.get('glyphName')
                if name not in self._selectedGlyphs:
                    self._selectedGlyphs.append(name)
        if self._selectedGlyphs:
            self.pane_report.markGlyphButton.enable(True)
            self.pane_report.openGlyphButton.enable(True)
            self.pane_report.selectGlyphButton.enable(True)
            print(self._selectedGlyphs)                
            
    def validate(self, sender=None):
        # validate with the designspaceProblems checker
        #print('self.ds', self.ds)
        print('validating', sender)

        self.pane_report.openGlyphButton.enable(False)
        self.pane_report.report.set([])
        bar = ProgressBar()
        print('bar', bar)
        try:
            self.ds = ufoProcessor.DesignSpaceProcessor()
            self.ds.read(self.path)
            bar.label("Loading fonts…")
            self.ds.loadFonts()
            checker = designspaceProblems.DesignSpaceChecker(self.ds)
            print('checker', checker)
            bar.label("Checking designspace geometry…")
            checker.checkDesignSpaceGeometry()
            bar.label("Checking sources…")
            checker.checkSources()
            bar.label("Checking instances…")
            checker.checkInstances()
            #print('hasStructuralProblems', checker.hasStructuralProblems())
            if not checker.hasStructuralProblems():
                # font specific
                self.ds.loadFonts()
                checker.nf = self.ds.getNeutralFont()
                bar.label("Checking kerning…")
                checker.checkKerning()
                bar.label("Checking font info…")
                checker.checkFontInfo()
                bar.label("Checking glyphs…")
                checker.checkGlyphs()
                bar.label("Checking rules…")
                checker.checkRules()

            instanceNames = []
            for instanceDescriptor in self.ds.instances:
                if instanceDescriptor.path is not None:
                    n = []
                    n.append(os.path.basename(instanceDescriptor.path))
                    n.append("({}-{})".format(instanceDescriptor.familyName, instanceDescriptor.styleName))
                    instanceNames.append(" ".join(n))

            report = []
            bar.label("Sorting results…")
            for problem in checker.problems:
                #print('problem', problem)
                icon = ""
                cat, desc = problem.getDescription()
                if problem.category in [0,1,2]:
                    icon="❗️"
                elif problem.category in [3,4,5]:
                    icon="❕"
                data = ""
                if problem.data:
                    t = []
                    for k, v in problem.data.items():
                        if k == "fontObj":
                            k = "font"
                            v = os.path.basename(v.path)
                        if k == "unicodes":
                            new = ["0x{:X}".format(a) for a in v if a is not None]
                            if None in v:
                                new.append("None")
                            v = ", ".join(new)
                        if k == "instances":
                            names = []
                            for i in v:
                                if hasattr(i, "familyName"):
                                    # make sure we're looking at an instanceDescriptor
                                    if i.filename is not None:
                                        path = " ({})".format(os.path.basename(i.filename))
                                    else:
                                        path = ""
                                    names.append("{} {}{}".format(i.familyName, i.styleName, path))
                            v = ", ".join(names)
                        t.append("%s: %s" % (k, str(v)))
                    data = ", ".join(t)
                d = dict(problemIcon=icon, problemCode="{}.{}".format(problem.category, problem.problem), problemClass=cat, problemDescription=desc, problemData=data, problem=problem)
                report.append(d)
            self.pane_report.report.set(report)
        finally:
            if bar is not None:
                bar.close()

    def draw(self):
        pass
        # draw pretty things for the problems
        #print(self.currentProblem)
        #rect(10,10,100,100)
        #print('aa')

if __name__ == "__main__":
    ds = getSkateboardDesignSpace()
    print(ds)
    if ds is not None:
        a = ProblemWindow(ds)
        a.validate()
        
        