# small window that shows which events Skateboard is publishing
import AppKit
import vanilla
import time, os
from mojo.events import addObserver, publishEvent, removeObserver
import mojo.extensions
from mutatorMath.objects.mutator import Location

skateboardBundle = mojo.extensions.ExtensionBundle("SkateBoard")

class SkateboardEventTester(object):
    toolNameForTitle = "⌘Skateboard"
    def __init__(self):
        self._documentIconImagePath = os.path.join(skateboardBundle.resourcesPath(), "skateboard_icon_512.pdf")
        self.w = vanilla.Window((300, 300), "%s Event printer" % self.toolNameForTitle)
        self._documentIconImage = AppKit.NSImage.alloc().initWithContentsOfFile_(self._documentIconImagePath)
        self.w.logo = vanilla.ImageView((0,0,0,0))    #, scale="proportional")
        self.w.logo.setImage(imageObject = self._documentIconImage)
        self.w.bind("close", self.cbClose)
        self.events = []
        addObserver(self, "previewLocationChangedCallback", 'skateboard.previewLocation.changed')
        addObserver(self, "previewLocationChangedCallback", "skateboard.interactionSources.changed")
        addObserver(self, "previewLocationChangedCallback", "skateboard.window.opened")
        addObserver(self, "previewLocationChangedCallback", "skateboard.window.closed")
        addObserver(self, "previewLocationChangedCallback", "skateboard.document.opened")
        addObserver(self, "previewLocationChangedCallback", "skateboard.document.closed")
        addObserver(self, "previewLocationChangedCallback", "skateboard.running.ping")
        addObserver(self, "glyphProblemCallback", "skateboard.problems.glyph")
        
        print("-----", time.time(), "Skateboard EventTester start")
        publishEvent("skateboard.running")
        self.w.open()
    
    def cbClose(self, sender=None):
        removeObserver(self, "skateboard.previewLocation.changed")
        removeObserver(self, "skateboard.interactionSources.changed")
        removeObserver(self, "skateboard.window.opened")
        removeObserver(self, "skateboard.window.closed")
        removeObserver(self, "skateboard.document.opened")
        removeObserver(self, "skateboard.document.closed")
        removeObserver(self, "skateboard.running.ping")
        print("-----", time.time(), "Skateboard EventTester stop")
    
    def glyphProblemCallback(self, notification):
        # notification when a glyph fails
        print("-----", time.time(), "compatibility failure glyph", notification['glyphName'])
        
    def previewLocationChangedCallback(self, notification):
        #print("previewLocationChangedCallback", notification)
        loc = notification.get('location')
        if loc is not None:
            loc = Location(loc).asString()
        else:
            loc = "no location"
        ds = notification.get('designspace')
        if ds is None:
            ds = "no document"
        print("-----", time.time(), notification['interactionSources'], notification['notificationName'])    #, loc, ds)

if __name__ == "__main__":
    OpenWindow(SkateboardEventTester)
