# a new structure for skateboard.

from defconAppKit.windows.baseWindow import BaseWindowController
import vanilla
import AppKit
import mojo
import os, importlib

import ufoProcessor
import ufoProcessor.varModels
#import ufoProcessor.sp3
from ufoProcessor.emptyPen import checkGlyphIsEmpty

import fontTools.designspaceLib as _designspaceLib
from fontTools.pens.areaPen import AreaPen
from fontTools.designspaceLib import DesignSpaceDocument, SourceDescriptor, InstanceDescriptor, AxisDescriptor, RuleDescriptor, processRules
from fontTools.varLib.models import normalizeValue, normalizeLocation
from mutatorMath.objects.mutator import Location, buildMutator
import mutatorMath.objects.mutator
from fontMath.mathGlyph import MathGlyph

from getDesignSpaceWindow import CurrentDesignSpace, getDesignSpaceDocuments
import textWindow
#importlib.reload(textWindow)
import spaceWindow
#importlib.reload(spaceWindow)

import storageObject

import traceback

from getDesignSpaceWindow import CurrentDesignSpace, getDesignSpaceDocuments
import skateboardDesignspaceProcessor
importlib.reload(skateboardDesignspaceProcessor)

from mojo.UI import (GlyphWindowToPDF,
    PutFile,
    ProgressBar,
    CurrentFontWindow,
    CurrentGlyphWindow,
    OpenGlyphWindow,
    setDefault,
    getDefault,
    AllFontWindows,
    AllGlyphWindows,
    UpdateCurrentGlyphView,
    OpenSpaceCenter,
    AskString,
    AskYesNoCancel,
    GetFile,
    SetCurrentLayerByName)
from mojo.UI import getDefault

DEVELOP = False


try:
    from variableFontGenerator import BatchDesignSpaceProcessor
    hasVariableFontGenerator = True
except ImportError:
    hasVariableFontGenerator = False
try:
    from designSpaceEditorwindow import DesignSpaceEditor
    hasDesignSpaceEditor = True
except ImportError:
    hasDesignSpaceEditor = False

try:
    import glyphNameFormatter
    hasGlyphNameFormatter = True
except ImportError:
    hasGlyphNameFormatter = False

try:
    import designspaceProblems
    hasDesignspaceProblems = True
except ImportError:
    hasDesignspaceProblems = False



if DEVELOP:
    pathForBundle = os.path.dirname(__file__)
    resourcePathForBundle = os.path.join(os.path.dirname(pathForBundle), "resources")
    skateboardBundle = mojo.extensions.ExtensionBundle(path=pathForBundle, resourcesName=resourcePathForBundle)
else:
    skateboardBundle = mojo.extensions.ExtensionBundle("SkateBoard")


class SkateBoardWindow(BaseWindowController):
    toolNameForTitle = "⌘Skateboard"
    buttonWidth = 60
    vanillaWindowWidth = 500

    # the structure of the window
    # the panes
    
    def __init__(self):
        self.document = SkateboardDocumentController(self)
        self.mainWindow_loadImageResources()
        self.mainWindow_buildVanillaWindow()
        self.w.open()
        self.mainWindow_enableInterface(False)

    # main_window
    def mainWindow_loadImageResources(self):
        self._emptyImagePath = os.path.join(skateboardBundle.resourcesPath(), "ufo_is_empty.pdf")
        self._emptyImage = AppKit.NSImage.alloc().initWithContentsOfFile_(self._emptyImagePath)
        self._ufoNeutralImagePath = os.path.join(skateboardBundle.resourcesPath(), "ufo_is_neutral.pdf")
        self._ufoNeutralImage = AppKit.NSImage.alloc().initWithContentsOfFile_(self._ufoNeutralImagePath)
        self._ufoMasterImage = AppKit.NSImage.alloc().initWithSize_((20, 20))

        self._axisControlImage = AppKit.NSImage.alloc().initWithSize_((20, 20))
        self._axisControlImage = self._ufoNeutralImage

        self._openDocImagePath = os.path.join(skateboardBundle.resourcesPath(), "button_txt_open.pdf")
        self._openDocImage = AppKit.NSImage.alloc().initWithContentsOfFile_(self._openDocImagePath)
        self._closeDocImagePath = os.path.join(skateboardBundle.resourcesPath(), "button_txt_close.pdf")
        self._closeDocImage = AppKit.NSImage.alloc().initWithContentsOfFile_(self._closeDocImagePath)
        self._saveDocImagePath = os.path.join(skateboardBundle.resourcesPath(), "button_txt_save.pdf")
        self._saveDocImage = AppKit.NSImage.alloc().initWithContentsOfFile_(self._saveDocImagePath)

        self._ufoMutedImagePath = os.path.join(skateboardBundle.resourcesPath(), "ufo_is_muted.pdf")
        self._ufoMutedImage = AppKit.NSImage.alloc().initWithContentsOfFile_(self._ufoMutedImagePath)
        self._snappedImagePath = os.path.join(skateboardBundle.resourcesPath(), "snapped.pdf")
        self._snappedImage = AppKit.NSImage.alloc().initWithContentsOfFile_(self._snappedImagePath)
        self._ufoAvailableImagePath = os.path.join(skateboardBundle.resourcesPath(), "ufo_is_available.pdf")
        self._ufoAvailableImage = AppKit.NSImage.alloc().initWithContentsOfFile_(self._ufoAvailableImagePath)
        self._ufoOpenImagePath = os.path.join(skateboardBundle.resourcesPath(), "ufo_is_open.pdf")
        self._ufoOpenImage = AppKit.NSImage.alloc().initWithContentsOfFile_(self._ufoOpenImagePath)
        self._ufoCurrentImagePath = os.path.join(skateboardBundle.resourcesPath(), "ufo_is_current.pdf")
        self._ufoCurrentImage = AppKit.NSImage.alloc().initWithContentsOfFile_(self._ufoCurrentImagePath)
        self._ufoMissingImagePath = os.path.join(skateboardBundle.resourcesPath(), "ufo_is_missing.pdf")
        self._ufoMissingImage = AppKit.NSImage.alloc().initWithContentsOfFile_(self._ufoMissingImagePath)
        
        self._documentIconImagePath = os.path.join(skateboardBundle.resourcesPath(), "logo_designspace_5.pdf")
        self._documentIconImage = AppKit.NSImage.alloc().initWithContentsOfFile_(self._documentIconImagePath)

        self._newGlyphImagePath = os.path.join(skateboardBundle.resourcesPath(), "button_explore.pdf")
        self._newGlyphImage = AppKit.NSImage.alloc().initWithContentsOfFile_(self._newGlyphImagePath)

        self._textButtonImagePath = os.path.join(skateboardBundle.resourcesPath(), "button_txt_text.pdf")
        self._textButtonImage = AppKit.NSImage.alloc().initWithContentsOfFile_(self._textButtonImagePath)
        self._spaceButtonImagePath = os.path.join(skateboardBundle.resourcesPath(), "button_txt_space.pdf")
        self._spaceButtonImage = AppKit.NSImage.alloc().initWithContentsOfFile_(self._spaceButtonImagePath)
        self._exportButtonImagePath = os.path.join(skateboardBundle.resourcesPath(), "button_txt_ufo.pdf")
        self._exportButtonImage = AppKit.NSImage.alloc().initWithContentsOfFile_(self._exportButtonImagePath)
        self._addGlyphButtonImagePath = os.path.join(skateboardBundle.resourcesPath(), "button_txt_glyph.pdf")
        self._addGlyphButtonImage = AppKit.NSImage.alloc().initWithContentsOfFile_(self._addGlyphButtonImagePath)

        self._makeFontImagePath = os.path.join(skateboardBundle.resourcesPath(), "button_txt_variablefont.pdf")
        self._makeFontButtonImage = AppKit.NSImage.alloc().initWithContentsOfFile_(self._makeFontImagePath)
        self._editDocImagePath = os.path.join(skateboardBundle.resourcesPath(), "button_txt_edit.pdf")
        self._editDocButtonImage = AppKit.NSImage.alloc().initWithContentsOfFile_(self._editDocImagePath)

    def mainWindow_enableInterface(self, state=True):
        # $$
        print('mainWindow_enableInterface', state)
        if state:
            # enable everything
            self.pane_doc.l.enable(True)
            self.pane_previewAxes.locationList.enable(True)
            #self.pane_locs.filterLocations.enable(True)
            #self.pane_locs.locationsList.enable(True)
            #self.locationsList_setFilterLocationButton()
            #self.pane_previewAxes.show(True)
            #self.pane_locs.show(True)
            #self.pane_doc.show(True)
            self.w.panes.show(True)
            self.w.file.saveButton.show(True)
            self.w.frontPage.show(False)
            self.w.file.textWindowButton.show(True)
            self.w.file.spaceWindowButton.show(True)
            self.w.file.closeButton.show(True)
            self.w.debugButton.show(True)
            #self.w.frontPage.openButton.show(False)
            self.w.file.exportUFOButton.show(True)
            #self.w.file.addGlyphButton.show(True)
            #if hasDesignSpaceEditor:
            #    self.w.file.editDSButton.show(True)
            #if hasVariableFontGenerator:
            #    self.w.file.makeVFButton.show(True)            
        else:
            # disable everything
            #self.pane_doc.l.enable(False)
            #self.pane_previewAxes.locationList.enable(False)
            #self.pane_locs.filterLocations.enable(False)
            #self.pane_locs.locationsList.enable(False)
            #self.locationsList_setFilterLocationButton(reset=True)
            #self.pane_previewAxes.show(False)
            #self.pane_locs.show(False)
            #self.pane_doc.show(False)
            self.w.file.saveButton.show(False)
            self.w.frontPage.show(True)
            self.w.file.textWindowButton.show(False)
            self.w.file.spaceWindowButton.show(False)
            self.w.file.closeButton.show(False)
            #self.w.frontPage.openButton.show(True)
            self.w.file.exportUFOButton.show(False)
            #self.w.file.addGlyphButton.show(False)
            self.w.panes.show(False)
            self.w.debugButton.show(False)
            if hasDesignSpaceEditor:
               self.w.file.editDSButton.show(False)
            if hasVariableFontGenerator:
               self.w.file.makeVFButton.show(False)      
                
    def mainWindow_buildVanillaWindow(self):

        start_fileThings = 0
        height_fileThings = 40
        
        start_sourceThings = start_fileThings + height_fileThings
        height_sourceThings = 170
        
        start_previewThings = start_sourceThings + height_sourceThings
        height_previewThings = 150
        
        start_locationThings = start_previewThings + height_previewThings
        height_locationThings = 150

        self.w = vanilla.Window((self.vanillaWindowWidth, 679),
            self.toolNameForTitle,
            minSize=(500, 500),
            maxSize=(800,2000),
            autosaveName = "com.letterror.skateboard.mainWindow",
            fullScreenMode = None,
            titleVisible = True,
            )

        ww = 50
        wy = 40

        self.w.frontPage = vanilla.Group((0,0,0,0))
        self.w.frontPage.logo = vanilla.ImageView((0,50,0,0), horizontalAlignment="center", verticalAlignment="center", scale="proportional")
        self.w.frontPage.logo.setImage(imageObject = self._documentIconImage)
        self.w.frontPage.openButton = vanilla.ImageButton((0, 0, self.buttonWidth, wy),
            imageObject=self._openDocImage,
            bordered=False, 
            callback=self.document.document_openDesignSpaceCallback,
            sizeStyle="regular")

        self.w.file = vanilla.Group((0, 0,0, height_fileThings))
        self.w.file.l = vanilla.HorizontalLine((0,-1, 0, 1))

        self.pane_doc = vanilla.Group((0, 0,0, 0))
        self.pane_previewAxes = vanilla.Group((0, 0, 0, 0))
        self.pane_locs = vanilla.Group((0, 0, 0, 0))

        sourceListColumnDescriptions = [
            {'title': '', 'key':"isMutedImage", 'width': 18, 'cell':vanilla.ImageListCell()},
            {'title': "", 'key':"isOpenImage", 'width': 18, 'cell':vanilla.ImageListCell()},
            {'title': "", 'key':"isNeutralImage", 'width': 18, 'cell':vanilla.ImageListCell()},
            #{'title': "id", 'key':"_fontid", 'width': 100},
            #{'title': "Object", 'key':"_fontObject", 'width': 50},
            {'title': "UFO", 'key':"name", 'width': 280},
            #{'title': "Glyph", 'key':"_glyphObject", 'width': 100},
            {'title': "Layer", 'key':'layer', 'width': 200},
            #{'title': "Factor", 'key':'factor', 'width': 100}
            ]

        self.pane_doc.l = vanilla.List((0,30,0,0), [],
            columnDescriptions=sourceListColumnDescriptions,
            #doubleClickCallback=self.sourcesList_selectFontCallback,
            #selectionCallback=self.sourcesList_selectionChangedCallback,
            #menuCallback = self.sourcesMenu_buildMenu,
            )
        self.pane_doc.l.getNSTableView().setUsesAlternatingRowBackgroundColors_(False)
        self.pane_doc.caption = vanilla.TextBox((5,5,-80,20), "Sources & Layers")
        #imagePath = "./resources/icon_16x16.png"
        self.pane_previewAxes.caption = vanilla.TextBox((5,5,-80,20), "Preview")
        self.previewAxesList_doNotUpdate = False    # flag: set to prevent previewAxesList_handleEditsCallback from updating all the time

        interactionSourcesPopup = vanilla.PopUpButtonListCell(self.document.interactionSourcesNames)
        previewAxesList_columnDescriptions = [
            {'title': '', 'key':"isMutedImage", 'width': 18, 'cell':vanilla.ImageListCell()},
            {'title': '', 'key':"isSnappedImage", 'width': 18, 'cell':vanilla.ImageListCell()},
            {'title': "Name", 'key':"axisKey", 'width': 90, 'editable':False},
            {'title': "Value", 'key':"valueKey", 'width': 80, 'editable':True},
            {'title': "Mapped", 'key':"mappedValueKey", 'width': 80, 'editable':False},
            {'title': "Normalised", 'key':"normalisedValueKey", 'width': 80, 'editable':False},
            #{'title': "Min", 'key':"minimumKey", 'width': 60, 'editable':False},
            #{'title': "Default", 'key':"defaultKey", 'width': 60, 'editable':False},
            #{'title': "Max", 'key':"maximumKey", 'width': 60, 'editable':False},
            {'title': "Interaction", 'key':"controlKey", 'width': 100, 'cell': interactionSourcesPopup,  "binding": "selectedValue"},
            ]

        self.pane_previewAxes.locationList = vanilla.List((0,30,0,0), [],
            columnDescriptions=previewAxesList_columnDescriptions,
            #editCallback=self.previewAxesList_handleEditsCallback,
            #allowsMultipleSelection=False,
            #menuCallback = self.previewAxesMenu_buildMenu,
            )

        ### group 2
        self.pane_locs = vanilla.Group((0, 0, 0, 0))
        self.pane_locs.caption = vanilla.TextBox((5, 5, 150,20), "Locations")

        locationListDescriptions = [
            #{'title': "", 'key':"control", 'width': 30},
            {'title': "•", 'key':"description", 'width': 350},
            {'title': "Flavor", 'key':"flavor", },#'width': 100},
            ]

        filterLocationDescriptions = [
            dict(title="Sources"),
            dict(title="Supports"),
            dict(title="Instances"),
            dict(title="Interesting"),
            dict(title="Open"),
            ]
        self.pane_locs.filterLocations = vanilla.SegmentedButton((160, 3, -5, 22),
            segmentDescriptions=filterLocationDescriptions, selectionStyle="any",
            #callback = self.filterLocationsCallback
            )
        #self.locationsList_setFilterLocationButton()
        #self.locationsList_doNotUpdate = False
        self.pane_locs.locationsList = vanilla.List((0, 30, 0, 0), [],
            columnDescriptions=locationListDescriptions,
            #selectionCallback=self.locationsList_selectLocationCallback,
            #allowsMultipleSelection=True,
            #menuCallback = self.locationsMenu_buildMenu,
            )
        self.pane_locs.locationsList.getNSTableView().setUsesAlternatingRowBackgroundColors_(False)

        # column descriptions for the source list
        # sourceListColumnDescriptions = [
        #     {'title': '', 'key':"isMutedImage", 'width': 18, 'cell':vanilla.ImageListCell()},
        #     {'title': "", 'key':"isOpenImage", 'width': 18, 'cell':vanilla.ImageListCell()},
        #     {'title': "", 'key':"isNeutralImage", 'width': 18, 'cell':vanilla.ImageListCell()},
        #     #{'title': "id", 'key':"_fontid", 'width': 100},
        #     #{'title': "Object", 'key':"_fontObject", 'width': 50},
        #     {'title': "UFO", 'key':"name", 'width': 280},
        #     #{'title': "Glyph", 'key':"_glyphObject", 'width': 100},
        #     {'title': "Layer", 'key':'layer', 'width': 200},
        #     #{'title': "Factor", 'key':'factor', 'width': 100}
        #     ]

        # self.pane_doc.l = vanilla.List((0,30,0,0), [],
        #     columnDescriptions=sourceListColumnDescriptions,
        #     #doubleClickCallback=self.sourcesList_selectFontCallback,
        #     #selectionCallback=self.sourcesList_selectionChangedCallback,
        #     #menuCallback = self.sourcesMenu_buildMenu,
        #     )


        h = 100
        paneDescriptors = [
            dict(   view=self.pane_doc,
                    identifier="pane_doc",
                    size=h * 2,
                    minSize=100,
                    canCollapse=False),

            dict(   view=self.pane_previewAxes,
                    identifier="pane_previewAxes",
                    size=h * 1,
                    minSize=100,
                    canCollapse=False),

            dict(   view=self.pane_locs,
                    identifier="pane_locs",
                    size=h * 2,
                    minSize=100,
                    canCollapse=False),

            ]
        self.w.panes = vanilla.SplitView((0, 40, -0, -30),
            paneDescriptors,
            isVertical=False,
            #dividerThickness=6,
            dividerStyle="thin",
            autosaveName = "com.letterror.skateboard.panes",
            )


        buttonSlotsLeft = [self.buttonWidth * i for i in range(10)]
        buttonSlotsRight = [- self.buttonWidth * i for i in range(1,10)]
        self.w.file.closeButton = vanilla.ImageButton((buttonSlotsLeft[0], 0, self.buttonWidth, wy),
                imageObject=self._closeDocImage, bordered=False,
                callback=self.document.document_closeDesignSpaceCallback,
                sizeStyle="regular")
        self.w.file.saveButton = vanilla.ImageButton((buttonSlotsLeft[1], 0, self.buttonWidth, wy),
                imageObject=self._saveDocImage, bordered=False,
                callback=self.document.document_saveDesignSpaceCallback_checkForSaveAs,
                sizeStyle="regular")
        self.w.file.saveButton.enable(False)
        self.w.file.closeButton.show(False)
        self.w.file.spaceWindowButton = vanilla.ImageButton((buttonSlotsRight[1], 0, self.buttonWidth, wy),
                imageObject=self._spaceButtonImage, bordered=False,
                callback=self.mainWindow_openSpaceWindow,
                sizeStyle="regular")
        self.w.file.textWindowButton = vanilla.ImageButton((buttonSlotsRight[2], 0, self.buttonWidth, wy),
                imageObject=self._textButtonImage, bordered=False,
                callback=self.mainWindow_openTextWindow,
                sizeStyle="regular")
        self.w.file.exportUFOButton = vanilla.ImageButton((buttonSlotsRight[0], 0, self.buttonWidth, wy),
                imageObject=self._exportButtonImage, bordered=False,
                callback=self.document.export_generateUFOCallback,
                sizeStyle="regular")
        if hasVariableFontGenerator:
            self.w.file.makeVFButton = vanilla.ImageButton((buttonSlotsLeft[3], 0, self.buttonWidth, wy),
                imageObject=self._makeFontButtonImage, bordered=False,
                callback=self.document.export_makeVariableFontCallback,
                sizeStyle="regular")
        if hasDesignSpaceEditor:
            self.w.file.editDSButton = vanilla.ImageButton((buttonSlotsLeft[2], 0, self.buttonWidth, wy),
                imageObject=self._editDocButtonImage, bordered=False,
                callback=self.mainWindow_openDSEditorCallback,
                sizeStyle="regular")
    
        self.w.debugButton = vanilla.Button((5, -25, 60, 20), "Debugs", callback=self.debug_makeObjectBrowserCallback, sizeStyle="small")
        self.w.versionText = vanilla.TextBox((-50, -20, 100, 20), "v %s" % str(skateboardBundle.version), sizeStyle="small")

    def mainWindow_openTextWindow(self, sender):
        print("mainWindow_openTextWindow")

    def mainWindow_openSpaceWindow(self, sender):
        print("mainWindow_openSpaceWindow", sender)
    
    def mainWindow_openDSEditorCallback(self, sender):
        print("mainWindow_openDSEditorCallback", sender)

    def debug_makeObjectBrowserCallback(self, sender):
        try:
            methodNames = []
            m = [getattr(self, method) for method in dir(self)]
            for item in m:
                if type(item) == MethodType:
                    methodNames.append(item.__name__)
            methodNames.sort()
            t = time.strftime("%a, %d %b %Y %H:%M:%S", time.localtime())
            browserWindow = vanilla.Window((1000, 600), "Skateboard Debug %s" %(t), minSize=(400,500))
            data = {}
            if self.instanceLocation is not None:
                data['instanceLocation'] = Location(self.instanceLocation).asString()
            data['openFonts'] = [os.path.basename(p) for p in list(self.openFonts.keys())]
            data['needFonts'] = [os.path.basename(p) for p in list(self.needFonts.keys())]
            data['designspace document'] = self.ds
            if self.ds is not None:
                data['fonts'] = list([(k, v.path) for k, v in self.ds.fonts.items()])
                data['designspace path'] = self.ds.path
                data['document.sources'] = self.ds.sources
                data['document.instances'] = self.ds.instances
                data['document.rules'] = self.ds.rules
            data['math model'] = self.previewMathPref
            data['show sparks pref'] = self.showSparksPref
            data['rounding'] = self.previewRoundPref
            data['textWindow'] = self.textWindow
            data['spaceWindow'] = self.spaceWindow
            data['skateboardBundle.version'] = skateboardBundle.version
            data['previewGlyphName'] = self.currentGlyphName
            data['defcon glyphobservers'] = [str(g) for g in self._defconObservers_glyphs.values()]
            if self.checker:
                data['designspaceProblems'] = [str(p) for p in self.checker.problems]
            browserWindow.browser = vanilla.ObjectBrowser((0,0,0,0), data)
            browserWindow.open()
            
        except:
            print("Skateboard: Ironically, can't open the debug window because of a problem.", traceback.format_exc())

    # document
class SkateboardDocumentController(object):
    # this does all the work

    navigatorPrefsLibKey = "com.letterror.skateboard.navigator"
    superpolatorDataLibKey = "com.superpolator.data"    # used by superpolator
    interestingLocationLibKey = "com.letterror.skateboard.interestingLocation"
    interactionSourcesLibKey = "com.letterror.skateboard.interactionSources"
    mutedSourcesLibKey = "com.letterror.skateboard.mutedSources"
    mutedAxesLibKey = "com.letterror.skateboard.mutedAxes"
    previewLocationLibKey = "com.letterror.skateboard.previewLocation"
    previewTextLibKey = "com.letterror.skateboard.previewText"
    glyphSourceLocationLibKey = "com.letterror.skateboard.sourceLocation"    # the location a glyph outline originated from


    def __init__(self, window):
        self.locationsList_doNotUpdate = False
        self.window = window

        self.ds = None
        self.checker = None

        self.interactionSourcesNames = [ "vertical", "horizontal", "ignore"]    #, "animate" ]
        # this maps the identifier to a list of axis names
        self.interactionSources = {"vertical":[], "horizontal":[], "ignore":[]}

        self.maxCollectors = 10
        #self._lastFactors = []
        self._lastOpenNodes = None
        self._lastDrawingValues = None    # values collected in drwing the background
        self.pointLoad = 0      # total points in the preview
        self._maxPoints = 2000  # number of points we can handle without problems
        self.axisRange = 250   # arbitrary units to map interaction sources to 
        self.axisSteps = 100    # number of steps along one axis with cursor keys, should be pref
        self.verbose = False
        self._allowExtrapolation = False
        self.previewTextRaw = "SKATEBOARD"
        self.snapValues = {}
        self.snapped = []    # list of axisnames that are snapping at the moment
        self._listSources = False
        self._listSupports = True
        self._listInstances = True
        self._listInteresting = True
        self._listOpen = False
        self.documentNeedSave = None
        self.currentlyDragging = False
        self.currentSupport = None
        self.currentSource = None
        self.useAxisMaps = False
        self.axisMapper = None
        self.storage_mainPreviews = storageObject.StorageContainer()
        self.storage_supportPreviews = storageObject.StorageContainer()
        self.storage_fontInfo = storageObject.StorageContainer()

        # initialise attributes
        self.seriesLocations = []     # list for all locations in this document
        self.pointsOfInterest = []    # list of user picked locations
        self.sourceGlyphObjects = []    # for processing selections
        self.currentGlyphName = None
        self.currentFontPath = None    # the path of the current font 
        self.currentGlyph = None     # the glyph in the current edit window
        self.instanceGlyphObject = RGlyph()    # RGlyph object of the current instance
        self.instanceGlyphObject_noSupport = RGlyph()
        self.hasNoSupportGlyph = False
        self.contouredInstanceGlyphObject = RGlyph()    # RGlyph object with all the contours
        self.instanceFont = RFont(showInterface=False)
        self.instanceFont.info.familyName = "Skateboard Preview"
        self.instanceFont.info.styleName = ""
        self.instanceFont.info.unitsPerEm = 2000
        self.instanceFont.info.descender = -1000
        self.instanceFont.info.ascender = 750
        self.previewFonts = {}
        
        self.textMarginScale = 1.2     # scale factor for the vertical metrics of the text window. JUst some padding to make it fit.
        self.instanceSpaceCenter = None    #OpenSpaceCenter(self.instanceFont)
        self.previewLocationPref = "previewCentered"
        self.previewDrawPref = "previewSolid"
        self.previewMathPref = "previewVarlib"
        self.previewRoundPref = False
        self.drawingWithToolActivePref = False
        self.textGlyphNames = ["S", "K"]
        self.instanceFontNeededGlyphs = []
        self.selectNearOffCurves = True
        self.selectedSegments = []
        self.fontPaths = []
        self.glyphMutators = {}
        self.lastSupports = None
        self.openFonts = {}    # fonts, actually open. key:fontpath, value: fontobject
        self.needFonts = {}    # fonts, needed, but closed
        self.missingFonts = {} # fonts, missting
        self.mutedFonts = {}   # fonts, muted
        self.mutedAxes = {}    # axes, muted
        self.instanceLocation = {}    # preview location for the background
        self.instanceLocationMapped = {}
        self.allSelectedLocations = [] # list of all selected locations
        self.axisNames = []
        self.axisExtremes = {}
        self.lastSignal = None
        self.interval = None
        #self.somethingSelector = 0        # test for 3rd encoder
        
        self.button1State = False
        self.button2State = False
        self.button3State = False
        self.instanceKerningMutator = None
        self.instanceInfoMutator = None
        self.lastKerning = None    # the last kerning generated
        self.lastKerningPairs = None    # the pairs for the last kerning
        self.lastKerningLocation = None # the location for the last kerning
        
        self.currentSelection = {}
        self.currentSelectedClusters = {}
        self.trackedClusters = {}
        self.colinearTolerance = 100    # area of polygon
        self.currentGlyphViews = {}    # dict with glyphnames of all windows that are open

        self.textWindow = None
        self.spaceWindow = None
        self.spaceWindowData = {}
        self.textWindowMode = "text"
        self.slowAnimationTimer = None
        self.slowAnimationInterval = .2
        self.slowAnimationTimerTick = 0
        self.slowAnimationTimerRunning = False
        self.miniKBDScrubbingTimer = None
        self._miniKBDScrubbingTimerHits = 0
        self.miniKBDScrubbingTimerInterval = .45
        self.redrawTimer = None
        self._redrawTimerHits = 0
        self.redrawStillDraggingTimerInterval = .008
        self.mouseUpTimer = None
        self.textRedrawTimer = None
        self.mouseUpDraggingDoneTimerInterval = .05
        self.textRedrawInterval = .05
        self.textRedrawGlyphHistory = {}
        self.textRedrawOrder = []
        self.textRedrawComps = []
        self.drawDetails = True
        self.navigatorTool = None
        self.glyphChanges = {}    # keep track of edits
        self.glyphCache = {}     # cacher

        #self.localEventMonitor = AppKit.NSEvent.addLocalMonitorForEventsMatchingMask_handler_(AppKit.NSKeyDownMask, self.mainWindow_monitorLocalEventsCallback_)


    def document_closeDesignSpaceCallback(self, sender):
        print("SkateboardDocumentController document_closeDesignSpaceCallback", sender)
        # clicking the "Close" button
        self.mainWindow_closeDocument()

    def mainWindow_closeDocument(self):
        if self.documentNeedSave:
            # handle saving dialog
            result = AskYesNoCancel("Save Designspace?",
                title="Skateboard",
                informativeText="The current Designspace has unsaved changes.")
            if result == 1:
                # save dialog
                self.document_saveDesignSpaceCallback()
        self.mainWindow_closeDocument_closing()

    def mainWindow_closeDocument_closing(self):
        if self._defconObservers_glyphs:
            #print("mainWindow_closeDocument_closing - - - - - - - ")
            for i, g in self._defconObservers_glyphs.items():
                #%% defcon.glyph observer
                #print("\tremoving defcon observer for", g)
                #g.endSelfNotificationObservation()
                g.removeObserver(self, "Glyph.Changed")
                g.removeObserver(self, "Glyph.SelectionChanged")
        self._defconObservers_glyphs = {}
        if self.ds is not None:
            self.ds.update = None
        self.ds = None
        #self.w.frontPage.openButton.setTitle("Open")
        self.window.w.file.saveButton.enable(False)
        self.window.mainWindow_enableInterface(False)
        self.instanceLocation = None
        self.document_updateCurrentDesignSpace()
        self.data_instanceLocationChanged()
        self.pointsOfInterest = []
        if self.textWindow is not None:
            self.textWindow.mainWindowIsClosing()
        if self.spaceWindow is not None:
            self.spaceWindow.mainWindowIsClosing()
        if self.navigatorTool is not None:
            self.navigatorTool.RFPreviewCallback = None
            uninstallTool(self.navigatorTool)
            self.navigatorTool = None
        publishEvent("skateboard.document.closed",
            location=None,
            designspace=None,
            interactionSources=None,
            axes=None,
            instanceGlyph = self.instanceGlyphObject,
            previewGlyph = self.contouredInstanceGlyphObject,
            )
        self.storage_mainPreviews = StorageContainer()
        self.storage_supportPreviews = StorageContainer()
        self.storage_kerning = StorageContainer()
        self.storage_fontInfo = StorageContainer()

        self.window.w.getNSWindow().setRepresentedURL_(None)
        self.window.w.setTitle(self.toolNameForTitle)
        self.currentGlyphViews = {}
        self.document_modificationTime = None

        app = AppKit.NSApp()
        app.getNextSkateboardMasterCallback = None
        #print('--- app.getNextSkateboardMasterCallback removed')

    def document_saveDesignSpaceCallback(self, sender=None):
        # vanilla callback
        # save the current designsace.
        try:
            interestingItems = [(Location(a).asDict(), b) for a, b in self.pointsOfInterest]
            if interestingItems:
                self.ds.lib[self.interestingLocationLibKey] = interestingItems
            # save the prefs
            if not self.interactionSourcesLibKey in self.ds.lib:
                self.ds.lib[self.interactionSourcesLibKey] = {}
            self.ds.lib[self.interactionSourcesLibKey] = self.interactionSources
            # muted fonts
            mutedFontsForSave = []
            for path, layerName in self.mutedFonts.keys():
                fileName = os.path.basename(path)
                if layerName is None:
                    layerName = "foreground"
                mutedFontsForSave.append((fileName, layerName))
            if mutedFontsForSave:
                self.ds.lib[self.mutedSourcesLibKey] = mutedFontsForSave
            else:
                # remove the key from the lib, otherwise old data remains
                if self.mutedSourcesLibKey in self.ds.lib:
                    del self.ds.lib[self.mutedSourcesLibKey]
            # muted axes
            if self.mutedAxes:
                self.ds.lib[self.mutedAxesLibKey] = list(self.mutedAxes.keys())
            else:
                # remove the key from the lib, otherwise old data remains
                if self.mutedAxesLibKey in self.ds.lib:
                    del self.ds.lib[self.mutedAxesLibKey]
            self.ds.lib[self.previewLocationLibKey] = Location(self.instanceLocation).asDict()
            self.ds.lib[self.previewTextLibKey] = self.previewTextRaw
            self.ds.write(self.ds.path)
            self.document_setNeedSave(False)
            self.document_modificationTime = self.document_getModificationTime()
        except:
            print("Skateboard: Problem while saving..", traceback.format_exc())
    
    def document_saveDesignSpaceCallback_checkForSaveAs(self, sender):
        print("SkateboardDocumentController document_saveDesignSpaceCallback_checkForSaveAs", sender)
    
    def document_openDesignSpaceCallback(self, sender):
        print("SkateboardDocumentController document_openDesignSpaceCallback", sender)

    def document_openDesignSpaceCallback(self, sender):
        # clicking the "Open" button
        resultPath = GetFile(message="Open a designspace or sp3 file",
            title="Open", directory=None, fileName=None,
            allowsMultipleSelection=False, fileTypes=["designspace", 'sp3'])
        if resultPath is None:
            if self.ds is not None:
                # we had one open, went to the open dialog and changed our mind. 
                return
            self.ds = None
        else:
            self.mainWindow_openDocumentWithPath(resultPath)
        
    def export_generateUFOCallback(self ,sender):
        print("SkateboardDocumentController export_generateUFOCallback", sender)
    
    def export_makeVariableFontCallback(self, sender):
        print("SkateboardDocumentController export_makeVariableFontCallback", sender)

    def mainWindow_openDocumentWithPath(self, path):
        # open the document at this path
        print("SkateboardDocumentController mainWindow_openDocumentWithPath")
        ext = os.path.splitext(path)[-1]
        if ext == ".designspace":
            self._openDesignspaceDocument(path)
        elif ext == ".sp3":
            self._convertSuperpolatorDocument(path)
    
    def _openDesignspaceDocument(self, path):
        print("SkateboardDocumentController _openDesignspaceDocument", path)
        # first, make sure we can reasonably ready this file
        shouldOpen = True
        msg = "This designspace has structural problems."
        if hasDesignspaceProblems:
            # we should have the module
            checker = designspaceProblems.DesignSpaceChecker(path)
            #checker.checkEverything()
            checker.checkDesignSpaceGeometry()
            checker.checkSources()
            checker.checkInstances()
            #pprint(checker.problems)
            if (0,0) in checker.problems:
                # we really can't continue
                self.reportError(msg, "The file can't be parsed.")
                return
            elif (2, 10) in checker.problems:
                # we really can't continue
                self.reportError(msg, "A source has an anisotropic location.")
                return
            elif (1, 9) in checker.problems or (1, 9) in checker.problems:
                # we really can't continue
                self.reportError(msg, "Sorry, the axis values don't make sense.")
                return
            elif (2, 7) in checker.problems:
                # we really can't continue
                self.reportError(msg, "No source defined for default location.")
                return
                
        else:
            print("Don't have the designspaceProblems package")

        # this is a designspace path
        try:
            openDocs = self.getOpenDesignspacePaths()
            if path in openDocs:
                editor = self.getDesignspaceWindowFromPath(path)
                #editor.save(None)
                editor.close()
                #editor.w.getNSWindow().makeKeyAndOrderFront_(editor)
            

            self.ds = skateboardDesignspaceProcessor.SkateboardDesignSpaceProcessor()
            self.ds.useVarlib = True
            # XX Wrap in try / except
            self.ds.read(path)
            
            self.ds.findDefault()
            self.ds.loadFonts()    # reload
            self.checker = designspaceProblems.DesignSpaceChecker(self.ds)
            #self.checker.checkEverything()
            self.checker.checkDesignSpaceGeometry()
            self.checker.checkSources()
            self.checker.checkInstances()
            self.ds.update = self.document_updateCurrentDesignSpace # add skateboard update callback
            self.document_modificationTime = self.document_getModificationTime()
            self.axisNames = []
            self.document_checkFonts()
            self.document_checkAxes()
            self.spaceWindowData = {}
            self.interactionSources = {"vertical":[], "horizontal":[], "ignore":[]}
            self._defconObservers_glyphs = {}
            try:
                self.document_readSuperpolatorPrefs()
            except:
                # some error reading superpolator prefs
                print("Skateboard: Failed to read superpolator preferences")
                print(traceback.format_exc())
            self.document_readPrefs()
            self.window.w.file.saveButton.enable(True)
            #self.w.frontPage.openButton.setTitle("Close")
            self.document_setNeedSave(False)
            self.window.mainWindow_enableInterface(True)
            if self.interestingLocationLibKey in self.ds.lib:
                pointData = []
                try:
                    pointData = [(self.data_validateLocation(a), b) for a, b in self.ds.lib[self.interestingLocationLibKey]]
                finally:
                    self.pointsOfInterest = pointData
            self.w.getNSWindow().setRepresentedURL_(AppKit.NSURL.fileURLWithPath_(self.ds.path))
            self.w.setTitle(str(os.path.basename(self.ds.path)))

            for window in AllGlyphWindows():
                key = id(window.editGlyphView)
                self.currentGlyphViews[key] = window.getGlyph()
                    
            if self.previewLocationLibKey in self.ds.lib:
                prefLoc = self.data_validateLocation(self.ds.lib[self.previewLocationLibKey])
                self.instanceLocation = prefLoc
            else:
                self.instanceLocation = Location(self.ds.newDefaultLocation())
            # set defaults for interaction sources
            tags = self.data_getAxisTags()
            self.axisTagToInteractionSourceDefaults = dict(wdth='horizontal', wght='vertical', opsz="vertical", slnt='horizontal' )
            if not self.hasInteractionSources():
                axesNames = self.instanceLocation.keys()
                for axisName in axesNames:
                    if self.hasInteractionSources(axisName):
                        continue
                    t = tags.get(axisName)
                    preferredDirectionForAxis = self.axisTagToInteractionSourceDefaults.get(t, "ignore")
                    if not axisName in self.interactionSources[preferredDirectionForAxis]:
                        self.interactionSources[preferredDirectionForAxis].append(axisName)
            publishEvent("skateboard.document.opened",
                location = self.instanceLocation, 
                designspace = self.ds, 
                interactionSources = self.interactionSources, 
                axes = self.axisExtremes,
                instanceGlyph = self.instanceGlyphObject,
                previewGlyph = self.contouredInstanceGlyphObject,
                )
            self.document_updateCurrentDesignSpace()
            self.data_instanceLocationChanged()
            self.navigatorTool = navigatorTool.NavigatorTool()
            installTool(self.navigatorTool)
            self.navigatorTool.RFPreviewCallback = self.glyphWindow_drawPreviewCallbackForNavigatorTool
            app = AppKit.NSApp()
            app.getNextSkateboardMasterCallback = self.getNextSkateboardMaster
            #print('--- app.getNextSkateboardMasterCallback added', app.getNextSkateboardMasterCallback)
        except:
            #print("Skateboard: Failed to open the document", path)
            print(traceback.format_exc())

    def _convertSuperpolatorDocument(self, path):
        print("SkateboardDocumentController _convertSuperpolatorDocument", path)

    def document_setNeedSave(self, state=True):
        self.documentNeedSave = state
        if state:
            self.window.w.file.saveButton.enable(True)
            self.window.w.getNSWindow().setDocumentEdited_(True)
        else:
            self.window.w.file.saveButton.enable(False)
            self.window.w.getNSWindow().setDocumentEdited_(False)

    def document_updateCurrentDesignSpace(self):
        if self.ds is None:
            self.window.pane_doc.l.set([])
            self.locationsList_safelySetItems()
            self.seriesLocations = []
            return
        self.ds.loadFonts()    # reload?
        self.document_checkFonts()
        self.document_checkAxes()
        self.locationsList_updateItemsCallback()

    def document_checkFonts(self):
        # find out which fonts are needed by the designspace
        # and then which fonts we have open in RoboFont.
        self.fontPaths = []
        self.openFonts = {}    # fonts, actually open. key:fontpath, value: fontobject
        self.needFonts = {}    # fonts, needed, but closed
        self.missingFonts = {}
        show = []
        if self.ds is None:
            self.pane_doc.l.set([])
            return
        fp = set()
        sourceDescriptors = {}
        for sd in self.ds.sources:
            # here we have access to the layer name
            k = (sd.path, sd.layerName)
            if k not in self.fontPaths:
                self.fontPaths.append(k)
                sourceDescriptors[k] = sd
            if sd.path is None:
                self.missingFonts[sd.path] = True
            elif not os.path.exists(sd.path):
                self.missingFonts[sd.path] = True
        af = AllFonts()
        currentFont = []
        if af:
            currentFont = af[0]
        #hasNeutral = self.ds.getNeutralFont()
        hasNeutral = self.ds.findDefault()
        defaultLayerName = None
        defaultPath = None
        ufoStatus = None
        if hasNeutral is not None:
            defaultLayerName = hasNeutral.layerName
            defaultPath = hasNeutral.path
        for p, layerName in self.fontPaths:
            ufoIsNeutral = False
            fontIsMuted = False
            layerNameString = layerName
            if layerName is None:
                layerNameString = ""
            fontObj = None
            glyphObj = None
            fontIsNeutralImage = self.window._emptyImage
            if os.path.exists(p):
                
                fontIsOpen = self.window._ufoAvailableImage
                ufoStatus = "available"
            else:
                fontIsOpen = self._ufoMissingImage
                ufoStatus = "missing"
            if p == defaultPath and layerName == defaultLayerName:
                fontIsNeutralImage = self.window._ufoNeutralImage
                ufoIsNeutral = True
            currentFontDone = False
            for f in AllFonts():
                if f.path == p and p is not None:
                    if f == currentFont:
                        fontIsOpen = self._ufoCurrentImage
                        ufoStatus = "current"
                    else:
                        fontIsOpen = self._ufoOpenImage
                        ufoStatus = "open"
                    fontObj = f
                    self.openFonts[p] = fontObj
                    break
            if fontObj is None:
                # the font is not open
                # so: open it
                if os.path.exists(p):
                    try:
                        fontObj = RFont(p, showInterface=False)
                        self.needFonts[p] = fontObj
                    except:
                        # faulty thing?
                        print("Skateboard: Trouble reading ufo from", p)
            if fontObj:
                if self.currentGlyph is None:
                    glyphObj = "-"
                else:
                    if self.currentGlyph.name in fontObj:
                        glyphObj = fontObj[self.currentGlyph.name].__class__.__name__ + " " + self.currentGlyph.name
                    else:
                        glyphObj = "-"
            names = []
            key = (p,layerName)
            if key in self.mutedFonts:
                fontIsMutedImage = self.window._ufoMutedImage
            else:
                fontIsMutedImage = self.window._emptyImage
            show.append(dict(
                isNeutralImage=fontIsNeutralImage,
                isOpenImage=fontIsOpen,
                isMutedImage = fontIsMutedImage,
                name=os.path.basename(p ),
                _fontObject=str(fontObj.__class__.__name__),
                _glyphObject=glyphObj,
                _fontid = "%X"%id(fontObj),
                layer=layerNameString,
                #function="",
                factor="",
                _sourceDescriptorKey = key,    # used by getSourceDescriptorFromKey
                _ufoStatus = ufoStatus,
                _ufoIsNeutral = ufoIsNeutral
                ))
        self.window.pane_doc.l.set(show)

    def document_checkAxes(self):
        # get the axes data from the designspace,
        # provide the axes
        newLocation = None
        newAxisNames = []
        newAxisExtremes = {}
        loc = {}
        for aD in self.ds.axes:
            # set default location, safest to go to the default. 
            loc[aD.name] = aD.default
            newAxisNames.append(aD.name)
            newAxisExtremes[aD.name] = (aD.minimum, aD.default, aD.maximum)
        geometryChanged = False
        if self.axisNames != newAxisNames:
            self.axisNames = newAxisNames
            geometryChanged = True
        if self.axisExtremes != newAxisExtremes:
            self.axisExtremes = newAxisExtremes
            geometryChanged = True
        if geometryChanged:
            self.instanceLocation = Location(loc)
        self.axisMapper = ufoProcessor.varModels.AxisMapper(self.ds.axes)
        self.data_collectSnapValues()

    def document_readPrefs(self):
        # read the document skateboard prefs
        # interaction sources
        docInteractionSources = self.ds.lib.get(self.interactionSourcesLibKey)
        orderedAxisNames = [a.name for a in self.ds.axes]
        validInteractionSources = {'horizontal': [], 'vertical': [], 'ignore': []}
        if docInteractionSources:
            for k, axisNames in docInteractionSources.items():
                if k in validInteractionSources:
                    for axisName in axisNames:
                        if axisName in orderedAxisNames:
                            validInteractionSources[k].append(axisName)
        self.interactionSources = validInteractionSources
        # previewtext
        rawText = self.ds.lib.get(self.previewTextLibKey)
        if rawText is not None:
            self.previewTextRaw = rawText
        # muted fonts
        mutedSourcesFromDoc = self.ds.lib.get(self.mutedSourcesLibKey)
        if mutedSourcesFromDoc is not None:
            mutedSources = {}
            for ufoName, layerName in mutedSourcesFromDoc:
                for sD in self.ds.sources:
                    if os.path.basename(sD.path) == ufoName:
                        if layerName == "foreground":
                            layerName = None
                        if layerName == sD.layerName:
                            sourceDescriptorKey = (sD.path, sD.layerName)
                            mutedSources[sourceDescriptorKey] = sD
            self.mutedFonts = mutedSources                            
        # muted axes
        mutedAxesFromDoc = self.ds.lib.get(self.mutedAxesLibKey)
        orderedAxisNames = [a.name for a in self.ds.axes]
        if mutedAxesFromDoc is not None:
            mutedAxes = {}
            for axisName in mutedAxesFromDoc:
                if axisName not in orderedAxisNames:
                    continue
                mutedAxes[axisName] = True
            self.mutedAxes = mutedAxes
        
    def document_readSuperpolatorPrefs(self):
        # read the document, or superpolator prefs
        superpolatorPrefs = self.ds.lib.get(self.superpolatorDataLibKey)
        orderedAxisNames = [a.name for a in self.ds.axes]
        if not superpolatorPrefs:
            if len(orderedAxisNames)>0:
                self.interactionSources["horizontal"] = [orderedAxisNames[0]]
            if len(orderedAxisNames)>1:
                self.interactionSources["vertical"] = [orderedAxisNames[1]]
            return
        if 'verticalPreviewAxis' in superpolatorPrefs:
            wantAxisName = superpolatorPrefs['verticalPreviewAxis']
            if wantAxisName in orderedAxisNames:
                self.interactionSources["vertical"] = [wantAxisName]
        else:
            if len(orderedAxisNames)>0:
                self.interactionSources["vertical"] = [orderedAxisNames[0]]
        if 'horizontalPreviewAxis' in superpolatorPrefs:
            wantAxisName = superpolatorPrefs['horizontalPreviewAxis']
            if wantAxisName in orderedAxisNames:
                self.interactionSources["horizontal"] = [wantAxisName]
        else:
            self.interactionSources["horizontal"] = [orderedAxisNames[1]]
        #self.previewTextRaw = superpolatorPrefs.get('previewtext', None)

    def data_collectSnapValues(self):
        # collect all the values from all the axes
        if self.ds is None:
            self.snapValues = {}
            return
        self.snapValues = {}
        axes = {}
        units = {}
        for ad in self.ds.axes:
            self.snapValues[ad.name] = []
            units[ad.name] =  self.data_getAxisIncrementUnit(ad.name)
            axes[ad.name] = self.data_getAxisValues(ad.name)
        for sd in self.ds.sources:
            if not sd.location: continue
            for axisName, value in sd.location.items():
                t = (value-.5*units[axisName], value+.5*units[axisName])
                self.snapValues[axisName].append((t, value))
        for sd in self.ds.instances:
            if not sd.location: continue
            for axisName, value in sd.location.items():
                if type(value) == type(()):
                    for v in value:
                        t = (v-.5*units[axisName], v+.5*units[axisName])
                        self.snapValues[axisName].append((t, v))
                else:
                    t = (value-.5*units[axisName], value+.5*units[axisName])
                    self.snapValues[axisName].append((t, value))
        # ading interesting to snap values
        for location, name in self.pointsOfInterest:
            for axisName, value in location.items():
                t = (value-.5*units[axisName], value+.5*units[axisName])
                self.snapValues[axisName].append((t, value))
        # add open locations
        # add rules values

    def data_getAxisValues(self, axisName=None):
        # return the minimum / default / maximum for the axis
        # it's possible we ask for an axis that is not in the document.
        if self.ds is None:
            return None
        if axisName is None:
            # get all of them
            axes = {}
            for ad in self.ds.axes:
                axes[ad.name] = (ad.minimum, ad.default, ad.maximum)
            return axes
        for ad in self.ds.axes:
            if ad.name == axisName:
                return ad.minimum, ad.default, ad.maximum
        return None

    def data_getAxisIncrementUnit(self, axisName):
        # calculate our preferred smallest step along this axis
        axisUnit = 0
        axisValues = self.data_getAxisValues(axisName)
        if axisValues:
            mn, df, mx = axisValues
            axisUnit = (mx - mn)/self.axisSteps
        return axisUnit

    def locationsList_safelySetItems(self, items=None):
        self.locationsList_doNotUpdate = True
        if items is None:
            items = []
        self.window.pane_locs.locationsList.set(items)
        self.locationsList_doNotUpdate = False

    def getOpenDesignspacePaths(self):
        # make a dict with all open designspace editor files
        wrappers = getDesignSpaceDocuments()
        docs = []
        for obj in wrappers:
            if obj.designSpacePath is not None:
                docs.append(obj.designSpacePath)
        return docs

    def reportError(self, text, tb):
        if tb.find("Traceback") == 0:
            tb = "Something went wrong. At this time it may not be possible to assign blame and point fingers. So, pardon the traceback posted below, it may be helpful in a report. Thanks!\n\n%s" % tb
        vanilla.dialogs.message("%s" %text, tb, parentWindow=self.window.w)

    def document_getModificationTime(self):
        if self.ds is None:
            return None
        return os.path.getmtime(self.ds.path)

if __name__ == "__main__":
    sw = SkateBoardWindow()
    