# small window that shows which events Skateboard is publishing
import AppKit
import vanilla
import time, os
from mojo.events import addObserver, publishEvent, removeObserver
from defconAppKit.controls.glyphLineView import GlyphLineView
from mojo.UI import MultiLineView
import mojo.extensions

skateboardBundle = mojo.extensions.ExtensionBundle("SkateBoard")

class StreamingGlyphWindow(object):
    toolNameForTitle = "⌘Skateboard"
    glyphLibLocationKey = 'com.skateboard.preview.location'
    def __init__(self):
        self.w = vanilla.Window((1200, 200), "%s Streamer" % self.toolNameForTitle, minSize=(100, 200))
        self.isStreaming = False

        self.font = RFont(showInterface=False)
        self.font.info.unitsPerEm = 1500
        self.font.info.descender = -500
        self.font.info.ascender = 1000
        self.allGlyphs = []
        self.lastGlyphAndLocation = None

        self.displayPrefs = {'Show Kerning': True,
            'Center': False,
            'Show Space Matrix': False,
            'Single Line': False,
            'showLayers': [],
            'displayMode': u'Multi Line',
            'Stroke': False,
            'xHeight Cut': False,
            'Upside Down': False,
            'Beam': False,
            'Water Fall': False,
            'Inverse': False,
            'Show Template Glyphs': True,
            'Multi Line': True,
            'Right to Left': False,
            'Show Metrics': False,
            'Show Control glyphs': True,
            'Fill': True,
            'Left to Right': True}

        #self.w.line = GlyphLineView((0,0,0,0), pointSize=None, applyKerning=False)
        self.w.mline = MultiLineView((0,0,0,0), pointSize=80, lineHeight=160,
            selectionCallback=self.callbackGlyphLineClick, doubleClickCallback=self.callbackGlyphLineDoubleClick)
        self.w.mline.setFont(self.font)
        self.w.mline.setDisplayStates(self.displayPrefs)
        #self.w.line.setBackgroundColor(AppKit.NSColor.whiteColor())
        #self.w.line.setGlyphColor(AppKit.NSColor.colorWithCalibratedRed_green_blue_alpha_(0, 0, 0, 1))
        self.w.bind("close", self.cbClose)
        self.events = []
        addObserver(self, "previewLocationChangedCallback", 'skateboard.previewLocation.changed')
        addObserver(self, "docChangedCallback", "skateboard.interactionSources.changed")
        addObserver(self, "docChangedCallback", "skateboard.window.opened")
        addObserver(self, "docChangedCallback", "skateboard.window.closed")
        addObserver(self, "docChangedCallback", "skateboard.document.opened")
        addObserver(self, "docChangedCallback", "skateboard.document.closed")
        addObserver(self, "confirmRunningCallback", "skateboard.running.ping")
        publishEvent("skateboard.running")
        self.w.open()

    def callbackGlyphLineClick(self, sender):
        print("callbackGlyphLineClick", sender.getSelectedGlyph())
        g = sender.getSelectedGlyph()
        if g is not None:
            if self.glyphLibLocationKey in g.lib:
                print("sender.getSelectedGlyph()", g.lib[self.glyphLibLocationKey])
    
    def callbackGlyphLineDoubleClick(self, sender):
        print("callbackGlyphLineDoubleClick", sender.getSelectedGlyph())
    
    def docChangedCallback(self, notification):
        pass
        
    def addGlyph(self, g, loc):
        if loc is None:
            return
        name = "%s_%f" %(g.name, time.time())
        key = (g.name, loc.items())
        #print("a", key == self.lastGlyphAndLocation, key, self.lastGlyphAndLocation)
        #if key == self.lastGlyphAndLocation:
        #    print("key skip", key)
        #    return
        #print("key add", key, key==self.lastGlyphAndLocation)
        
        self.lastGlyphAndLocation = key
        self.font.newGlyph(name)
        new = self.font[name]
        new.width = g.width
        p = new.getPointPen()
        g.drawPoints(p)
        #print("g", g, len(g.contours), len(g.components))
        #print("new", new, len(new.contours), len(new.components))
        #print('new.naked()', new.naked())
        #self.allGlyphs = [g.naked()] + self.allGlyphs
        if loc is not None:
            new.lib[self.glyphLibLocationKey] = loc
        self.allGlyphs = [new] + self.allGlyphs
        self.allGlyphs = self.allGlyphs[:50]
        self.w.mline.set(self.allGlyphs)
        
    def cbClose(self, sender=None):
        removeObserver(self, "skateboard.previewLocation.changed")
        removeObserver(self, "skateboard.interactionSources.changed")
        removeObserver(self, "skateboard.window.opened")
        removeObserver(self, "skateboard.window.closed")
        removeObserver(self, "skateboard.document.opened")
        removeObserver(self, "skateboard.document.closed")
        removeObserver(self, "skateboard.running.ping")
        self.font = None
    
    def confirmRunningCallback(self, notification):
        self.isStreaming = True
        
    def previewLocationChangedCallback(self, notification):
        g = notification.get('previewGlyph')
        loc = notification.get('location')
        if g is not None:
            self.addGlyph(g, loc)

if __name__ == "__main__":
    OpenWindow(StreamingGlyphWindow)
