import vanilla
from mojo.UI import setDefault, getDefault
from getDesignSpaceWindow import getSkateboard

# preferences window
# save all to 

openInstancesPrefKey = "com.letterror.skateboard.openInstances"
followOpticalSizePrefKey = "com.letterror.skateboard.followOpticalSize"
colorSchemePrefKey = "com.letterror.skateboard.colorScheme"
problemsResponsePrefKey = "com.letterror.skateboard.problemsResponse"

def getColorScheme():
    return getDefault(colorSchemePrefKey, None)

def getProblemsResponse():
    return getDefault(problemsResponsePrefKey, None)
    
    
class SkateboardPreferences(object):
    def __init__(self):
        self.w = vanilla.FloatingWindow((1, 1), "Skateboard Preferences")
        self.w.instancesPrefCheck = vanilla.CheckBox("auto", "Open instance UFO’s after generating.", callback=self.prefCallback)
        self.w.instancesPrefCheck.prefKey = openInstancesPrefKey
        self.setPrefKeyValues(self.w.instancesPrefCheck)
        
        self.w.colorSchemeCaption = vanilla.TextBox("auto", "Skateboard Preview Use")
        self.w.colorScheme = vanilla.RadioGroup("auto", 
                    ['Bright Colors', 'Dark Colors'], isVertical=False, callback=self.colorSchemeCallback)
        scheme = getDefault(colorSchemePrefKey)
        if scheme == 'bright':
            self.w.colorScheme.set(0)
        elif scheme == 'dark':
            self.w.colorScheme.set(1)
        else:
            self.w.colorScheme.set(0)

        self.w.problemsResponseCaption = vanilla.TextBox("auto", "Glyph compatibility problems")
        self.w.problemsResponse = vanilla.RadioGroup("auto", 
                    #['Show', 'Prepolate', "Notify"], isVertical=False, callback=self.problemsResponseCallback)
                    ['Show', 'Hide'], isVertical=False, callback=self.problemsResponseCallback)
        response = getDefault(problemsResponsePrefKey)
        if response == None or response == "show":
            self.w.problemsResponse.set(0)
        elif response == "prepolate":
            self.w.problemsResponse.set(1)
        elif response == "notify":
            self.w.problemsResponse.set(2)

        self.w.okButton = vanilla.Button("auto", "Close", callback=self.closeButtonCallback)
        rules = [
            "H:|-[instancesPrefCheck]-|",
            "H:|-[colorSchemeCaption]-|",
            "H:|-[colorScheme]-|",
            "H:|-[problemsResponseCaption]-|",
            "H:|-[problemsResponse]-|",
            "H:|-[okButton]-|",
            "V:|-[instancesPrefCheck]-space-[colorSchemeCaption]-[colorScheme]-space-[problemsResponseCaption]-[problemsResponse]-[okButton]-|",
        ]
        metrics = {
            "half": 150,
            'space': 30,
            }
        self.w.addAutoPosSizeRules(rules, metrics)
        self.w.open()
    
    def problemsResponseCallback(self, sender):
        # set the choice for response to compatibility issues
        v = sender.get()
        if v == 0:
            response = "show"
        elif v == 1:
            response = "prepolate"
        elif v == 2:
            response = "notify"
        setDefault(problemsResponsePrefKey, response)
        sb = getSkateboard()
        if sb is not None:
            #print("switching to", scheme)
            sb.resetProblemsResponse()
            sb.data_instanceLocationChanged(dragging=False)
        
    def colorSchemeCallback(self, sender):
        # set the choice for the color scheme
        scheme = "bright"
        v = sender.get()
        if v == 0:
            scheme = "bright"
        elif v == 1:
            scheme = "dark"
        setDefault(colorSchemePrefKey, scheme)
        sb = getSkateboard()
        if sb is not None:
            #print("switching to", scheme)
            sb.resetColorScheme()
            sb.data_instanceLocationChanged(dragging=False)
            
    def setPrefKeyValues(self, item):
        value = getDefault(item.prefKey)
        #print("pref \"{}\" {}".format(item.getTitle(), value))
        if not value:
            value = 0
        else:
            value = 1
        item.set(value)
        
    def closeButtonCallback(self, sender):
        self.w.close()
        
    def prefCallback(self, sender):
        #print(sender.get(), sender.getTitle(), sender.prefKey)
        value = sender.get()
        setDefault(sender.prefKey, value)

if __name__ == "__main__":
    OpenWindow(SkateboardPreferences)