import AppKit
import vanilla
from defconAppKit.windows.baseWindow import BaseWindowController
from glyphNameFormatter.reader import n2u, u2n

class AddGlyphsSheet(BaseWindowController):
    _title = "Skateboard Add Glyphs Sheet"
    _cancelButtonTitle = "Cancel"
    _OKButtonTitle = "Add Glyphs"
    _sheetWidth = 300
    _sheetHeight = 400

    def __init__(self, data, cancelCallback, applyCallback, parentWindow=None):
        self.data = data
        self.names = []
        for name, uniValue in data['proposedNames']:
            # ignore the provided unicode value
            # take the one from RF instead
            uni = n2u(name)
            if uni is None:
                uni = ""
            else:
                uni = hex(uni)
            self.names.append(dict(name=name, unicodeValue=uni))
        self.cancelCallback = cancelCallback
        self.applyCallback = applyCallback
        self.parentWindow = parentWindow
        self.buildBaseWindow(parentWindow)

    def buildBaseWindow(self, parentWindow):
        if parentWindow is None:
            thisHeight = self._sheetHeight-30
            self.w = vanilla.Window((self._sheetWidth, self._sheetHeight),
                title = self._title,
                miniaturizable=False,
                minSize=(self._sheetWidth, self._sheetHeight),
                )
        else:
            self.w = vanilla.Sheet((self._sheetWidth, self._sheetHeight),
                parentWindow,
                minSize=(self._sheetWidth, self._sheetHeight),
            )
        self.w.caption = vanilla.TextBox((10, 5, -10, 40), "Add these glyphs to this designspace?\n(Closed UFOs will be opened first.)")
        self.w.cancelButton = vanilla.Button((-210, -30, -110, 20), 
            self._cancelButtonTitle,
            callback=self.callbackCancelButton,
            sizeStyle='small')
        self.w.applyButton = vanilla.Button((-100, -30, -10, 20), 
            self._OKButtonTitle,
            callback=self.callbackApplyButton,
            sizeStyle='small')
        self.w.setDefaultButton(self.w.applyButton)
        
        y = -200
        # list based on currently selected glyph in the open master.
        # maybe this should just be an input box for names, and then
        # a list overview of what is about to be added
        self.w.namesInputBox = vanilla.EditText((10, 50, -10, y), "", placeholder="Write / paste glyphnames", callback=self.analyseGlyphNamesCallback)
        columnDescriptions = [
            {'title': 'Name', 'key':"name", 'editable': False},
            {'title': 'Unicode', 'key':"unicodeValue", 'editable': False},
            ]
        margin = 50
        self.w.glyphNames = vanilla.List((10, y+10, -10, -40), self.names, columnDescriptions=columnDescriptions)
        self.w.open()
        names = []
        for d in self.names:
            names.append(d['name'])
        names.sort()
        self.w.namesInputBox.set(" ".join(names))
        self.analyseGlyphNamesCallback()
    
    def analyseGlyphNamesCallback(self, sender=None):
        names = self.w.namesInputBox.get()
        proposal = []
        data = []
        for n in names.split(' '):
            n = n.strip()
            if not n: continue
            uni = n2u(n)
            if uni is None:
                uniT = ""
            else:
                uniT = hex(uni)
            proposal.append(dict(name=n, unicodeValue=uniT))
            data.append((n,uni))
        self.names = proposal
        self.data['proposedNames'] = data
        self.w.glyphNames.set(proposal)
        
    def callbackApplyButton(self, sender):
        """ This applies the edited records to the controller. """
        if self.applyCallback is not None:
            self.applyCallback(self.data)
        self.w.close()

    def callbackCancelButton(self, sender):
        """ This applies the edited records to the controller. """
        if self.cancelCallback is not None:
            self.cancelCallback(self.data)
        self.w.close()
        
        
class NewLayerNameEditSheet(BaseWindowController):
    _title = "Skateboard new Layer Name Edit Sheet"
    _cancelButtonTitle = "Cancel"
    _OKButtonTitle = "Apply"
    _sheetWidth = 300
    _sheetHeight = 70

    def __init__(self, data, cancelCallback, applyCallback, parentWindow=None):
        self.data = data
        self.cancelCallback = cancelCallback
        self.applyCallback = applyCallback
        self.parentWindow = parentWindow
        self.buildBaseWindow(parentWindow)

    def buildBaseWindow(self, parentWindow):
        if parentWindow is None:
            thisHeight = self._sheetHeight-30
            self.w = vanilla.Window((self._sheetWidth, self._sheetHeight),
                title = self._title,
                miniaturizable=False,
                minSize=(self._sheetWidth, self._sheetHeight),
                )
        else:
            self.w = vanilla.Sheet((self._sheetWidth, self._sheetHeight),
                parentWindow,
                minSize=(self._sheetWidth, self._sheetHeight),
            )

        self.w.cancelButton = vanilla.Button((-210, -30, -110, 20), 
            self._cancelButtonTitle,
            callback=self.callbackCancelButton,
            sizeStyle='small')
        self.w.applyButton = vanilla.Button((-100, -30, -10, 20), 
            self._OKButtonTitle,
            callback=self.callbackApplyButton,
            sizeStyle='small')
        self.w.setDefaultButton(self.w.applyButton)
        
        margin = 50
        proposedLayerName = self.data.get("proposedLayerName", "New-support-layer")
        self.w.layerName = vanilla.EditText((10, 10, -10, 22), proposedLayerName, placeholder="New Layername")
        self.w.open()

    def callbackApplyButton(self, sender):
        """ This applies the edited records to the controller. """
        proposedLayerName = self.w.layerName.get()
        if proposedLayerName in self.data['existingLayerNames']:
            AppKit.NSBeep()
            return
        self.data['newLayerName'] = proposedLayerName
        if self.applyCallback is not None:
            self.applyCallback(self.data)
        self.w.close()

    def callbackCancelButton(self, sender):
        """ This applies the edited records to the controller. """
        if self.cancelCallback is not None:
            self.cancelCallback(self.data)
        self.w.close()
        
            
class LocationNameEditSheet(BaseWindowController):
    _title = "Skateboard Interesting Location Label Sheet"
    _cancelButtonTitle = "Cancel"
    _OKButtonTitle = "Apply"
    _sheetWidth = 300
    _sheetHeight = 70

    def __init__(self, data, cancelCallback, applyCallback, parentWindow=None):
        self.data = data
        self.cancelCallback = cancelCallback
        self.applyCallback = applyCallback
        self.parentWindow = parentWindow
        self.buildBaseWindow(parentWindow)

    def buildBaseWindow(self, parentWindow):
        if parentWindow is None:
            thisHeight = self._sheetHeight-30
            self.w = vanilla.Window((self._sheetWidth, self._sheetHeight),
                title = self._title,
                miniaturizable=False,
                minSize=(self._sheetWidth, self._sheetHeight),
                )
        else:
            self.w = vanilla.Sheet((self._sheetWidth, self._sheetHeight),
                parentWindow,
                minSize=(self._sheetWidth, self._sheetHeight),
            )

        self.w.cancelButton = vanilla.Button((-210, -30, -110, 20), 
            self._cancelButtonTitle,
            callback=self.callbackCancelButton,
            sizeStyle='small')
        self.w.applyButton = vanilla.Button((-100, -30, -10, 20), 
            self._OKButtonTitle,
            callback=self.callbackApplyButton,
            sizeStyle='small')
        self.w.setDefaultButton(self.w.applyButton)
        
        margin = 50
        self.w.locationName = vanilla.EditText((10, 10, -10, 22), self.data['currentLocationName'])
        self.w.open()

    def callbackApplyButton(self, sender):
        """ This applies the edited records to the controller. """
        self.data['newLocationName'] = self.w.locationName.get()
        if self.applyCallback is not None:
            self.applyCallback(self.data)
        self.w.close()

    def callbackCancelButton(self, sender):
        """ This applies the edited records to the controller. """
        if self.cancelCallback is not None:
            self.cancelCallback(self.data)
        self.w.close()
    
class LocationEditorSheet(BaseWindowController):

    _title = "Skateboard Location Editor Sheet"
    _cancelButtonTitle = "Cancel"
    _OKButtonTitle = "Apply"
    _sheetWidth = 250
    _sheetHeight = 150

    def __init__(self, data, cancelCallback, applyCallback, parentWindow=None, axisValues=None):
        self.data = data
        self.data['newLocation'] = data['location'].copy()
        self.cancelCallback = cancelCallback
        self.applyCallback = applyCallback
        self.parentWindow = parentWindow
        self.axisValues = axisValues
        self.buildBaseWindow(parentWindow)

    def buildBaseWindow(self, parentWindow):

        lookingAtDescriptions = [
            {'title': "Axis", 'key':"axisKey", 'width': 90, 'editable':False, 'typingSensitive': True},
            {'title': "Value", 'key':"valueKey", 'width': 100, 'editable':True,},
            ]

        if parentWindow is None:
            thisHeight = self._sheetHeight-30
            self.w = vanilla.Window((self._sheetWidth, self._sheetHeight),
                title = self._title,
                miniaturizable=False,
                minSize=(self._sheetWidth, self._sheetHeight),
                )
        else:
            self.w = vanilla.Sheet((self._sheetWidth, self._sheetHeight),
                parentWindow,
                minSize=(self._sheetWidth, self._sheetHeight),
            )
        self.w.cancelButton = vanilla.Button((-210, -30, -110, 20), 
            self._cancelButtonTitle,
            callback=self.callbackCancelButton,
            sizeStyle='small')
        self.w.applyButton = vanilla.Button((-100, -30, -10, 20), 
            self._OKButtonTitle,
            callback=self.callbackApplyButton,
            sizeStyle='small')
        self.w.setDefaultButton(self.w.applyButton)
        margin = 50
        self.w.axes = vanilla.List(
            (0,0,0,-40), [], 
            editCallback=self.editAxisValueCallback,
            columnDescriptions=lookingAtDescriptions,
            menuCallback = self.makeLocationPrefsMenu
            )
        self.w.axes.set(self.dataToDicts(self.data['newLocation']))
        self.w.open()
    
    def makeLocationPrefsMenu(self, sender):
        items = []
        sel = sender.getSelection()
        if not sel:
            AppKit.NSBeep()
            return
        if len(sel) == 1:
            listItem = self.w.axes[sel[0]]
            items.append(dict(title="Axis %s" % listItem['axisKey']))
            items.append('----')
            if self.axisValues:
                if listItem['axisKey'] in self.axisValues:
                    aMin, aDef, aMax = self.axisValues[listItem['axisKey']]
                    items.append(dict(title="Minimum %3.3f" % aMin))
                    if aDef != aMin and aDef != aMax:
                        items.append(dict(title="Default %3.3f" % aDef, callback=self.setAxisValue))
                    items.append(dict(title="Maximum %3.3f" % aMax, callback=self.setAxisValue))
        return items    
    
    def setAxisValue(self, sender):
        # set the value from the menu to this axis.
        title = sender.title()
        sel = self.w.axes.getSelection()
        if not sel:
            AppKit.NSBeep()
            return
        if len(sel) == 1:
            listItem = self.w.axes[sel[0]]
            if "Maximum" in title:
                self.setAxisValueToData(listItem['axisKey'], self.axisValues[listItem['axisKey']][2])
            elif "Minimum" in title:
                self.setAxisValueToData(listItem['axisKey'], self.axisValues[listItem['axisKey']][0])
            elif "Default" in title:
                self.setAxisValueToData(listItem['axisKey'], self.axisValues[listItem['axisKey']][1])
            
    def editAxisValueCallback(self, sender):
        editedLocation = {}
        validLocation = True
        didEditLocation = False
        newValues = {}
        for thing in sender.get():
            v = None
            try:
                v = float(thing['valueKey'])
                didEditLocation = True
            except ValueError:
                AppKit.NSBeep()
                validLocation = False
                break
            editedLocation[thing['axisKey']] = v
        if validLocation:
            self.w.applyButton.enable(True)
        else:
            self.w.applyButton.enable(False)
        new = {}
        self.data['newLocation'] = editedLocation
    
    def setAxisValueToData(self, axisName, value):
        self.data['newLocation'][axisName] = value
        self.w.axes.set(self.dataToDicts(self.data['newLocation']))
        
    def dataToDicts(self, location):
        d = []
        for axisName, axisValue in location.items():
            d.append(dict(axisKey=axisName, valueKey=axisValue))
        return d
        
    def callbackApplyButton(self, sender):
        """ This applies the edited records to the controller. """
        #selectedSource = self.w.availableSources.get()
        #self.data['selectedSource'] = self.data['availableSources'][self.w.availableSources.get()]
        #self.data['selectedName'] = self.w.layerName.get()
        if self.applyCallback is not None:
            self.applyCallback(self.data)
        self.w.close()

    def callbackCancelButton(self, sender):
        """ This applies the edited records to the controller. """
        if self.cancelCallback is not None:
            self.cancelCallback(self.data)
        self.w.close()
    
class SupportSheet(BaseWindowController):

    _title = "Skateboard New Support Sheet"
    _cancelButtonTitle = "Cancel"
    _OKButtonTitle = "Add Support"
    _sheetWidth = 450
    _sheetHeight = 220

    def __init__(self, data, cancelCallback, applyCallback, parentWindow=None):
        self.data = data
        self.cancelCallback = cancelCallback
        self.applyCallback = applyCallback
        self.parentWindow = parentWindow
        self.buildBaseWindow(parentWindow)

    def buildBaseWindow(self, parentWindow):
        if parentWindow is None:
            thisHeight = self._sheetHeight-30
            self.w = vanilla.Window((self._sheetWidth, self._sheetHeight),
                title = self._title,
                miniaturizable=False,
                minSize=(self._sheetWidth, self._sheetHeight),
                )
        else:
            self.w = vanilla.Sheet((self._sheetWidth, self._sheetHeight),
                parentWindow,
                minSize=(self._sheetWidth, self._sheetHeight),
            )

        self.w.cancelButton = vanilla.Button((-210, -30, -110, 20), 
            self._cancelButtonTitle,
            callback=self.callbackCancelButton,
            sizeStyle='small')
        self.w.applyButton = vanilla.Button((-100, -30, -10, 20), 
            self._OKButtonTitle,
            callback=self.callbackApplyButton,
            sizeStyle='small')
        self.w.setDefaultButton(self.w.applyButton)
        
        margin = 50
        self.w.glyphName = vanilla.TextBox((10, 10, -10, 20), "New support for glyph \"%s\"" % self.data['glyphName'])
        self.w.layerNameCaption = vanilla.TextBox((10, 40, -10, 20), "Name for the new layer:", sizeStyle="small")
        self.w.layerName = vanilla.EditText((margin, 65, -10, 20), self.data['proposedLayerName'], sizeStyle="small")
        self.w.availableSourcesCaption = vanilla.TextBox((10, 100, -10, 20), "Select open UFO for the new layer:", sizeStyle="small")
        self.w.availableSources = vanilla.PopUpButton((margin, 125, -10, 20), self.data['ufoNames'])

        self.w.dialogCaption = vanilla.TextBox((10, -50, -10, 20), "Note: you need to save the designspace as well as the UFO.", sizeStyle="small")
        self.w.open()

    def callbackApplyButton(self, sender):
        """ This applies the edited records to the controller. """
        selectedSource = self.w.availableSources.get()
        self.data['selectedSource'] = self.data['availableSources'][self.w.availableSources.get()]
        self.data['selectedName'] = self.w.layerName.get()
        if self.applyCallback is not None:
            self.applyCallback(self.data)
        self.w.close()

    def callbackCancelButton(self, sender):
        """ This applies the edited records to the controller. """
        if self.cancelCallback is not None:
            self.cancelCallback(self.data)
        self.w.close()

if __name__ == "__main__":
    def cancelCallback(sender):
        pass
    def applyCallback(data):
        print(data)
    
    #data = dict(availableSources = ["a", "b", "c"], glyphName="A", proposedLayerName="NewLayerNameShouldBeEditable", ufoNames=['a', 'b', 'c'])
    #OpenWindow(SupportSheet, data, cancelCallback=cancelCallback, applyCallback=applyCallback)
    
    #data = dict(location=dict(axis1=100, axis2=200, axis3=300, axis4=400))
    #OpenWindow(LocationEditorSheet, data, cancelCallback=cancelCallback, applyCallback=applyCallback, axisValues=dict(axis1=(0, 100, 300), axis2=(0, 100, 300), axis3=(0, 100, 300)))
    
    #data = dict(currentLocationName="Current.Location.Name")
    #OpenWindow(LocationNameEditSheet, data, cancelCallback=cancelCallback, applyCallback=applyCallback)
    
    #data = dict(proposedLayerName="aaa", existingLayerNames=['bb', 'cc'])
    #OpenWindow(NewLayerNameEditSheet, data, cancelCallback=cancelCallback, applyCallback=applyCallback)
    
    
    data = dict(proposedNames = [('arrowup',None), ('heh',None), ('dollar',None), ('Euro',None)])
    OpenWindow(AddGlyphsSheet, data, cancelCallback=cancelCallback, applyCallback=applyCallback)
    