import AppKit
from defconAppKit.windows.baseWindow import BaseWindowController
from defconAppKit.controls.glyphLineView import GlyphLineView
from mutatorMath.objects.mutator import Location, buildMutator
import vanilla
import random
from mojo.events import addObserver, removeObserver, installTool, uninstallTool, publishEvent
from mojo.UI import PutFile, setDefault, getDefault
from mojo.canvas import CanvasGroup
from mojo.drawingTools import *
from mojo.events import publishEvent

import importlib
import sbx.tools.colorMachine
importlib.reload(sbx.tools.colorMachine)

class Container(object):
    pass

import colorsys

def getColor(steps, index, saturation=1):
    return colorsys.hsv_to_rgb(index/(steps*1.0), saturation, 1)

def roundedRect(x, y, w, h, offset=25):
    if w < 0:
        w = abs(w)
        x -= w
    if h < 0:
        h = abs(h)
        y -= h
    newPath()
    moveTo((x+offset, y))
    lineTo((x+w-offset, y))
    curveTo((x+w, y), (x+w, y), (x+w, y+offset))
    lineTo((x+w, y+h-offset))
    curveTo((x+w, y+h), (x+w, y+h), (x+w-offset, y+h))
    lineTo((x+offset, y+h))
    curveTo((x, y+h), (x, y+h), (x, y+h-offset))
    lineTo((x, y+offset))
    curveTo((x, y), (x, y), (x+offset, y))
    closePath()
    drawPath()

def drawPlaceMarker(x, y):
    s = 7
    save()
    newPath()
    translate(x, y)
    green = (0,.8,0.8)
    stroke(*green)
    strokeWidth(.6)
    fill(None)
    line((-s, 0), (s, 0))
    line((0,-s), (0, s))
    drawPath()
    restore()
    
class SkateboardSpaceWindow(BaseWindowController):
    # a slave window that draws color fields and heatmaps
    toolNameForTitle = "⌘Skateboard Space"
    showLabelsPrefKey = "com.letterror.skateboard.showSpaceWindowLabels"
    marginValuePrefKey = "com.letterror.skateboard.marginValue"
    
    def __init__(self, closeCallback=None, requestSamplesCallback=None):
        # http://robofont.com/documentation/building-tools/api/mojo/mojo-canvas/
        self.closeCallback = closeCallback
        #self.requestSamplesCallback = requestSamplesCallback
        self.darkMode = "bright"
        self.canExtrapolate = False
        self.drawLabels = getDefault(self.showLabelsPrefKey, False)
        self.margin = getDefault(self.marginValuePrefKey, 100)
        self.snapped = []
        self.unitSize = 30    #40    #12
        self.snapDistanceLine = 10
        self.snapDistanceIcon = 20
        self.fadeThreshold = 500
        self.fadeLength = 100
        self.inset = 20
        self.isMain = False
        self.masters = None        # set directly from skateboard
        self.instances = None      # set directly from skateboard
        self.snapValues = None     # set directly from skateboard
        self.axisRanges = {}       # set directly from skateboard
        self.instanceLoc = None    # window coordinates for instancelocation. calculated by skateboard
        self.supports = None       # set directly from skateboard
        self.rules = None          # set directly from skateboard
        self.interesting = None    # set directly from skateboard
        self.openLocations = None  # set directly from skateboard  
        self.masterIconSize = 12
        self._supportColors = []
        self._doSamples = False
        self._samples = {}
        self._done = []
        self._toDo = {}
        self.batchSize = 72
        self.snapColor = (0,.7,1)
        self.interactionSources = None
        self.previewLocation = None
        self.axes = None
        self.processSamplesInterval = .01
        self.processSamplesTimer = None
        self.blue = (2/255, 193/255, 255/255)
        self.cursorLineDark = (1,1,1)
        self.cursorLineLight = (.5,.5,.5)
        self.zones = {}
        self.horizontalSnapped = False
        self.verticalSnapped = False
        #self._glyphViewPreviewBackgroundColor = getDefault("glyphViewPreviewBackgroundColor")

        self.w = vanilla.Window((600, 500),
            minSize=(500, 300),
            title=self.toolNameForTitle,
            autosaveName = "com.letterror.skateboard.spaceWindow",
            fullScreenMode = "auxiliary",
            )
        self.w.pdfButton = vanilla.Button((2,-23,50,20), "PDF", self.savePDFCallback, sizeStyle="small")
        self.w.showLabels = vanilla.CheckBox((-100, -23, 80, 20), "Show labels", sizeStyle="small", callback=self.showLabelsCallback)
        self.w.showLabels.set(self.drawLabels)
        self.w.marginSlider = vanilla.Slider((-200, -23, 80, 20), minValue=50, maxValue=200, value=self.margin, sizeStyle="small", callback= self.marginSliderCallback)
        self.w.c = CanvasGroup((0,0,0,-25), delegate=self)
        self.w.bind("close", self.cbClose)
        self.w.bind("became main", self.cbBecameMain)
        self.w.bind("resigned main", self.cbResignedMain)
        self.w.bind("resize", self.cbResize)
        addObserver(self, "previewLocationChangedCallback", 'skateboard.previewLocation.changed')
        addObserver(self, "interactionSourcesChangedCallback", 'skateboard.interactionSources.changed')
        self.w.open()
        self.w.c.update()

    def extrapolationPrefChanged(self, state):
        self.canExtrapolate = state
        self.w.c.update()
        
    def interactionSourcesChangedCallback(self, notification):
        self.interactionSources = notification['interactionSources']
        self.previewLocation = notification['location']
        self.axes = notification['axes']
        self.masters = notification.get('masters')
        #self.resample()
        self.w.c.update()

    def update(self):
        self.w.c.update()
    
    def marginSliderCallback(self, sender):
        self.margin = sender.get()
        setDefault(self.marginValuePrefKey, self.margin)
        self.w.c.update()
        
    def showLabelsCallback(self, sender):
        # vanilla callback for showLabels checkbpx
        self.drawLabels = sender.get() == 1
        setDefault(self.showLabelsPrefKey, self.drawLabels)
        self.w.c.update()
        
    def previewLocationChangedCallback(self, notification):
        # observer callback for previewLocationChangedCallback
        self.interactionSources = notification['interactionSources']
        self.previewLocation = notification['location']
        self.axes = notification['axes']
        self.w.c.update()
        
    def cbClose(self, sender):
        # vanilla callback space window
        if self.closeCallback:
            self.closeCallback()
        removeObserver(self, 'skateboardPreviewLocationChanged')
        removeObserver(self, 'skateboard.interactionSources.changed')
        self.interactionSources = None
        self.previewLocation = None
        self.axes = None
        self.masters = None
        
    def mainWindowIsClosing(self, sender=None):
        # callback when mainwindow is closing and spacewindow is still open.
        self.instanceFont = None
        self.w.close()
        
    def cbResize(self, sender):
        # vanilla callback space window
        #self.resample()
        pass
    
    def cbBecameMain(self, sender):
        # vanilla callback space window
        self.isMain = True
        #self.scheduleProcessSamplesTimer()
        self.w.c.update()

    def cbResignedMain(self, sender):
        # vanilla callback space window
        self.isMain = False
        if self.processSamplesTimer is not None:
            self.processSamplesTimer.invalidate()
            self.processSamplesTimer = None
    
    def mouseDown(self, event):
        if self.interactionSources is None:
            return
        p = event.locationInWindow()
        p.y -= 25
        loc = self._windowToLocation((p.x, p.y))
        publishEvent("skateboardSpaceWindowEvent", data=dict(location=loc, mouse="mouseDown"), snapped=self.snapped)
    
    def mouseUp(self, event):
        if self.interactionSources is None:
            return
        p = event.locationInWindow()
        p.y -= 25
        loc = self._windowToLocation((p.x, p.y))
        publishEvent("skateboardSpaceWindowEvent", data=dict(location=loc, mouse="mouseUp"), snapped=self.snapped)
        
    def mouseDragged(self, event):
        # callback for canvasgroup delegate
        if self.interactionSources is None:
            return
        p = event.locationInWindow()
        p.y -= 25
        loc = self._windowToLocation((p.x, p.y))
        publishEvent("skateboardSpaceWindowEvent", data=dict(location=loc, mouse="mouseDragged"), snapped=self.snapped)

    def _windowToLocation(self, windowPos):
        if self.interactionSources is None:
            return
        
        #print("spaceWindow _windowToLocation self.canExtrapolate", self.canExtrapolate)
        horizontalAxis = self.interactionSources.get('horizontal')
        verticalAxis = self.interactionSources.get('vertical')
        
        w = self.w.c.width()
        h = self.w.c.height()
        we = w - 2 * self.margin # effective width
        he = h - 2 * self.margin
        px = windowPos[0] - self.margin
        py = windowPos[1] - self.margin
        fx = px / we
        fy = py / he

        # check for zone snapping
        self.snapped = []
        self.horizontalSnapped = False
        self.verticalSnapped = False
        if not horizontalAxis:
            fx = 0
        if not verticalAxis:
            fy = 0
        loc = {}
        if horizontalAxis:
            for axisName in horizontalAxis:
                minimum, default, maximum = self.axes[axisName]
                loc[axisName] = minimum + fx*(maximum-minimum)    # XXX this will not work for axes -1, 0, 1
                if maximum-minimum == 2:
                    loc[axisName] += 1
                # need tp make sure the axis values are within min/max range
                # otherwise the rules kick out
                # this was an issue for Stephen Nixon, 20190807
                # 20191217 this clipping prevents extrapolation to be done in the space window.
                if not self.canExtrapolate:
                    loc[axisName] = max(min(maximum, loc[axisName]), minimum)
        if verticalAxis:
            for axisName in verticalAxis:
                minimum, default, maximum = self.axes[axisName]
                loc[axisName] = minimum + fy*(maximum-minimum)    # XXX this will not work for axes -1, 0, 1
                if maximum-minimum == 2:
                    loc[axisName] += 1
                # also
                # 20191217 this clipping prevents extrapolation to be done in the space window.
                if not self.canExtrapolate:
                    loc[axisName] = max(min(maximum, loc[axisName]), minimum)
        for axisName, value in loc.items():
            if not self.snapValues: break
            if axisName in self.snapValues:
                for z, v in self.snapValues[axisName]:
                    if z[0] <= value <= z[1]:
                        loc[axisName] = v
                        if axisName not in self.snapped:
                            self.snapped.append(axisName)
                            if horizontalAxis:
                                if axisName in horizontalAxis:
                                    self.horizontalSnapped = True
                            if verticalAxis:
                                if axisName in verticalAxis:
                                    self.verticalSnapped = True
        return loc

    def setSupports(self, supports=None):
        self.supports = supports
    
    def setSnapValues(self, snapValues=None):
        self.snapValues = snapValues
    
    def setRules(self, rules=None):
        self.rules = rules
        
    def setMasters(self, masters, instanceLoc, axisRanges=None):
        # accept the masters data from skateboard
        self.masters = masters
        self.instanceLoc = instanceLoc
        if axisRanges is None:
            axisRanges = {}
        self.axisRanges = axisRanges
        self.w.c.update()
        
    def presentEmptySpace(self):
        w = self.w.c.width()
        h = self.w.c.height()
        fill(.5)
        rect(0, 0, w, h)
        font("Menlo-Regular")
        fontSize(14)
        fill(1)
        text("Skateboard Space\nNo interaction sources", (self.margin, .5*h))
        
    
    def draw(self):
        # callback for canvasgroup delegate
        # spacewindow.draw
        self.zones = {}
        dd = Container()
        dd.w = self.w.c.width()
        dd.h = self.w.c.height()
        dd.we = dd.w - 2 * self.margin # effective width
        dd.he = dd.h - 2 * self.margin
        if self.interactionSources == None:
            # we have nothing
            save()
            self.presentEmptySpace()
            restore()
            return
        save()
        dd.xBoxes = max(2, int(dd.we / self.unitSize))
        dd.yBoxes = max(2, int(dd.he / self.unitSize))
        dd.xUnit = dd.we / dd.xBoxes
        dd.yUnit = dd.he / dd.yBoxes
    
        if self.darkMode == "dark":
            fill(.2)
        else:
            fill(1)
        rect(0, 0, dd.w, dd.h)
        
        dd.horizontalAxis = self.interactionSources.get('horizontal')
        dd.horizontalRange = self.axisRanges.get("horizontal")
        dd.verticalAxis = self.interactionSources.get('vertical')
        dd.verticalRange = self.axisRanges.get("vertical")
        stroke(None)
        for sample in self._done:
            srect = sample['rect']
            r, g, b = sample['color']
            if self.darkMode == "dark":
                a = .9
            else:
                a = .15
            fill(r, g, b, a)
            rect(*srect)
        
        save()
        dd.xOffsetForVerticalCaptions = 0
        dd.yOffsetForVerticalCaptions = 0
        
        # just the lines
        dd.yMin = dd.yDef = dd.yMax = .5 * dd.h
        dd.xDef = dd.xMin = dd.xMax = .5 * dd.w
        dd.xScale = dd.yScale = 0
        a, b, c, h = self.w.getPosSize()
        dd.fade = 1
        if h <= self.fadeThreshold-self.fadeLength:
            dd.fade = 0
        elif self.fadeThreshold-self.fadeLength < h <= self.fadeThreshold:
            dd.fade = (h-(self.fadeThreshold-self.fadeLength))/self.fadeLength
        if dd.horizontalAxis:
            if dd.horizontalRange == None:
                # 1
                dd.xDef = dd.xMin = dd.xMax = .5 * dd.w
                dd.xScale = 0
            elif dd.horizontalRange == (-1.0, 0.0):
                # 2
                dd.xMax = self.margin
                dd.xDef = dd.xMin = dd.w - self.margin
                dd.xScale = 1
            elif dd.horizontalRange == (-1.0, 0.0, 1.0):
                # 3
                dd.xMin = self.margin
                dd.xMax = dd.w - self.margin
                dd.xDef = .5 * dd.w
                dd.xScale = .5
            elif dd.horizontalRange == (0.0, 1.0):
                # 4
                dd.xMin = dd.xDef = self.margin
                dd.xMax = dd.w - self.margin
                dd.xScale = 1
        if dd.verticalAxis:
            if dd.verticalRange == None:
                # 1
                dd.yMin = dd.yDef = dd.yMax = .5 * dd.h
                dd.yScale = 0
            elif dd.verticalRange == (-1.0, 0.0):
                # 2
                dd.yDef = dd.yMax = dd.h - self.margin
                dd.yMin = self.margin
                dd.yScale = 1
            elif dd.verticalRange == (-1.0, 0.0, 1.0):
                # 3
                dd.yMin = self.margin
                dd.yMax = dd.h-self.margin
                dd.yDef =.5 * dd.h
                dd.yScale = .5
            elif dd.verticalRange == (0, 1.0):
                # 4
                dd.yMin = dd.yDef = self.margin
                dd.yMax = dd.h-self.margin
                dd.yScale = 1
        # draw axes lines            
        stroke(*self.blue)
        strokeWidth(3)
        fill(None)
        lineDash(5, 5)
        line((dd.xMin, dd.yDef), (dd.xMax, dd.yDef))
        line((dd.xDef, dd.yMin), (dd.xDef, dd.yMax))
        fill(*self.blue)
        stroke(None)
        self.smallMarker(dd.xDef, dd.yDef)
        self.smallMarker(dd.xMin, dd.yDef)
        self.smallMarker(dd.xMax, dd.yDef)
        self.smallMarker(dd.xDef, dd.yMin)
        self.smallMarker(dd.xDef, dd.yMax)

        # draw the items in the space
        font("Menlo-Regular")
        axisCaptionSize = 8.5
        fontSize(axisCaptionSize)
        ww = self.snapDistanceLine
        # rules
        if self.rules:
            self.presentRuleZones(dd)
        # cursor lines
        cursorX = cursorY = None
        if self.instanceLoc:
            px = (dd.xScale * dd.we * self.instanceLoc.get('horizontal', 0)) + dd.xDef
            py = (dd.yScale * dd.he * self.instanceLoc.get('vertical', 0)) + dd.yDef
            dd.cursorX = px - dd.xDef
            dd.cursorY = py - dd.yDef
            lineDash(3,3)
            if self.verticalSnapped:
                stroke(self.snapColor[0], self.snapColor[1], self.snapColor[2])
            else:
                if self.darkMode == "dark":
                    stroke(*self.cursorLineDark)
                else:
                    stroke(*self.cursorLineLight)
            fill(None)
            strokeWidth(1)
            line((min(px, dd.xMin), py), (max(px, dd.xMax), py))
            if self.horizontalSnapped:
                stroke(self.snapColor[0], self.snapColor[1], self.snapColor[2])
            else:
                if self.darkMode == "dark":
                    stroke(*self.cursorLineDark)
                else:
                    stroke(*self.cursorLineLight)
            line((px, min(py, dd.yMin)), (px, max(py, dd.yMax)))
            if self.darkMode == "dark":
                fill(*self.cursorLineDark)
            else:
                fill(*self.cursorLineLight)
            stroke(None)
            self.smallMarker(dd.xMax, py)
            self.smallMarker(dd.xMin, py)
            self.smallMarker(px, dd.yMin)
            self.smallMarker(px, dd.yMax)
        translate(dd.xDef, dd.yDef)
        if self.supports is not None:
            save()
            try:
                self.presentSupports(dd)
            finally:
                restore()
        padding = 2
        itemsForDrawing = []
        if self.masters:
            itemsForDrawing += self.masters
        if self.instances:
            itemsForDrawing += self.instances
        if self.interesting:
            itemsForDrawing += self.interesting
        if self.openLocations:
            itemsForDrawing += self.openLocations
        if itemsForDrawing:
            for item in itemsForDrawing:
                flavor = item['flavor']
                px = (dd.xScale * dd.we * item['location'].get('horizontal', 0))
                py = (dd.yScale * dd.he * item['location'].get('vertical', 0))
                if px == 0 and py == 0:
                    flavor = "default"
                self.masterMarker(flavor, px, py, item.get('muted', False))
                if item.get('muted', False):
                    fill(.75)
                else:
                    fill(0)
                stroke(None)
                if flavor == "open":
                    # don't draw labels for open spots
                    continue
                if self.drawLabels:
                    name = item.get("layername")
                    t = "%s" %(item['locationString'])
                    if name:
                        t += "\n%s" % name
                    lines = len(t.split('\n'))+3
                    tw, th = textSize(t)
                    tx,ty = (px-11, py-lines*axisCaptionSize-8)
                    tx = px - .5 * self.masterIconSize
                    ty = py - th - .75 * self.masterIconSize
                    if not self.darkMode == "dark":
                        fill(1,1,1,0.8 * dd.fade)
                        rect(tx-2, ty, tw+4, th)
                        fill(0,0,0,dd.fade)
                    else:
                        fill(1)
                    stroke(None)
                    text(t, (tx,ty))
        if self.instanceLoc:
            save()
            try:
                self.presentCursor(dd)
            finally:
                restore()
        restore()
    
    def presentRuleZones(self, dd):
        # [
        #     [
        #         {'horizontal': (0.0, 0.328), 'vertical': (0.0, 1.0), 'name': 'fold_I_serifs'}
        #     ],
        #     [
        #         {'horizontal': (0.0, 1.0), 'vertical': (0.0, 0.82), 'name': 'fold_S_terminals'}
        #     ]
        # ]
        fill(None)
        ruleRed = (.8,.125,0.3, .15*dd.fade)
        stroke(*ruleRed)
        inset = 10
        strokeWidth(2*inset)
        for r in self.rules:
            for cd in r:
                if cd['horizontal'] is None or cd['vertical'] is None:
                    continue
                xMin, xMax = cd['horizontal']
                yMin, yMax = cd['vertical']
                pMinx = (dd.xScale * dd.we * xMin) + dd.xDef
                pMiny = (dd.yScale * dd.he * yMin) + dd.yDef
                pMaxx = (dd.xScale * dd.we * xMax) + dd.xDef
                pMaxy = (dd.yScale * dd.he * yMax) + dd.yDef
                fill(None)
                stroke(*ruleRed)
                roundedRect(pMinx+inset, pMiny+inset, pMaxx-pMinx-2*inset, pMaxy-pMiny-2*inset, 5)
                
                cx, cy = .5*(pMinx+pMaxx), .5*(pMiny+pMaxy)
                fill(ruleRed[0],ruleRed[1],ruleRed[2],1*dd.fade)
                stroke(None)
                if self.drawLabels:
                    ruleLabel = cd['name']
                    tw, th = textSize(ruleLabel)
                    text(ruleLabel, (cx-.5*tw,cy-.5*th))
        
    def presentCursor(self, dd):
        if self.verticalSnapped or self.horizontalSnapped:
            fill(self.snapColor[0], self.snapColor[1], self.snapColor[2])
        else:

            if self.darkMode == "dark":
                fill(*self.cursorLineDark)
            else:
                fill(*self.cursorLineLight)
        stroke(None)
        oval(dd.cursorX-.5*self.masterIconSize, dd.cursorY-.5*self.masterIconSize, self.masterIconSize, self.masterIconSize)

    def presentSupports(self, dd):
        #newSupports [
        #     {'horizontal': None, 'vertical': None}, 
        #     {'horizontal': (-1.0, -1.0, 0), 'vertical': None}, 
        #     {'horizontal': (0, 1.0, 1.0), 'vertical': None}, 
        #     {'horizontal': None, 'vertical': (0, 1.0, 1.0)}, 
        #     {'horizontal': (0, 1.0, 1.0), 'vertical': (0, 1.0, 1.0)}, 
        #     {'horizontal': (0, 0.309398, 1.0), 'vertical': (0, 0.635547, 1.0)}, 
        #     {'horizontal': (0.309398, 0.704963, 1.0), 'vertical': (0, 0.36699200000000004, 1.0)}, 
        #     {'horizontal': (0.309398, 0.853447, 1.0), 'vertical': (0.36699200000000004, 0.835073, 1.0)}
        # ]
        self._supportColors = []
        l = len(self.supports)
        for i, item in enumerate(self.supports):
            h = item.get('horizontal')
            v = item.get('vertical')
            if h is None and v is None: continue    # the default
            if h is not None:
                hMin, hDef, hMax = h
                hMin = (dd.xScale * dd.we * hMin)
                hMax = (dd.xScale * dd.we * hMax)
            else:
                hMin = dd.xMin - dd.xDef
                hMax = dd.xMax - dd.xDef
            if v is not None:
                vMin, vDef, vMax = v
                vMin = (dd.yScale * dd.he * vMin)
                vMax = (dd.yScale * dd.he * vMax)
            else:
                vMin = dd.yMin - dd.yDef
                vMax = dd.yMax - dd.yDef
            w = hMax-hMin
            h = vMax-vMin
            rgb = getColor(l, i)
            self._supportColors.append(rgb)
            thisInset = self.inset
            if h <= 2*self.inset or w <= 2*self.inset:
                stroke(None)
                fill(rgb[0], rgb[1], rgb[2], .1*dd.fade)
                thisInset = 0
                rr = 5
            else:
                rr = .5 * min(w, h)
                fill(None)
                stroke(rgb[0], rgb[1], rgb[2], .1*dd.fade)
                strokeWidth(2*thisInset)
            roundedRect(hMin + thisInset, vMin + thisInset, w-2*thisInset, h-2*thisInset, rr)
            
    def savePDFCallback(self, sender):
        path = PutFile(message="Save this designspace as PDF:", fileName="skateboard.designspace.pdf")
        view = self.w.c.getNSView()
        frame = view.frame()
        pdfData = view.dataWithPDFInsideRect_(frame)
        pdfData.writeToFile_atomically_(path, True)
        
    def masterMarker(self, flavor, x, y, muted=False, color=None):
        if muted:
            alpha = .25
        else:
            alpha = 1
        save()
        lineDash(None)
        hasDrawn = False
        if flavor == "master":
            fill(1)
            stroke(1,	44/255,	121/255, alpha)
            strokeWidth(3)
            d = self.masterIconSize
        elif flavor == "instance":
            fill(1)
            stroke(1,	191/255, 	0, alpha)
            strokeWidth(3)
            d = self.masterIconSize
        elif flavor == "support":
            fill(1)
            stroke(.5, .5, .5, alpha)
            strokeWidth(3)
            d = self.masterIconSize
        elif flavor == "default":
            fill(1)
            stroke(121/255,	44/255,	1, alpha)
            strokeWidth(3)
            d = self.masterIconSize
        elif flavor == "interesting":
            stroke(None)
            fill(.5, .5, .5, alpha)
            d = .5 * self.masterIconSize
            drawPlaceMarker(x, y)
            hasDrawn = True
        else:
            stroke(None)
            fill(.25, .65, .65, alpha)
            d = .25 * self.masterIconSize
        if not hasDrawn:
            oval(x-.5*d, y-.5*d, d, d)
        restore()
    
    def smallMarker(self, x, y, markerSize=6):
        save()
        lineDash(None)
        oval(x-.5*markerSize, y-.5*markerSize, markerSize, markerSize)
        restore()
                
if __name__ == "__main__":
    from skateboard import SkateboardDesignSpaceProcessor
    class TestWrapper(object):
        def __init__(self, testPath):
            self.doc = SkateboardDesignSpaceProcessor()
            self.doc.read(testPath)
            self.space = SkateboardSpaceWindow()
            
    path = "/Users/erik/code/skateboardExtension/ufo/ComplexTest/dp.designspace"
    OpenWindow(TestWrapper, path)
    