import AppKit
from defconAppKit.windows.baseWindow import BaseWindowController
from defconAppKit.controls.glyphLineView import GlyphLineView
import vanilla
from mojo.canvas import CanvasGroup
from mojo.drawingTools import *
from mojo.events import publishEvent
from mojo.UI import MultiLineView
from fontParts.fontshell import RFont
from sbx.ui.preferencesWindow import followOpticalSizePrefKey

class SkateboardTextWindow(BaseWindowController):
    # a slave window that resembles a RF spacecenter, but shows preview interpolations.
    toolNameForTitle = "⌘Skateboard Text"
    def __init__(self, closeCallback=None, instanceFont=None, requestGlyphsCallback=None, selectGlyphCallback=None, previewText=None):
        self.closeCallback = closeCallback
        self.selectGlyphCallback = selectGlyphCallback

        self.displayPrefs = {'Show Kerning': True,
            'Center': False,
            'Show Space Matrix': False,
            'Single Line': True,
            'showLayers': [],
            'displayMode': u'Multi Line',
            'Stroke': False,
            'xHeight Cut': False,
            'Upside Down': False,
            'Beam': False,
            'Water Fall': True,
            'Inverse': False,
            'Show Template Glyphs': True,
            'Multi Line': True,
            'Right to Left': False,
            'Show Metrics': False,
            'Show Control glyphs': True,
            'Fill': True,
            'Left to Right': True}

        self.font = RFont(showInterface=False)
        self.font.info.unitsPerEm = 1500
        self.font.info.descender = -500
        self.font.info.ascender = 1000

        self.glyphOrder = []
        self.mode = "text"
        self.w = vanilla.Window((800, 400),
            minSize=(300, 200),
            title=self.toolNameForTitle,
            autosaveName = "com.letterror.skateboard.textWindow",
            fullScreenMode = "auxiliary",
            )
        self.w.mline = MultiLineView((0,0,0,-30), pointSize=80, lineHeight=0,
            selectionCallback=self.callbackGlyphLineClick,
            #doubleClickCallback=self.callbackGlyphLineDoubleClick
            )
        #print(dir(self.w.mline))
        #print(self.w.mline.getDisplayStates())
        self.w.mline.setFont(self.font)
        self.w.mline.setDisplayStates(self.displayPrefs)
        self.w.mline.setLineHeight(0)
        buttonHeight = -27
        if previewText is None:
            previewText = "SKATEBOARD"
        self.w.text = vanilla.EditText((5,buttonHeight,300,25), previewText, callback=requestGlyphsCallback, placeholder="SKATEBOARD")
        textWindowModeDescriptions = [
            dict(title="Text"),
            dict(title="Glyph"),
            ]
        self.w.textWindowMode = vanilla.SegmentedButton((320,buttonHeight, 100, 22),
            segmentDescriptions=textWindowModeDescriptions,
            callback = self.textWindowModeCallback)
        self.w.scaleOptical = vanilla.CheckBox((-100, buttonHeight, 90, 22), "opsz", callback=self.changeScaleOpticalCallback)
        colorModeDescriptions = [
            dict(title="Positive"),
            dict(title="Negative"),
            ]
        self.w.colorMode = vanilla.SegmentedButton((420,buttonHeight, 140, 22),
            segmentDescriptions=colorModeDescriptions,
            callback = self.colorModeCallback)
        self.w.bind("close", self.cbClose)
        #self.w.bind("became main", self.cbBecameMain)
        self.w.open()
        self.w.textWindowMode.set([1])
        self.w.colorMode.set([1])
    
    def changeScaleOpticalCallback(self, sender):
        pass
        #print('changeScaleOpticalCallback', sender.get())
            
    def shouldDrawBackground(self):
        # callback for canvasgroup delegate
        return False
    
    def textWindowModeCallback(self, sender):
        #print("textWindowModeCallback", sender.get())
        choice = sender.get()
        if choice == 0:
            # text mode
            self.mode = "text"
            self.w.text.enable(True)
        elif choice == 1:
            # glyph explore mode
            self.mode = "glyph"
            self.w.text.enable(False)
        publishEvent("skateboardTextWindow.mode", mode = self.mode)
        #self.showing = list(set([g.name for g in glyphsForPreview]))
        
    def colorModeCallback(self, sender):
        choice = sender.get()
        if choice == 0:
            self.displayPrefs['Inverse'] = False
            self.w.mline.setDisplayStates(self.displayPrefs)
        elif choice == 1:
            self.displayPrefs['Inverse'] = True
            self.w.mline.setDisplayStates(self.displayPrefs)
        
    # def draw(self):
    #     # callback for canvasgroup delegate
    #     # http://robofont.com/documentation/building-tools/api/mojo/mojo-canvas/
    #     units = 8
    #     w = self.w.c.width()
    #     h = self.w.c.height()
    #     fill(1,0,0)
    #     rect(100,100, 10, 10,)
        
    def getRawText(self):
        return self.w.text.get()

    def setOpticalPointSize(self, opticalSize):
        if self.w.scaleOptical.get() == 1:
            if opticalSize is not None:
                self.w.mline.setPointSize(opticalSize)
        
    def setGlyphs(self, glyphsForPreview, opticalSize=None, unitsPerEm=1500, ascender=1000, descender=500):
        self.font.info.unitsPerEm = unitsPerEm
        self.font.info.ascender = ascender
        self.font.info.descender = descender
        self.w.mline.setFont(self.font)        
        try:
            self.w.mline.set(glyphsForPreview)
            if opticalSize is not None:
                self.w.mline.setPointSize(opticalSize)
        except TypeError:
            print("textwindow error setGlyphs @@")

    def cbClose(self, sender):
        # vanilla callback space window
        if self.closeCallback:
            self.closeCallback()
    
    def mainWindowIsClosing(self, sender=None):
        # callback when mainwindow is closing and spacewindow is still open.
        self.font = None
        self.w.close()
        
    def cbBecameMain(self, sender):
        # vanilla callback space window
        #print("SkateboardTextWindow.cbBecameMain", sender)
        pass
        
    def callbackGlyphLineClick(self, sender):
        # will call skateboard.textWindow_selectGlyph
        #print("SkateboardTextWindow.callbackGlyphLineClick", sender)
        if self.selectGlyphCallback is not None:
            selection = self.w.mline.getSelectedGlyph()
            #print("SkateboardTextWindow.callbackGlyphLineClick selection", selection)
            self.selectGlyphCallback(sender)

    def callbackGlyphLineDoubleClick(self, sender):
        #
        #print("SkateboardTextWindow.callbackGlyphLineDoubleClick")
        pass
