# coding: utf-8
# the best and smallest viable designspace implementation in RF3

from mojo.roboFont import version

"""
    
    new Math objects:
        - kernlet
        - font proportions
        - glyph margins and width


    Something about timers
https://developer.apple.com/library/archive/documentation/Performance/Conceptual/EnergyGuide-iOS/MinimizeTimerUse.html

xx
"""

import AppKit
import os, time, math, random, sys
import weakref
import importlib
import traceback
from types import MethodType
import webbrowser

import sbx
import sbx.tools.sp3

from fontTools.misc.transform import Transform
try:
    AllFonts()
except NameError:
    try:
        from lib.fontObjects.fontPartsWrappers import *
    except:
        print("Skateboard: Can't import AllFonts")

try:
    from variableFontGenerator import BatchDesignSpaceProcessor
    hasVariableFontGenerator = True
except ImportError:
    hasVariableFontGenerator = False
try:
    from designSpaceEditorwindow import DesignSpaceEditor
    hasDesignSpaceEditor = True
except ImportError:
    hasDesignSpaceEditor = False

try:
    import glyphNameFormatter
    hasGlyphNameFormatter = True
except ImportError:
    hasGlyphNameFormatter = False

try:
    import designspaceProblems
    hasDesignspaceProblems = True
except ImportError:
    hasDesignspaceProblems = False


from sbx.tools.editions import getDaysRemainingFromSkateboardBundle   
from pprint import pprint
 

import vanilla
import defconAppKit
from defconAppKit.windows.baseWindow import BaseWindowController
from mojo.events import addObserver, removeObserver, installTool, uninstallTool, publishEvent
from mojo.UI import GlyphWindowToPDF, PutFile, ProgressBar, CurrentFontWindow, CurrentGlyphWindow, OpenGlyphWindow, setDefault, getDefault, AllFontWindows, AllGlyphWindows, UpdateCurrentGlyphView, OpenSpaceCenter, AskString, GetFile, SetCurrentLayerByName
from mojo.UI import AskYesNoCancel as mojo_AskYesNoCancel
from vanilla.dialogs import askYesNoCancel as vanilla_AskYesNoCancel
from vanilla.dialogs import getFile as vanilla_getFile

import ufoProcessor
#importlib.reload(ufoProcessor)
import ufoProcessor.varModels
#import ufoProcessor.sp3
from ufoProcessor.emptyPen import checkGlyphIsEmpty

import fontTools.designspaceLib as _designspaceLib
from fontTools.pens.areaPen import AreaPen
from fontTools.designspaceLib import DesignSpaceDocument, SourceDescriptor, InstanceDescriptor, AxisDescriptor, RuleDescriptor, processRules
#from fontTools.varLib.models import normalizeValue, normalizeLocation
import sbx.tools.normalize
importlib.reload(sbx.tools.normalize)
from mutatorMath.objects.mutator import Location, buildMutator
import mutatorMath.objects.mutator
from fontMath.mathGlyph import MathGlyph

from getDesignSpaceWindow import CurrentDesignSpace, getDesignSpaceDocuments
import sbx.tools.reader
importlib.reload(sbx.tools.reader)
import sbx.ui.textWindow
importlib.reload(sbx.ui.textWindow)
import sbx.ui.spaceWindow
importlib.reload(sbx.ui.spaceWindow)

import mojo.drawingTools as ctx
from mojo.UI import MultiLineView
from mojo.compile import executeCommand
from fontTools.pens.basePen import BasePen
from fontTools.pens.transformPen import TransformPen
#from defcon.objects.font import Font as DefconFont
#from fontParts.fontshell import RFont
#from fontParts.fontshell import RGlyph
from mojo.roboFont import RFont, RGlyph, RLayer

import mojo.extensions

import navigatorTool
importlib.reload(navigatorTool)
import sbx.ui.sheets
importlib.reload(sbx.ui.sheets)
import sbx.tools.reader
import random

from AppKit import NSPasteboardTypeString, NSPasteboard, NSImageCell, NSImageAlignTop, NSScaleNone, NSImageFrameNone
from fontTools.ufoLib.glifLib import writeGlyphToString
from sbx.tools.sortLocations import findOpenNodes
import sbx.tools.pens
importlib.reload(sbx.tools.pens)
from sbx.tools.pens import *
import sbx.tools.storageObject
import sbx.tools.findCluster
importlib.reload(sbx.tools.findCluster)

from defcon import addRepresentationFactory
import mojo.canvas
import sbx.ui.preferencesWindow
importlib.reload(sbx.ui.preferencesWindow)
import sbx.tools.colorMachine
importlib.reload(sbx.tools.colorMachine)

def GlyphEmptyFactory(glyph, font=None):
    from ufoProcessor.emptyPen import checkGlyphIsEmpty
    return checkGlyphIsEmpty(glyph)

addRepresentationFactory("Skateboard.isGlyphEmpty", GlyphEmptyFactory)


class Container(object):
    pass



DEVELOP = False
if DEVELOP:
    pathForBundle = os.path.dirname(__file__)
    resourcePathForBundle = os.path.join(os.path.dirname(pathForBundle), "resources")
    skateboardBundle = mojo.extensions.ExtensionBundle(path=pathForBundle, resourcesName=resourcePathForBundle)
else:
    skateboardBundle = mojo.extensions.ExtensionBundle("SkateBoard")

class SkateboardDesignSpaceProcessor(ufoProcessor.DesignSpaceProcessor):
    def __init__(self, readerClass=None, writerClass=None, fontClass=None, ufoVersion=3, useVarlib=False):
        super(SkateboardDesignSpaceProcessor, self).__init__(readerClass=readerClass, writerClass=writerClass)

    def update(self):
        # placeholder 
        pass
        
    def _instantiateFont(self, path):
        """ Return a instance of a font object with all the given subclasses"""
        for f in AllFonts():
            if f.path == path and f.path is not None:
                return f
        return RFont(path, showInterface=False)
    
    def loadFonts(self, reload=False):
        # Load the fonts and find the default candidate based on the info flag
        # pay attention:
        #     1. different sources can reference different layers in the same ufo
        #     2. also: the same source can appear in different places in the designspace
        # so maybe the sourcedescriptor.name is not a good identifier
        if self._fontsLoaded and not reload:
            return
        names = set()
        for sourceDescriptor in self.sources:
            if not sourceDescriptor.name in self.fonts:
                pathOK = True
                if sourceDescriptor.path is not None:
                    if os.path.exists(sourceDescriptor.path):
                        self.fonts[sourceDescriptor.name] = self._instantiateFont(sourceDescriptor.path)
                        # this is not a problem, why report it as one?
                        names = names | set(self.fonts[sourceDescriptor.name].keys())
                else:
                    pathOK = False
                if not pathOK:
                    self.fonts[sourceDescriptor.name] = None
                    self.problems.append("can't load master from %s"%(sourceDescriptor.path))
        self.glyphNames = list(names)
        self._fontsLoaded = True



class SkateBoard(BaseWindowController):
    buyButtonURL = "https://extensionstore.robofont.com/extensions/skateboard/"
    editionLibKey = 'com.letterror.skateboardEdition'    # edition info from the bundle
    navigatorPrefsLibKey = "com.letterror.skateboard.navigator"
    superpolatorDataLibKey = "com.superpolator.data"    # used by superpolator
    interestingLocationLibKey = "com.letterror.skateboard.interestingLocation"
    interactionSourcesLibKey = "com.letterror.skateboard.interactionSources"
    mutedSourcesLibKey = "com.letterror.skateboard.mutedSources"
    mutedAxesLibKey = "com.letterror.skateboard.mutedAxes"
    previewLocationLibKey = "com.letterror.skateboard.previewLocation"
    previewTextLibKey = "com.letterror.skateboard.previewText"
    showSparksPrefKey = "com.letterror.skateboard.showSparks"
    previewMathPrefKey = "com.letterror.mathModelPref"
    glyphSourceLocationLibKey = "com.letterror.skateboard.sourceLocation"    # the location a glyph outline originated from
    toolNameForTitle = "⌘Skateboard"
    previewFolderName = "Skateboard Previews"    
    instancesFolderName = "Skateboard Instances"    
    buttonWidth = 60
    vanillaWindowWidth = 500
 
    def _getEdition(self):
        remaining = getDaysRemainingFromSkateboardBundle()
        expireDate, edition = skateboardBundle.infoDictionary.get(self.editionLibKey, (None, None))
        return remaining, edition   

    def hasInteractionSources(self, axisName = None):
        if axisName:
            for direction, axes in self.interactionSources.items():
                if axisName in axes:
                    return True
            return False
        for direction, axes in self.interactionSources.items():
            if axes:
                return True
        return False
        
    def __init__(self, path=None):

        self.textWindow = None
        self.spaceWindow = None
        self.spaceWindowData = {}
        self.textWindowMode = "text"
        self.textRawNames = []
        
        self._documentIconImage = None
        self.delegateDrawCalls = []
        
        # the indentifier and names of the interaction sources
        self.interactionSourcesNames = [ "vertical", "horizontal", "ignore"]    #, "animate" ]
        # this maps the identifier to a list of axis names
        self.interactionSources = {"vertical":[], "horizontal":[], "ignore":[]}
        self._defconObservers_glyphs = {}
        self._defconObservers_fonts = {}
        
        self._working = {}    # store information about which glyphs are working
        self.document_modificationTime = None
        self.thingsWhileDragging = None    # container for things that don't change while dragging
        
        self._drawPreviewWithNavigator = True
        self._drawPreviewsElsewhere = True
        
        self._problemResponseScheme = sbx.ui.preferencesWindow.getProblemsResponse()
        self._colorsTransparent = False
        self._colorScheme = None
        self.resetColorScheme()        

        self._measureLineCurveOffset = 40
        self._captionTextSize = 8
        self._captionTextFont = "Menlo-Regular"
        self.documentAscender = None
        self.documentDescender = None
        self.documentUnitsPerEm = None


        value = getDefault(self.showSparksPrefKey)
        if value is None:
            self.showSparksPref = False
        else:
            self.showSparksPref = True
        self._drawAnchors = True
        self._showMasters = True
        self._drawMetrics = True
        self._trackSelection = True
        self._selectedMarkerSizeScale = 2.2
        self._markerSize = 2.5
        self.marginIndicatorLineHeight = 30
        
        self.maxCollectors = 10
        #self._lastFactors = []
        self._lastOpenNodes = None
        self._lastDrawingValues = None    # values collected in drwing the background
        self.pointLoad = 0      # total points in the preview
        self._maxPoints = 2000  # number of points we can handle without problems
        self.axisRange = 250   # arbitrary units to map interaction sources to 
        self.axisSteps = 100    # number of steps along one axis with cursor keys, should be pref
        self.verbose = False
        self._allowExtrapolation = False
        self.previewTextRaw = "SKATEBOARD"
        self.snapValues = {}
        self.snapped = []    # list of axisnames that are snapping at the moment
        self._listSources = True
        self._listSupports = False
        self._listInstances = True
        self._listInteresting = True
        self._listOpen = False
        self.documentNeedSave = None
        self.currentlyDragging = False
        self.currentSupport = None
        self.currentSource = None
        self.useAxisMaps = False
        self.axisMapper = None
        self.storage_mainPreviews = sbx.tools.storageObject.StorageContainer()
        self.storage_supportPreviews = sbx.tools.storageObject.StorageContainer()
        self.storage_fontInfo = sbx.tools.storageObject.StorageContainer()
        
        self.placeholderGlyph = RGlyph()
        p = self.placeholderGlyph.getPen()
        p.moveTo((20,0))
        p.lineTo((800,780))
        p.lineTo((780,800))
        p.lineTo((0,20))
        p.closePath()
        p.moveTo((0,780))
        p.lineTo((780,0))
        p.lineTo((800,20))
        p.lineTo((20,800))
        p.closePath()
        self.placeholderGlyph.leftMargin = 100
        self.placeholderGlyph.rightMargin = 100
        
        self._attrMargins = {
            AppKit.NSFontAttributeName: AppKit.NSFont.fontWithName_size_(self._captionTextFont, self._captionTextSize),
            AppKit.NSForegroundColorAttributeName: AppKit.NSColor.blackColor(),
        }
        #@@@
        self._attrMeasurements = {
            AppKit.NSFontAttributeName: AppKit.NSFont.fontWithName_size_(self._captionTextFont, self._captionTextSize),
            AppKit.NSForegroundColorAttributeName: AppKit.NSColor.redColor(),
        }

        defaultEditWindowWidth = getDefault("glyphViewDefaultWidth")
        if defaultEditWindowWidth < 2000:
            setDefault("glyphViewDefaultWidth", 2000)

        self.ds = None
        self.checker = None

        # initialise attributes
        self.locationFonts = []    # list of Font objects for each location
        self.seriesLocations = []     # list for all locations in this document
        self.pointsOfInterest = []    # list of user picked locations
        self.sourceGlyphObjects = []    # for processing selections
        self.currentGlyphName = None
        self.currentFontPath = None    # the path of the current font 
        self.currentGlyph = None     # the glyph in the current edit window
        self.instanceGlyphObject = RGlyph()    # RGlyph object of the current instance
        self.instanceGlyphObject_noSupport = RGlyph()
        self.hasNoSupportGlyph = False
        self.contouredInstanceGlyphObject = RGlyph()    # RGlyph object with all the contours
        self.instanceFont = RFont(showInterface=False)
        self.instanceFont.info.familyName = "Skateboard Preview"
        self.instanceFont.info.styleName = ""
        self.instanceFont.info.unitsPerEm = 2000
        self.instanceFont.info.descender = -1000
        self.instanceFont.info.ascender = 750
        self.previewFonts = {}
        
        self.textMarginScale = 1.2     # scale factor for the vertical metrics of the text window. JUst some padding to make it fit.
        self.instanceSpaceCenter = None    #OpenSpaceCenter(self.instanceFont)
        self.previewLocationPref = "previewCentered"
        self.previewDrawPref = "previewSolid"
        self.previewMathPref = "previewVarLib"
        self.canExtrapolate = self.previewMathPref != "previewVarLib"
        self.previewRoundPref = False
        self.drawingWithToolActivePref = False
        self.textGlyphNames = ["S", "K"]
        self.instanceFontNeededGlyphs = []
        self.selectNearOffCurves = True
        self.selectedSegments = []
        self.fontPaths = []
        self.glyphMutators = {}
        self.lastSupports = None
        self.openFonts = {}    # fonts, actually open. key:fontpath, value: fontobject
        self.needFonts = {}    # fonts, needed, but closed
        self.missingFonts = {} # fonts, missting
        self.mutedFonts = {}   # fonts, muted
        self.mutedAxes = {}    # axes, muted
        self.instanceLocation = {}    # preview location for the background/designspace values
        #self.instanceLocationMapped = {}
        self.allSelectedLocations = [] # list of all selected locations
        self.axisNames = []
        self.axisExtremes = {}
        self.lastSignal = None
        self.interval = None
        #self.somethingSelector = 0        # test for 3rd encoder
        
        self.button1State = False
        self.button2State = False
        self.button3State = False
        self.instanceKerningMutator = None
        self.instanceInfoMutator = None
        self.lastKerning = None    # the last kerning generated
        self.lastKerningPairs = None    # the pairs for the last kerning
        self.lastKerningLocation = None # the location for the last kerning
        
        self.currentSelection = {}
        self.currentSelectedClusters = {}
        self.trackedClusters = {}
        self.colinearTolerance = 100    # area of polygon
        self.currentGlyphViews = {}    # dict with glyphnames of all windows that are open

        self.slowAnimationTimer = None
        self.slowAnimationInterval = .2
        self.slowAnimationTimerTick = 0
        self.slowAnimationTimerRunning = False
        self.miniKBDScrubbingTimer = None
        self._miniKBDScrubbingTimerHits = 0
        self.miniKBDScrubbingTimerInterval = .45
        self.redrawTimer = None
        self._redrawTimerHits = 0
        self.redrawStillDraggingTimerInterval = .008
        self.mouseUpTimer = None
        self.textRedrawTimer = None
        self.mouseUpDraggingDoneTimerInterval = .05
        self.textRedrawInterval = .05
        self.textRedrawGlyphHistory = {}
        self.textRedrawOrder = []
        self.textRedrawComps = []
        self.drawDetails = True
        self.navigatorTool = None
        self.glyphChanges = {}    # keep track of edits
        self.glyphCache = {}     # cacher

        self.localEventMonitor = AppKit.NSEvent.addLocalMonitorForEventsMatchingMask_handler_(AppKit.NSKeyDownMask, self.mainWindow_monitorLocalEventsCallback_)
        
        self.mainWindow_addObservers()
        self.mainWindow_buildVanillaWindow()

        self.w.open()
        self.mainWindow_enableInterface(False)
        self.getOpenDesignspacePaths()


        publishEvent("skateboard.window.opened",
            location=None,
            designspace=None,
            interactionSources=None,
            axes=None,
            instanceGlyph = self.instanceGlyphObject,
            previewGlyph = self.contouredInstanceGlyphObject,
            )
        self.data_instanceLocationChanged()
        if path is not None:
            self.mainWindow_openDocumentWithPath(path)
    
    def resetProblemsResponse(self):
        # handler for changes in problems response
        response = sbx.ui.preferencesWindow.getProblemsResponse()
        self._problemResponseScheme = response
        
    def resetColorScheme(self):
        # handler for changes in colorscheme preference
        self._colorScheme = colorScheme = sbx.ui.preferencesWindow.getColorScheme()
        self._colorsTransparent = getDefault("skateboard.glyphViewPreviewTransparent", False)
        self._glyphViewPreviewFillColor = getDefault("glyphViewPreviewFillColor")
        self._glyphViewPreviewBackgroundColor = getDefault("glyphViewPreviewBackgroundColor")
        
        if colorScheme == "bright" or colorScheme==None:
            self._fillPreview_varLibColor = sbx.tools.colorMachine.fillPreview_varLibColor_bright    #(226/255, 222/255, 222/255)
            self._fillPreview_mutMathColor = sbx.tools.colorMachine.fillPreview_mutMathColor_bright    #(219/255,227/255,227/255)
            self._strokePreview_filledStrokeColor = sbx.tools.colorMachine.strokePreview_filledStrokeColor_bright    #(170/255, 170/255, 170/255)
            self._measureLineColor = sbx.tools.colorMachine.measureLineColor_bright    #(1,0,0.4)
            self._sparkStrokeColor = (1,.5,0)
            self._offCurveColor = (1,.5,0)
        elif colorScheme == "dark":
            self._fillPreview_varLibColor = sbx.tools.colorMachine.fillPreview_varLibColor_dark    #((255-226)/255, (255-222)/255, (255-222)/255)
            self._fillPreview_mutMathColor = sbx.tools.colorMachine.fillPreview_mutMathColor_dark    #((255-219)/255,(255-227)/255,(255-227)/255)
            self._strokePreview_filledStrokeColor = sbx.tools.colorMachine.strokePreview_filledStrokeColor_dark    #((255-170)/255, (255-170)/255, (255-170)/255)
            self._measureLineColor = sbx.tools.colorMachine.measureLineColor_dark    #(1,0,0.4)
            self._sparkStrokeColor = (1,.5,0)
            self._offCurveColor = (1,.5,0)
        if self.spaceWindow is not None:
            self.spaceWindow.darkMode = colorScheme

    ## external scripting
    def getDesignSpace(self):
        return self.ds
        
    ## observers
    def mainWindow_addObservers(self):
        # observer stuff
        self._skateboard_observes_these = [
            ('observed_fontDidOpenCallback', "fontDidOpen"),
            ('observed_fontDidCloseCallback', "fontDidClose"),
            ('observed_currentGlyphChangedCallback', 'currentGlyphChanged'),
            ('observed_glyphWindowWillOpenCallback', 'glyphWindowWillOpen'),
            ('observed_glyphWindowWillCloseCallback', 'glyphWindowWillClose'),
            ('glyphWindow_drawBackgroundWrapper', 'drawBackground'),
            ('glyphWindow_drawBackgroundWrapper', 'drawBackgroundInactive'),
            ('navigatorTool_mouseDraggedCallback', 'mouseDragged'),
            ('navigatorTool_mouseUpCallback', 'mouseUp'),
            ('navigatorTool_data_locationContextCallback', 'navigatorToolEvent'),
            ('navigatorTool_menuCallback', 'navigatorToolEvent.menu'),
            ('navigatorTool_activeCallback', 'navigatorToolEvent.active'),
            ('navigatorTool_inactiveCallback', 'navigatorToolEvent.inactive'),
            ('navigatorTool_clickCallback', 'navigatorToolEvent.click'),
            ('navigatorTool_clickCallback', 'navigatorToolEvent.mousedown'),
            ('navigatorTool_clickCallback', 'navigatorToolEvent.mouseup'),
            ('textWindow_changedSettingsCallback', 'skateboardTextWindow.mode'),
            ('spaceWindow_dragEventCallback', 'skateboardSpaceWindowEvent'),
            ('mainWindow_confirmSkateboardRunningCallback', 'skateboard.running'),
            ('mainWindow_openNewDocument', 'skateboard.openDocument'),
            ('mainWindow_newGlyphsCallback', 'glyphbrowser.newGlyphs'),
            ('mainWindow_handleMiniKbdNotification', "minikbd.2encoder"),
            ('mainWindow_handleControlBoardNotification', "ControlBoardInput"),
            
            ('glyphWindow_glyphAdditionContextualMenuItems', 'glyphAdditionContextualMenuItems')
        ]
        for methodName, eventName in self._skateboard_observes_these:
            addObserver(self, methodName, eventName)

    def mainWindow_removeObservers(self):
        for methodName, eventName in self._skateboard_observes_these:
            removeObserver(self, eventName)

    def mainWindow_buildVanillaWindow(self):
        self._emptyImagePath = os.path.join(skateboardBundle.resourcesPath(), "ufo_is_empty.pdf")
        self._emptyImage = AppKit.NSImage.alloc().initWithContentsOfFile_(self._emptyImagePath)
        self._ufoNeutralImagePath = os.path.join(skateboardBundle.resourcesPath(), "ufo_is_neutral.pdf")
        self._ufoNeutralImage = AppKit.NSImage.alloc().initWithContentsOfFile_(self._ufoNeutralImagePath)
        self._ufoMasterImage = AppKit.NSImage.alloc().initWithSize_((20, 20))

        self._axisControlImage = AppKit.NSImage.alloc().initWithSize_((20, 20))
        self._axisControlImage = self._ufoNeutralImage

        self._openDocImagePath = os.path.join(skateboardBundle.resourcesPath(), "button_txt_open.pdf")
        self._openDocImage = AppKit.NSImage.alloc().initWithContentsOfFile_(self._openDocImagePath)
        self._closeDocImagePath = os.path.join(skateboardBundle.resourcesPath(), "button_txt_close.pdf")
        self._closeDocImage = AppKit.NSImage.alloc().initWithContentsOfFile_(self._closeDocImagePath)
        self._saveDocImagePath = os.path.join(skateboardBundle.resourcesPath(), "button_txt_save.pdf")
        self._saveDocImage = AppKit.NSImage.alloc().initWithContentsOfFile_(self._saveDocImagePath)

        self._ufoMutedImagePath = os.path.join(skateboardBundle.resourcesPath(), "ufo_is_muted.pdf")
        self._ufoMutedImage = AppKit.NSImage.alloc().initWithContentsOfFile_(self._ufoMutedImagePath)
        self._snappedImagePath = os.path.join(skateboardBundle.resourcesPath(), "snapped.pdf")
        self._snappedImage = AppKit.NSImage.alloc().initWithContentsOfFile_(self._snappedImagePath)
        self._ufoAvailableImagePath = os.path.join(skateboardBundle.resourcesPath(), "ufo_is_available.pdf")
        self._ufoAvailableImage = AppKit.NSImage.alloc().initWithContentsOfFile_(self._ufoAvailableImagePath)
        self._ufoOpenImagePath = os.path.join(skateboardBundle.resourcesPath(), "ufo_is_open.pdf")
        self._ufoOpenImage = AppKit.NSImage.alloc().initWithContentsOfFile_(self._ufoOpenImagePath)
        self._ufoCurrentImagePath = os.path.join(skateboardBundle.resourcesPath(), "ufo_is_current.pdf")
        self._ufoCurrentImage = AppKit.NSImage.alloc().initWithContentsOfFile_(self._ufoCurrentImagePath)
        self._ufoMissingImagePath = os.path.join(skateboardBundle.resourcesPath(), "ufo_is_missing.pdf")
        self._ufoMissingImage = AppKit.NSImage.alloc().initWithContentsOfFile_(self._ufoMissingImagePath)
        
        logoPath = "skateboard_icon_512{}.pdf"
        versionedPath = "skateboard_icon_512.pdf" # default
        v = None

        self._remaining = None
        remaining, edition = self._getEdition()
        for v in ['student', 'trial', 'develop']:
            if v in edition:
                versionedPath = logoPath.format("_"+v)
        proposedPath = os.path.join(skateboardBundle.resourcesPath(), versionedPath)
        if os.path.exists(proposedPath):
            #print('found image', proposedPath)
            self._documentIconImage = AppKit.NSImage.alloc().initWithContentsOfFile_(proposedPath)
        #else:
            #print("missing image", proposedPath)

        self._newGlyphImagePath = os.path.join(skateboardBundle.resourcesPath(), "button_explore.pdf")
        self._newGlyphImage = AppKit.NSImage.alloc().initWithContentsOfFile_(self._newGlyphImagePath)

        self._textButtonImagePath = os.path.join(skateboardBundle.resourcesPath(), "button_txt_text.pdf")
        self._textButtonImage = AppKit.NSImage.alloc().initWithContentsOfFile_(self._textButtonImagePath)
        self._spaceButtonImagePath = os.path.join(skateboardBundle.resourcesPath(), "button_txt_space.pdf")
        self._spaceButtonImage = AppKit.NSImage.alloc().initWithContentsOfFile_(self._spaceButtonImagePath)
        self._exportButtonImagePath = os.path.join(skateboardBundle.resourcesPath(), "button_txt_ufo.pdf")
        self._exportButtonImage = AppKit.NSImage.alloc().initWithContentsOfFile_(self._exportButtonImagePath)
        self._addGlyphButtonImagePath = os.path.join(skateboardBundle.resourcesPath(), "button_txt_glyph.pdf")
        self._addGlyphButtonImage = AppKit.NSImage.alloc().initWithContentsOfFile_(self._addGlyphButtonImagePath)

        self._makeFontImagePath = os.path.join(skateboardBundle.resourcesPath(), "button_txt_variablefont.pdf")
        self._makeFontButtonImage = AppKit.NSImage.alloc().initWithContentsOfFile_(self._makeFontImagePath)
        self._editDocImagePath = os.path.join(skateboardBundle.resourcesPath(), "button_txt_edit.pdf")
        self._editDocButtonImage = AppKit.NSImage.alloc().initWithContentsOfFile_(self._editDocImagePath)
        self._preferencesImagePath = os.path.join(skateboardBundle.resourcesPath(), "button_preferences.pdf")
        self._preferencesButtonImage = AppKit.NSImage.alloc().initWithContentsOfFile_(self._preferencesImagePath)

        # column descriptions for the source list
        sourceListColumnDescriptions = [
            {'title': '', 'key':"isMutedImage", 'width': 18, 'cell':vanilla.ImageListCell()},
            {'title': "", 'key':"isOpenImage", 'width': 18, 'cell':vanilla.ImageListCell()},
            {'title': "", 'key':"isNeutralImage", 'width': 18, 'cell':vanilla.ImageListCell()},
            #{'title': "id", 'key':"_fontid", 'width': 100},
            #{'title': "Object", 'key':"_fontObject", 'width': 50},
            {'title': "UFO", 'key':"name", 'width': 280},
            #{'title': "Glyph", 'key':"_glyphObject", 'width': 100},
            {'title': "Layer", 'key':'layer', 'width': 200},
            #{'title': "Factor", 'key':'factor', 'width': 100}
            ]
        locationListDescriptions = [
            #{'title': "", 'key':"control", 'width': 30},
            {'title': "•", 'key':"description", 'width': 350},
            {'title': "Flavor", 'key':"flavor", },#'width': 100},
            ]
        interactionSourcesPopup = vanilla.PopUpButtonListCell(self.interactionSourcesNames)
        previewAxesList_columnDescriptions = [
            {'title': '', 'key':"isMutedImage", 'width': 18, 'cell':vanilla.ImageListCell()},
            {'title': '', 'key':"isSnappedImage", 'width': 18, 'cell':vanilla.ImageListCell()},
            {'title': "Name", 'key':"axisKey", 'width': 90, 'editable':False},
            {'title': "Designspace", 'key':"valueKey", 'width': 80, 'editable':True},
            #{'title': "ValueAniso", 'key':"valueAnisoKey", 'width': 80, 'editable':True},
            {'title': "Userspace", 'key':"mappedValueKey", 'width': 80, 'editable':False},
            {'title': "Normalised", 'key':"normalisedValueKey", 'width': 80, 'editable':False},
            #{'title': "Min", 'key':"minimumKey", 'width': 60, 'editable':False},
            #{'title': "Default", 'key':"defaultKey", 'width': 60, 'editable':False},
            #{'title': "Max", 'key':"maximumKey", 'width': 60, 'editable':False},
            {'title': "Interaction", 'key':"controlKey", 'width': 100, 'cell': interactionSourcesPopup,  "binding": "selectedValue"},
            ]
        start_fileThings = 0
        height_fileThings = 40
        
        start_sourceThings = start_fileThings + height_fileThings
        height_sourceThings = 170
        
        start_previewThings = start_sourceThings + height_sourceThings
        height_previewThings = 150
        
        start_locationThings = start_previewThings + height_previewThings
        height_locationThings = 150
        self.w = vanilla.Window((self.vanillaWindowWidth, 679),
            self.toolNameForTitle,
            minSize=(500, 500),
            maxSize=(800,2000),
            autosaveName = "com.letterror.skateboard.mainWindow",
            fullScreenMode = None,
            titleVisible = True,
            )
            
        # effect = AppKit.NSVisualEffectView.alloc().init()
        # effect.setBlendingMode_(AppKit.NSVisualEffectBlendingModeBehindWindow)
        # self.w.getNSWindow().setTitlebarAppearsTransparent_(True)
        # self.w.getNSWindow().setTitleVisibility_(True)
        # mask = self.w.getNSWindow().styleMask()
        # self.w.getNSWindow().setStyleMask_(mask | AppKit.NSWindowStyleMaskFullSizeContentView)
        # self.w.getNSWindow().setContentView_(effect)
            
        self.w.file = vanilla.Group((0, 0,0, height_fileThings))
        #self.w.file.l = vanilla.HorizontalLine((0,-1, 0, 1))
        self.pane_doc = vanilla.Group((0, 0,0, 0))
        self.pane_previewAxes = vanilla.Group((0, 0, 0, 0))
        
        self.w.frontPage = vanilla.Group((0,0,0,0))
        self.w.frontPage.logo = vanilla.ImageView((0,50,0,0), horizontalAlignment="center", verticalAlignment="center", scale="proportional")
        self.w.frontPage.logo.setImage(imageObject = self._documentIconImage)
        self.w.frontPage.openDesignSpaces = vanilla.PopUpButton((10, 50, 300, 100), [], callback=self.document_selectCandidateDesignspace, sizeStyle="small")
        self.w.frontPage.openDesignSpaces.show(False)

        #self.w.frontPage.canvas = mojo.canvas.CanvasGroup((0,0,0,0), delegate=self)

        self.mainWindow_refreshCandidateDocumentsList()
        
        ww = 50
        wy = 40
        #self.w.frontPage.logoButton = vanilla.ImageButton((0, 0, ww, wy), imageObject=self._openDocImage, bordered=False, callback=self.document_selectAndOpenDesignSpaceCallback, sizeStyle="regular")
        self.w.frontPage.openButton = vanilla.ImageButton((0, 0, self.buttonWidth, wy),
            imageObject=self._openDocImage,
            bordered=False, 
            callback=self.document_openDesignSpaceCallback,
            sizeStyle="regular")
        buttonSlotsLeft = [self.buttonWidth * i for i in range(10)]
        buttonSlotsRight = [- self.buttonWidth * i for i in range(1,10)]
        self.w.file.closeButton = vanilla.ImageButton((buttonSlotsLeft[0], 0, self.buttonWidth, wy), imageObject=self._closeDocImage, bordered=False, callback=self.document_closeDesignSpaceCallback, sizeStyle="regular")
        self.w.file.saveButton = vanilla.ImageButton((buttonSlotsLeft[1], 0, self.buttonWidth, wy), imageObject=self._saveDocImage, bordered=False, callback=self.document_saveDesignSpaceCallback_checkForSaveAs, sizeStyle="regular")
        self.w.file.saveButton.enable(False)
        self.w.file.closeButton.show(False)
        self.w.file.spaceWindowButton = vanilla.ImageButton((buttonSlotsRight[1], 0, self.buttonWidth, wy), imageObject=self._spaceButtonImage, bordered=False, callback=self.mainWindow_openSpaceWindow, sizeStyle="regular")
        self.w.file.textWindowButton = vanilla.ImageButton((buttonSlotsRight[2], 0, self.buttonWidth, wy), imageObject=self._textButtonImage, bordered=False, callback=self.mainWindow_openTextWindow, sizeStyle="regular")
        self.w.file.exportUFOButton = vanilla.ImageButton((buttonSlotsRight[0], 0, self.buttonWidth, wy), imageObject=self._exportButtonImage, bordered=False, callback=self.export_generateUFOCallback, sizeStyle="regular")
        if hasVariableFontGenerator:
            self.w.file.makeVFButton = vanilla.ImageButton((buttonSlotsLeft[3], 0, self.buttonWidth, wy), imageObject=self._makeFontButtonImage, bordered=False, callback=self.export_makeVariableFontCallback, sizeStyle="regular")
        if hasDesignSpaceEditor:
            self.w.file.editDSButton = vanilla.ImageButton((buttonSlotsLeft[2], 0, self.buttonWidth, wy), imageObject=self._editDocButtonImage, bordered=False, callback=self.mainWindow_openDSEditorCallback, sizeStyle="regular")
        self.w.file.addGlyphButton = vanilla.ImageButton((buttonSlotsRight[3], 0, self.buttonWidth, wy), imageObject=self._addGlyphButtonImage, bordered=False, callback=self.mainWindow_newGlyphsCallback, sizeStyle="regular")
        self.w.file.addGlyphButton.enable(True)
        self.w.file.addGlyphButton.show(False)
        self.pane_doc.l = vanilla.List((0,30,0,0), [],
            columnDescriptions=sourceListColumnDescriptions,
            doubleClickCallback=self.sourcesList_selectFontCallback,
            selectionCallback=self.sourcesList_selectionChangedCallback,
            menuCallback = self.sourcesMenu_buildMenu,
            )
        self.pane_doc.l.getNSTableView().setUsesAlternatingRowBackgroundColors_(False)
        self.pane_doc.caption = vanilla.TextBox((5,5,-80,20), "Sources & Layers")
        #imagePath = "./resources/icon_16x16.png"
        self.pane_previewAxes.caption = vanilla.TextBox((5,5,150,20), "Preview")

        
        self.previewAxesList_doNotUpdate = False    # flag: set to prevent previewAxesList_handleEditsCallback from updating all the time
        self.pane_previewAxes.locationList = vanilla.List((0,30,0,0), [],
            columnDescriptions=previewAxesList_columnDescriptions,
            editCallback=self.previewAxesList_handleEditsCallback,
            allowsMultipleSelection=False,
            menuCallback = self.previewAxesMenu_buildMenu,
            )

        self.pane_previewAxes.locationList.getNSTableView().setUsesAlternatingRowBackgroundColors_(False)

        mathModelDescriptions = [
            dict(title="MutatorMath"),
            dict(title="VarLib"),
            ]
        segmentedLeftPos = 180 # left pos of the segmented buttons
        self.pane_previewAxes.mathModelPicker = vanilla.SegmentedButton((segmentedLeftPos, 3, -5, 22),
            segmentDescriptions=mathModelDescriptions, selectionStyle="one",
            callback = self.switchMathModelCallback)
        
        ### group 2
        self.pane_locs = vanilla.Group((0, 0, 0, 0))
        self.pane_locs.caption = vanilla.TextBox((5, 5, 150,20), "Locations")
        filterLocationDescriptions = [
            dict(title="Sources"),
            dict(title="Supports"),
            dict(title="Instances"),
            dict(title="Interesting"),
            dict(title="Open"),
            ]
        self.pane_locs.filterLocations = vanilla.SegmentedButton((segmentedLeftPos, 3, -5, 22),
            segmentDescriptions=filterLocationDescriptions, selectionStyle="any",
            callback = self.filterLocationsCallback)
        self.locationsList_setFilterLocationButton()
        self.locationsList_doNotUpdate = False
        self.pane_locs.locationsList = vanilla.List((0, 30, 0, 0), [],
            columnDescriptions=locationListDescriptions,
            selectionCallback=self.locationsList_selectLocationCallback,
            allowsMultipleSelection=True,
            menuCallback = self.locationsMenu_buildMenu,
            )
        self.pane_locs.locationsList.getNSTableView().setUsesAlternatingRowBackgroundColors_(False)
        
        h = 100
        paneDescriptors = [
            dict(   view=self.pane_doc,
                    identifier="pane_doc",
                    size=h * 2,
                    minSize=100,
                    canCollapse=False),

            dict(   view=self.pane_previewAxes,
                    identifier="pane_previewAxes",
                    size=h * 1,
                    minSize=100,
                    canCollapse=False),

            dict(   view=self.pane_locs,
                    identifier="pane_locs",
                    size=h * 2,
                    minSize=100,
                    canCollapse=False),

            ]
        self.w.panes = vanilla.SplitView((0, 40, -0, -30),
            paneDescriptors,
            isVertical=False,
            #dividerThickness=6,
            dividerStyle="thin",
            autosaveName = "com.letterror.skateboard.panes",
            )
        self.w.debugButton = vanilla.Button((30, -25, 60, 20), "Debugs", callback=self.debug_makeObjectBrowserCallback, sizeStyle="small")
        if edition == "develop":
            self.w.debugButton.show(True)
        else:
            self.w.debugButton.show(False)
        self.w.preferencesButton = vanilla.ImageButton((5, -25, 25, 20), imageObject=self._preferencesButtonImage, bordered=False, callback=self.mainWindow_showPreferencesWindow, sizeStyle="regular")

        self.w.versionText = vanilla.TextBox((-300, -20, 280, 20), "-", sizeStyle="small", alignment="right")
        self._remaining, edition = self._getEdition()
        if self._remaining:
            versionString = "this is a {} edition / v{} / {} days left".format(edition, skateboardBundle.version, self._remaining)
        else:
            versionString = "v {}".format(skateboardBundle.version)
        self.w.versionText.set(versionString)
        self.w.buyButton = vanilla.Button((30, -25, 60, 20), "Buy", self.callbackBuyButton, sizeStyle="small")
        if self._remaining is None:
            self.w.buyButton.show(False)
            
        self.w.controls = vanilla.Group((0,-30,0,30))
        self.w.bind("should close", self.mainWindow_shouldClose)
        self.w.bind("became main", self.mainWindow_becameMain)
        self.w.vanillaWrapper = weakref.ref(self)
        self.setUpBaseWindowBehavior()
    
    def callbackBuyButton(self, sender):
        #print('callbackBuyButton', sender)
        webbrowser.get().open(self.buyButtonURL)

    def mainWindow_showPreferencesWindow(self, sender):
        sbx.ui.preferencesWindow.SkateboardPreferences()
        
    # drop delegate
    def draw(self):
        AppKit.NSBeep()        
        #self.delegateDrawCalls.append(traceback.format_exc())
        # try:
        #     save()
        #     nsView = self.w.frontPage.canvas.getNSView()
        #     w = nsView.frame().size.width
        #     h = nsView.frame().size.height
        #     im = self._documentIconImage.size()
        #     c = (230/255, 248/255, 247/255)
        #     c = (100/255, 100/255, 247/255)
        #     fill(*c)
        #     rect(0,0,w,h)
        #     #image(self._documentIconImage, (.5*w-.5*im.width, .5*h-.5*im.height))
        #     restore()
        # except:
        #     print("xxx draw error", traceback.format_exc())
            
    def proposeDrop(self, pasteBoard, testing):
        files = pasteBoard.propertyListForType_(AppKit.NSFilenamesPboardType)
        spaces = []
        for name in files:
            if os.path.splitext(name)[-1] in ['.designspace', '.sp3']:
                spaces.append(name)
        if testing:
            if len(spaces) == 1:
                return True
            else:
                AppKit.NSBeep()
                return False
        else:
            if len(spaces) == 1:
                self.mainWindow_openDocumentWithPath(spaces[0])
        return True
    
    def pasteBoardTypes(self):
        return [AppKit.NSFilenamesPboardType]
    
    ## menucallbacks
    # sources list
    def sourcesMenu_buildMenu(self, sender):
        # called in self.pane_doc.l
        # all data for this list is set in self.document_checkFonts
        sel = sender.getSelection()
        items = []
        if not sel:
            # no selected items
            items.append(dict(title="Add Source…", callback=None))
            return items
        if len(sel) == 1:
            # one single selected items
            # let's find out what it is
            listItem = self.pane_doc.l[sel[0]]
            # which font is that?
            ufoStatus = listItem['_ufoStatus']
            isSupport = listItem['layer'] not in ['foreground', 'public.default']    # XXX Check - foreground might have a name
            ufoIsNeutral = listItem['_ufoIsNeutral']
            if ufoIsNeutral:
                items.append(dict(title="Default"))
            else:
                if isSupport:
                    items.append(dict(title="Source from layer"))
                else:
                    items.append(dict(title="Source from UFO"))
            #items.append(dict(title="Layer %s" % listItem['layer']))
            key = listItem['_sourceDescriptorKey']
            sd = self.getSourceDescriptorFromKey(key)
            items.append(dict(title=Location(sd.location).asString()))
            items.append("----")
            ufoName = os.path.basename(sd.path)
            supportGlyphs = None
            if ufoStatus == "available":
                items.append(dict(title="Open UFO", callback=self.sourcesMenu_openSelectedSources))
            if ufoStatus in ['current', 'open']:
                items.append(dict(title="Close UFO", callback=self.sourcesMenu_closeSelectedSource))
            if ufoStatus in ['current', 'available', 'open']:
                ok = True
                if ufoIsNeutral:
                    if not isSupport:
                        ok = False
                    else:
                        ok = True
                if ok:
                    mutingKey = sd.path, sd.layerName
                    if key in self.mutedFonts:
                        items.append(dict(title="Unmute", callback=self.sourcesMenu_unmuteSelectedSources))
                    else:
                        items.append(dict(title="Mute", callback=self.sourcesMenu_muteSelectedSources))
                    items.append(dict(title="Edit Location", callback=self.sourcesMenu_editSelectedSourceLocation))

            if not isSupport:
                # define a new support from an existing layer in this source
                items.append('----')
                availableSub = []
                for layerName, isFree in self.data_getAvailableLayers(sd):
                    if isFree:
                        availableSub.append(dict(title=layerName, callback=self.sourcesMenu_newSupportWithExistingLayerCallback))
                    else:
                        availableSub.append(dict(title=layerName))
                items.append(dict(title="New Support with Layer", items=availableSub))
            if ufoStatus == "missing":
                items.append(dict(title="Replace %s" % ufoName, callback=self.sourcesMenu_replaceSource))
            if isSupport:
                items.append('----')
                layerOK, supportGlyphs = self.data_findSupportGlyphs(sd)
                sub = []
                for glyphName in supportGlyphs:
                    sub.append(dict(title=glyphName, callback=self.sourcesMenu_openSupportGlyphCallback))
                if sub:
                    items.append(dict(title="Glyph Supports", items=sub))
                layerOK, glyphNames = self.data_findSupportGlyphs(sd)
                availableSub = []
                for layerName, isFree in self.data_getAvailableLayers(sd):
                    if isFree:
                        availableSub.append(dict(title=layerName, callback=self.sourcesMenu_reassignLayerCallback))
                    else:
                        availableSub.append(dict(title=layerName))
                availableSub.append("----")
                if not layerOK:
                    items.append(dict(title="Missing Layer %s" % listItem['layer']))
                items.append(dict(title="Reassign Support To Layer", items=availableSub))
                items.append(dict(title="New Layer for Support", callback=self.sourcesMenu_makeNewLayerCallback))
                items.append(dict(title="Delete Support", callback=self.sourcesMenu_deleteSupportLayerCallback))
            if ufoStatus in ['current', 'available', 'open']:
                items.append('----')
                items.append(dict(title="Show in Finder", callback=self.sourcesMenu_showSelectedSourceInFinder))
            return items
        else:
            selNames = []
            selStatus = []
            selNeutral = []
            selMuted = []
            for i in sel:
                listItem = self.pane_doc.l[i]
                key = self.pane_doc.l[i]['_sourceDescriptorKey']
                sd = self.getSourceDescriptorFromKey(key)
                ufoName = os.path.basename(sd.path)
                selNames.append(ufoName)
                ufoStatus = listItem['_ufoStatus']
                selStatus.append(ufoStatus)
                ufoIsNeutral = listItem['_ufoIsNeutral']
                selNeutral.append(ufoIsNeutral)
                if key in self.mutedFonts:
                    selMuted.append(True)
            if 'available' in selStatus:
                items.append(dict(title="Open", callback=self.sourcesMenu_openSelectedSources))
            if 'current' in selStatus or 'open' in selStatus:
                items.append(dict(title="Close", callback=self.sourcesMenu_closeSelectedSource))                
            items.append('----')
            if True in selMuted:
                items.append(dict(title="Unmute", callback=self.sourcesMenu_unmuteSelectedSources))
            items.append(dict(title="Mute", callback=self.sourcesMenu_muteSelectedSources))
        return items
    
    ## / edit locations
    def sourcesMenu_editSelectedSourceLocation(self, menuItem):
        # edit a location for the selection source
        for i in self.pane_doc.l.getSelection():
            key = self.pane_doc.l[i]['_sourceDescriptorKey']
            sourceDescriptor = self.getSourceDescriptorFromKey(key)
            data = dict(location=sourceDescriptor.location)
            s = sbx.ui.sheets.LocationEditorSheet(
                    data,
                    cancelCallback=self.sheet_cancel_genericCallback,
                    applyCallback=self.applySourceLocationEditCallback,
                    parentWindow=self.w,
                    axisValues = self.data_getAxisValues()
                    )
    
    def applySourceLocationEditCallback(self, data):
        for i in self.pane_doc.l.getSelection():
            key = self.pane_doc.l[i]['_sourceDescriptorKey']
            sourceDescriptor = self.getSourceDescriptorFromKey(key)
            sourceDescriptor.location.update(data['newLocation'])
            # master with new location
            # rebuild everything.
            self.document_checkFonts()
            self.spaceWindow_update()
            self.timerSchedule_navigatorTool_mouseUp_draggingDone()
            self.document_setNeedSave(True)
        
    def locationsMenu_editLocationCallback(self, menuItem):
        for i in self.pane_locs.locationsList.getSelection():
            item = self.pane_locs.locationsList[i]
            data = dict(location=item['location'])
            s = sbx.ui.sheets.LocationEditorSheet(
                    data,
                    cancelCallback=self.sheet_cancel_genericCallback,
                    applyCallback=self.applyinstanceLocationEditCallback,
                    parentWindow=self.w,
                    axisValues = self.data_getAxisValues()
                    )

    def applyinstanceLocationEditCallback(self, data):
        needsUpdate = False
        for i in self.pane_locs.locationsList.getSelection():
            item = self.pane_locs.locationsList[i]
            iD = item['obj']
            new = data['newLocation']
            for k, v in new.items():
                # check per axis if there is a new value
                if iD.location.get(k) != v:
                    needsUpdate = True
                    iD.location[k] = v
        if needsUpdate:
            self.document_updateCurrentDesignSpace()
            self.document_setNeedSave(True)
    
    def sourcesMenu_replaceSource(self, menuItem=None):
        # locate a new master for a missing source
        if menuItem is not None:
            print("sourcesMenu_replaceSource", menuItem)
        resultPath = GetFile(message="Replace UFO",
            title="Open", directory=None, fileName=None,
            allowsMultipleSelection=False, fileTypes=["ufo"])
        if resultPath is not None:
            needsUpdate = False
            for i in self.pane_doc.l.getSelection():
                key = self.pane_doc.l[i]['_sourceDescriptorKey']
                sourceDescriptor = self.getSourceDescriptorFromKey(key)
                sourceDescriptor.path = resultPath
                needsUpdate = True
                break
            if needsUpdate:
                self.ds.loadFonts(reload=True)
                self.document_checkFonts()
                self.document_updateCurrentDesignSpace()
                self.document_setNeedSave(True)
                
        
        
    def sourcesMenu_openSupportGlyphCallback(self, menuItem):
        # open a glyphwindow for the selected ufo + layer + glyphname
        # callback for sourcesMenu_buildMenu
        supportGlyphName = menuItem.title()
        for i in self.pane_doc.l.getSelection():
            key = self.pane_doc.l[i]['_sourceDescriptorKey']
            sourceDescriptor = self.getSourceDescriptorFromKey(key)
            f = OpenFont(sourceDescriptor.path, showInterface=True)
            w = CurrentFontWindow().window()
            w.show() # ? bizar
            OpenGlyphWindow(f[supportGlyphName])
            SetCurrentLayerByName(sourceDescriptor.layerName)
    
    def sourcesMenu_makeNewLayerCallback(self, something =None):
        # menu callback for making a new layer in the target UFO
        # this means a new sheet, yay
        newLayerName = something.title()
        for i in self.pane_doc.l.getSelection():
            key = self.pane_doc.l[i]['_sourceDescriptorKey']
            sourceDescriptor = self.getSourceDescriptorFromKey(key)
            targetUFO = self.ds.fonts.get(sourceDescriptor.name)
            data = dict(proposedLayerName="My-new-layer", existingLayerNames=targetUFO.layerOrder)
            s = sbx.ui.sheets.NewLayerNameEditSheet(
                    data,
                    cancelCallback=self.sheet_cancel_genericCallback,
                    applyCallback=self.sourcesMenu_sheet_applyNewLayerNameCallback,
                    parentWindow=self.w)

    def sourcesMenu_sheet_applyNewLayerNameCallback(self, something=None):
        newLayerName = something['newLayerName']
        for i in self.pane_doc.l.getSelection():
            key = self.pane_doc.l[i]['_sourceDescriptorKey']
            sourceDescriptor = self.getSourceDescriptorFromKey(key)
            targetUFO = self.ds.fonts.get(sourceDescriptor.name)
            targetUFO.newLayer(newLayerName)
            # validate the layer was added
            assert newLayerName in targetUFO.layerOrder
            sourceDescriptor.layerName = newLayerName
            self.document_checkFonts()        

    def sourcesMenu_newSupportWithExistingLayerCallback(self, something=None):
        # menu callback for reassign missing layer to an existing layer
        newLayerName = something.title()
        for i in self.pane_doc.l.getSelection():
            key = self.pane_doc.l[i]['_sourceDescriptorKey']
            sourceDescriptor = self.getSourceDescriptorFromKey(key)
            new = SourceDescriptor()
            new.path = sourceDescriptor.path
            new.location = Location(self.instanceLocation)
            new.layerName = newLayerName
            new.name = "skateboard_support_with_%s" % newLayerName
            new.copyLib = sourceDescriptor.copyLib
            new.copyInfo = sourceDescriptor.copyInfo
            new.copyGroups = sourceDescriptor.copyGroups
            new.copyFeatures = sourceDescriptor.copyFeatures
            new.muteKerning = sourceDescriptor.muteKerning
            new.muteInfo = sourceDescriptor.muteInfo
            new.mutedGlyphNames = sourceDescriptor.mutedGlyphNames
            new.familyName = sourceDescriptor.familyName
            new.styleName = sourceDescriptor.styleName
            self.ds.addSource(new)
            # rebuild sources list
            self.ds.loadFonts(reload=True)
            # why is the font loaded? AAA
            self.document_checkFonts()
            self.document_setNeedSave(True)

    def sourcesMenu_reassignLayerCallback(self, something=None):
        # menu callback for reassign missing layer to an existing layer
        newLayerName = something.title()
        result = mojo_AskYesNoCancel("Reassign this Support to layer %s?" % newLayerName, title="Skateboard", informativeText="")
        if result != 1:
            return
        for i in self.pane_doc.l.getSelection():
            key = self.pane_doc.l[i]['_sourceDescriptorKey']
            sourceDescriptor = self.getSourceDescriptorFromKey(key)
            sourceDescriptor.layerName = newLayerName
            # rebuild sources list
            self.document_checkFonts()
            self.document_setNeedSave(True)
    
    def sourcesMenu_deleteSupportLayerCallback(self, something=None):
        deleteSupportWithLayerName = something.title()
        result = mojo_AskYesNoCancel("Delete Support %s?" % deleteSupportWithLayerName, title="Skateboard", informativeText="This only removes the support from the designspace. The layers remain in the UFO.")
        if result != 1:
            return
        for i in self.pane_doc.l.getSelection():
            key = self.pane_doc.l[i]['_sourceDescriptorKey']
            sourceDescriptor = self.getSourceDescriptorFromKey(key)
            if sourceDescriptor in self.ds.sources:
                self.ds.sources.remove(sourceDescriptor)
            # rebuild sources list
            self.document_checkFonts()                
            self.document_setNeedSave(True)
    
    def data_updateDocumentVerticalMetrics(self, font=None):
        # update the vertical metrics with data from font
        if font is None:
            #print("data_updateDocumentVerticalMetrics resetting")
            self.documentUnitsPerEm = None
            self.documentAscender = None
            self.documentDescender = None
        else:
            if font.info.unitsPerEm>0:
                if self.documentUnitsPerEm is None:
                    self.documentUnitsPerEm = font.info.unitsPerEm
                else:
                    self.documentUnitsPerEm = max(font.info.unitsPerEm, self.documentUnitsPerEm)
            if font.info.ascender>0:
                if self.documentAscender is None:
                    self.documentAscender = font.info.ascender
                else:
                    self.documentAscender = max(font.info.ascender, self.documentAscender)
            if font.info.descender<0:
                if self.documentDescender is None:
                    self.documentDescender = font.info.descender
                else:
                    self.documentDescender = max(font.info.descender, self.documentDescender)
        #print("data_updateDocumentVerticalMetrics documentUnitsPerEm", self.documentUnitsPerEm)
        #print("data_updateDocumentVerticalMetrics documentDescender", self.documentDescender)
        #print("data_updateDocumentVerticalMetrics documentAscender", self.documentAscender)

    def data_getPreviewFont(self, location):
        if location is None:
            locationKey = "currentPreview"
        else:
            locationKey = location.asTuple()
        if locationKey in self.previewFonts:
            return self.previewFonts.get(locationKey)
        f = RFont(showInterface=False)
        f.info.familyName = "Skateboard Preview"
        f.info.styleName = ""
        f.info.unitsPerEm = self.documentUnitsPerEm
        f.info.descender = self.documentDescender
        f.info.ascender = self.documentAscender
        self.previewFonts[locationKey] = f
        return self.previewFonts.get(locationKey)

    def data_getAvailableLayers(self, sourceDescriptor):
        # find which layers are available in this font
        # ignore the default layer
        # ignore layers that are already in use
        usedLayers = []
        for sd in self.ds.sources:
            if sd.layerName:
                usedLayers.append(sd.layerName)
        s = []
        if sourceDescriptor.name in self.ds.fonts:
            thisFont = self.ds.fonts[sourceDescriptor.name]
            defaultLayerName = thisFont.defaultLayer.name
            available = thisFont.layerOrder
            # these are the layers we have already
            # filter out the ones that are in use
            for name in thisFont.layerOrder:
                if name in usedLayers or name == defaultLayerName:
                    s.append((name, False))
                else:
                    s.append((name, True))
        return s
        
    def data_findSupportGlyphs(self, sourceDescriptor):
        # find which glyphs are available in this sourceDescriptor

        thisFont = self.ds.fonts[sourceDescriptor.name]
        if not sourceDescriptor.layerName in thisFont.layerOrder:
            # add to muted fonts
            layerOK = False
            return False, []
        thisLayer  = thisFont.getLayer(sourceDescriptor.layerName)
        return True, thisLayer.keys()
    
    def muteSource(self, ufoPath, state):
        # mute this one
        hasChanged = False
        sources = []
        for sd in self.ds.sources:
            if sd.path == ufoPath:
                key = sd.path, sd.layerName
                if state:
                    # do mute
                    self.mutedFonts[key] = True
                    hasChanged = True
                else:
                    if key in self.mutedFonts:
                        del self.mutedFonts[key]
                        hasChanged = True
                if hasChanged:
                    sources.append(sd)
        if hasChanged:
            self.document_checkFonts()
            self.storage_mainPreviews.reset()
            self.storage_supportPreviews.reset()
            self.data_updateAllVisibleGlyphs()
            self.document_setNeedSave(True)

            self.spaceWindow_update()
            self.timerSchedule_navigatorTool_mouseUp_draggingDone()
        return sources
                    
        
    def sourcesMenu_unmuteSelectedSources(self, sender):
        # contextual menu callback for the sources list
        # unmute this master
        hasUnmuted = False
        for i in self.pane_doc.l.getSelection():
            sourceDescriptorKey = self.pane_doc.l[i]['_sourceDescriptorKey']
            if sourceDescriptorKey in self.mutedFonts:
                del self.mutedFonts[sourceDescriptorKey]
                hasUnmuted = True
        if hasUnmuted:
            self.document_checkFonts()
            self.storage_mainPreviews.reset()
            self.storage_supportPreviews.reset()
            self.data_updateAllVisibleGlyphs()
            self.document_setNeedSave(True)
            #for g in self.currentGlyphViews.values():
            #    self.data_prepareGlyph(g.name, force=True)
        self.spaceWindow_update()
        self.timerSchedule_navigatorTool_mouseUp_draggingDone()
    
    def sourcesMenu_muteSelectedSources(self, sender):
        # contextual menu callback for the sources list
        # mute this master
        # note: this affects all mutators and all glyphs
        hasMuted = False
        for i in self.pane_doc.l.getSelection():
            sourceDescriptorKey = self.pane_doc.l[i]['_sourceDescriptorKey']
            if not sourceDescriptorKey in self.mutedFonts:
                self.mutedFonts[sourceDescriptorKey] = True
                hasMuted = True
        if hasMuted:
            self.document_checkFonts()
            self.storage_mainPreviews.reset()
            self.storage_supportPreviews.reset()
            self.data_updateAllVisibleGlyphs()
            self.document_setNeedSave(True)
        self.spaceWindow_update()
        self.timerSchedule_navigatorTool_mouseUp_draggingDone()
    
    def data_updateAllVisibleGlyphs(self):
        # prepare new mutators for all the glyphs on display
        # this means all open windows, but also the glyphs in the text window
        for g in self.currentGlyphViews.values():
            if g is not None:
                self.data_prepareGlyph(g.name, force=True)
        
    def sourcesMenu_openSelectedSources(self, sender):
        # contextual menu callback for the sources list
        self.sourcesList_selectFontCallback(None)
    
    def sourcesMenu_closeSelectedSource(self, sender):
        # contextual menu callback for the sources list
        for i in self.pane_doc.l.getSelection():
            key = self.fontPaths[i]
            if key[0] in self.openFonts:
                font = self.openFonts[key[0]]
                if not font.hasChanged():
                    font.close()
                else:
                    font.save()
                    font.close()
                    
    def previewAxesMenu_buildMenu(self, sender):
        # make a menu for self.pane_previewAxes.locationList
        # the data is set in self.data_instanceLocationChanged
        # the location menu
        items = []
        sel = self.pane_previewAxes.locationList.getSelection()
        for i in sel:
            record = self.pane_previewAxes.locationList[i]
            axisName = record['axisKey']
            aMin = record['minimumKey']
            aMax = record['maximumKey']
            aDef = record['defaultKey']
            if aMin == aDef or aMax == aDef:
                # only min / max
                items.append(dict(title="Set %s to %s" % (axisName, aMin), callback=self.previewAxesMenu_setAxisValueToMinimum))
                items.append(dict(title="Set %s to %s" % (axisName, aMax), callback=self.previewAxesMenu_setAxisValueToMaximum))
            else:
                items.append(dict(title="Set %s to %s" % (axisName, aMin), callback=self.previewAxesMenu_setAxisValueToMinimum))
                items.append(dict(title="Set %s to %s" % (axisName, aDef), callback=self.previewAxesMenu_setAxisValueToDefault))
                items.append(dict(title="Set %s to %s" % (axisName, aMax), callback=self.previewAxesMenu_setAxisValueToMaximum))
            if not axisName in self.mutedAxes:
                items.append('----')
                items.append(dict(title="Mute %s axis" % (axisName), callback=self.previewAxesMenu_muteSelectedAxis))
            else:
                items.append('----')
                items.append(dict(title="Unmute %s axis" % (axisName), callback=self.previewAxesMenu_unmuteSelectedAxis))
        return items
    
    def previewAxesMenu_setAxisValueToMinimum(self, sender):
        # mapping
        sel = self.pane_previewAxes.locationList.getSelection()
        for i in sel:
            record = self.pane_previewAxes.locationList[i]
            axisName = record['axisKey']
            aMin = float(record['minimumKey'])
            self.instanceLocation[axisName] = aMin
            self.data_instanceLocationChanged(dragging=False)

    def previewAxesMenu_setAxisValueToMaximum(self, sender):
        # mapping
        sel = self.pane_previewAxes.locationList.getSelection()
        for i in sel:
            record = self.pane_previewAxes.locationList[i]
            axisName = record['axisKey']
            aMax = float(record['maximumKey'])
            self.instanceLocation[axisName] = aMax
            self.data_instanceLocationChanged(dragging=False)

    def previewAxesMenu_setAxisValueToDefault(self, sender):
        # mapping
        sel = self.pane_previewAxes.locationList.getSelection()
        for i in sel:
            record = self.pane_previewAxes.locationList[i]
            axisName = record['axisKey']
            aDef = float(record['defaultKey'])
            self.instanceLocation[axisName] = aDef
            self.data_instanceLocationChanged(dragging=False)
    
    def previewAxesMenu_muteSelectedAxis(self, sender):
        sel = self.pane_previewAxes.locationList.getSelection()
        thingsMuted = False
        for i in sel:
            record = self.pane_previewAxes.locationList[i]
            axisName = record['axisKey']
            self.mutedAxes[axisName] = True
            self.data_instanceLocationChanged(dragging=False)
            self.document_setNeedSave(True)
            thingsMuted = True
        # we really need to prepare all of them
        if thingsMuted:
            self.storage_mainPreviews.reset()
            self.storage_supportPreviews.reset()
            for g in self.currentGlyphViews.values():
                self.data_prepareGlyph(g.name, force=True)
            self.document_setNeedSave(True)

    def previewAxesMenu_unmuteSelectedAxis(self, sender):
        sel = self.pane_previewAxes.locationList.getSelection()
        thingsUnmuted = False
        for i in sel:
            record = self.pane_previewAxes.locationList[i]
            axisName = record['axisKey']
            if axisName in self.mutedAxes:
                del self.mutedAxes[axisName]
                self.data_instanceLocationChanged(dragging=False)
                thingsUnmuted = True
        if thingsUnmuted:
            self.storage_mainPreviews.reset()
            self.storage_supportPreviews.reset()
            for g in self.currentGlyphViews.values():
                self.data_prepareGlyph(g.name, force=True)
            self.document_setNeedSave(True)

    def locationsMenu_buildMenu(self, sender):
        # build menus for self.pane_locs.locationsList
        items = []
        sel = self.pane_locs.locationsList.getSelection()
        if len(sel) > 1:
            # multiple things
            # what do we want to offer?
            # generate UFO for instances
            instances = []
            for i in sel:
                item = self.pane_locs.locationsList[i]
                if item['flavor'] == "Instance":
                    instances.append(item)
            if instances:
                items.append(dict(title="Generate Selected Instances", callback=self.locationsMenu_generateSelectedInstances))
        elif len(sel) == 1:
            i = sel[0]
            item = self.pane_locs.locationsList[i]
            if item['flavor'] == "Instance":
                items.append(dict(title="Instance"))
                items.append(dict(title=Location(item['location']).asString()))
                items.append('----')
                items.append(dict(title="Edit Names"))
                items.append(dict(title="Edit Location", callback=self.locationsMenu_editLocationCallback))
                items.append(dict(title="Generate Instance", callback=self.locationsMenu_generateSelectedInstances))
                items.append(dict(title="Delete"))
            elif item['flavor'] == "Support":
                items.append(dict(title="Support"))
                items.append(dict(title=Location(item['location']).asString()))
            elif item['flavor'] == "Source":
                items.append(dict(title="Source"))
                items.append(dict(title=Location(item['location']).asString()))
                items.append('----')
                items.append(dict(title="Edit Location"))
            elif item['flavor'] == "Open":
                items.append(dict(title="Open Location"))
                items.append('----')
                items.append(dict(title="Add as Instance", callback=self.document_newInstanceCallback))
                items.append(dict(title="Insert Master"))
            elif item['flavor'] == "Interesting":
                items.append(dict(title="Interesting"))
                items.append(dict(title=Location(item['location']).asString()))
                items.append('----')
                items.append(dict(title="Edit Note", callback=self.locationsMenu_menuEditTitleCallback))
                items.append(dict(title="Edit Location", callback=self.locationsMenu_editInterestingCallback))
                items.append(dict(title="Add as Instance", callback=self.document_newInstanceCallback))
                items.append(dict(title="Delete", callback=self.locationsMenu_deleteSelectedInterestingLocationCallback))
            else:
                items.append(dict(title="Nothing!?"))
        return items
        
    def locationsMenu_generateSelectedInstances(self, sender):
        # vanilla callback for generating UFOs for the selected instances
        bar = None
        try:
            bar = ProgressBar()
            bar.label("Generating selected instances…")
            bar.tick()
            sel = self.pane_locs.locationsList.getSelection()
            instances = []
            if sel:
                self.ds.loadFonts(reload=True)

            for i in sel:
                item = self.pane_locs.locationsList[i]
                if item['flavor'] == "Instance":
                    instances.append(item['obj'])
                
                
            tempPath, generatingDesignSpace = self.data_prepareNewDesignSpace(
                igoreMuting = False,
                ignoreAnisotropic = True,
                save=True,
                clipLocations=False, 
                instances = instances,
                tagName="UFOExport",
                remapAxes=True,
                )
            # set the mathmodel here?
            if self.previewMathPref == "previewMutatorMath":
                bar.label("Generating instance with MutatorMath…")
                ufoNameMathTag = "MM"
                generatingDesignSpace.useVarlib = False
            else:
                bar.label("Generating instance with varlib…")
                ufoNameMathTag = "VL"
                generatingDesignSpace.useVarlib = True
            generatingDesignSpace.findDefault()
            generatingDesignSpace.loadFonts(reload=True)
            generatingDesignSpace._glyphMutators = {}
            neutral = generatingDesignSpace.getNeutralFont()

            for iD in instances:
                exportFont = generatingDesignSpace.makeInstance(iD, doRules=True, bend=True)
                if exportFont is not None:
                    if neutral is not None:
                        exportFont.lib.update(neutral.lib)
                    ufoName = "%s-%s_%s.ufo" %(iD.familyName, iD.styleName, ufoNameMathTag)
                    exportFolder = os.path.dirname(os.path.abspath(iD.path))
                    path = iD.path
                    if not os.path.exists(exportFolder):
                        os.makedirs(exportFolder)

                    exportPath = os.path.join(exportFolder, ufoName)
                    # make sure the preview is not open
                    for f in AllFonts():
                        if f.path == exportPath:
                            # close it
                            f.close()
                    exportFont.save(exportPath)
                    openInstancePref = getDefault(sbx.ui.preferencesWindow.openInstancesPrefKey)
                    if openInstancePref:
                        new = OpenFont(exportPath, showInterface=True)

        finally:
            if bar is not None:
                bar.close()
            if os.path.exists(tempPath):
                os.remove(tempPath)
        
    def locationsMenu_deleteSelectedInterestingLocationCallback(self, sender):
        # vanilla callback for the location editing segment button
        selectedLocation = self.pane_locs.locationsList.getSelection()
        selectedLocation.sort()
        selectedLocation.reverse()
        for s in selectedLocation:
            loc = self.seriesLocations[s]['location']
            remove = []
            for i, items in enumerate(self.pointsOfInterest):
                if items[0] == loc:
                    remove.append(i)
            for i in remove:
                del self.pointsOfInterest[i]
        self.locationsList_updateItemsCallback()
        self.document_setNeedSave(True)
    
    ## / sheets
    ## / edit the name of an interesting location
    def locationsMenu_menuEditTitleCallback(self, sender=None):
        # edit the title of the selected interesting location
        sel = self.pane_locs.locationsList.getSelection()
        try:
            for i in sel:
                item = self.pane_locs.locationsList[i]
                # open edit sheet here
                loc = Location(item['location']).asTuple()
                data = dict(currentLocationName=item['description'], location=loc)
                s = sbx.ui.sheets.LocationNameEditSheet(
                        data,
                        cancelCallback=self.sheet_cancel_genericCallback,
                        applyCallback=self.locationsMenu_sheet_apply_editLocationNameCallback,
                        parentWindow=self.w)
        except:
            if self.verbose:
                print("Skateboard: Error locationsMenu_menuEditTitleCallback", traceback.format_exc())
    
    def sheet_cancel_genericCallback(self, sender):
        pass

    def locationsMenu_sheet_apply_editLocationNameCallback(self, data):
        # callback for menu for editing the note / title / name of an interesting location
        sel = self.pane_locs.locationsList.getSelection()
        if sel:
            i = sel[0]
        else:
            return
        try:
            pointsOfInterest = self.pointsOfInterest[:]
            editedPointsOfInterest = []
            currentName = data['currentLocationName']
            newName = data['newLocationName']
            currentKey = (data['location'], currentName)
            if newName == currentName:
                return
            for location, name in pointsOfInterest:
                location = Location(location)
                thisKey = (location.asTuple(), name)
                if thisKey == currentKey:
                    editedPointsOfInterest.append((location, newName))
                else:
                    editedPointsOfInterest.append((location, name))
            self.pointsOfInterest = editedPointsOfInterest[:]
            # update the locations
            self.locationsList_updateItemsCallback()
            self.document_setNeedSave(True)
        except:
            if self.verbose:
                print("Skateboard: Error locationsMenu_sheet_apply_editLocationNameCallback", traceback.format_exc())

    ## / edit location of an interesting location        
    def locationsMenu_editInterestingCallback(self, sender):
        #LocationEditorSheet
        sel = self.pane_locs.locationsList.getSelection()
        for i in sel:
            item = self.pane_locs.locationsList[i]
            # open edit sheet here
            s = sbx.ui.sheets.LocationEditorSheet(
                    item,
                    cancelCallback=self.sheet_cancel_genericCallback,
                    applyCallback=self.locationsMenu_sheet_apply_editLocationCallback,
                    parentWindow=self.w,
                    axisValues = self.data_getAxisValues()
                    )

    
    def locationsMenu_sheet_apply_editLocationCallback(self, data):
        # callback for menu for editing the location of an interesting location
        sel = self.pane_locs.locationsList.getSelection()
        if sel:
            i = sel[0]
        else:
            return
        pointsOfInterest = self.pointsOfInterest[:]
        if data['newLocation'] == data['location']:
            return
        currentKey = (data['location'].asTuple(), data['description'])
        editedPointsOfInterest = []
        for i in sel:
            for location, name in pointsOfInterest:
                thisKey = (location.asTuple(), name)
                if thisKey == currentKey:
                    editedPointsOfInterest.append((Location(data['newLocation']), name))
                else:
                    editedPointsOfInterest.append((location, name))
            pointsOfInterest = editedPointsOfInterest[:]
        self.pointsOfInterest = pointsOfInterest
        # update the locations
        self.locationsList_updateItemsCallback()
        self.document_setNeedSave(True)

    ## / menu    
    def data_logWorkingGlyph(self, glyphName):
        # log that this glyph data_checkGlyphWorks
        self._working[glyphName] = True
    
    def data_publishFailingGlyphEvent(self, glyphName):
        # publish an event for this glyph
        publishEvent("skateboard.problems.glyph",
            location = self.instanceLocation, 
            designspace = self.ds, 
            interactionSources = self.interactionSources, 
            axes = self.axisExtremes,
            dragging=False,
            instanceGlyph = None,   #self.instanceGlyphObject,
            previewGlyph = None,    #self.contouredInstanceGlyphObject,
            glyphName = glyphName
            )
        
    def data_logFailingGlyph(self, glyphName):
        # log that this glyph does not work
        if self._working.get(glyphName) != False:
            # it was working and now it is not?
            self.data_publishFailingGlyphEvent(glyphName)
        self._working[glyphName] = False
        
    def data_checkGlyphWorks(self, glyphName):
        # get the status for this glyph
        return self._working.get(glyphName, None)
        
    # def keyForSource(self, sourceDescriptor):
    #     # Sometimes we need a unique key for the sourcedescriptor
    #     # Can't rely on just the UFO path, layerName or location
    #     layerName = sourceDescriptor.layerName
    #     if not layerName:
    #         layerName = "foreground"
    #     loc = Location(sourceDescriptor.location)
    #     key = "%s_%s_%s" % (os.path.basename(sourceDescriptor.path), layerName, loc.asString())
    #     return key

    def switchMathModelCallback(self, sender):
        # vanilla callback for segmented button switching the math model
        wouldLikeToUseVarLib = sender.get() == 1
        self.previewPane_handleMathModelChange(wouldLikeToUseVarLib)

    def previewPane_handleMathModelChange(self, wouldLikeToUseVarLib=None):
        # handle all the updating for a mathmodel change
        if wouldLikeToUseVarLib is None:
            wouldLikeToUseVarLib = 'previewVarLib'
        if wouldLikeToUseVarLib:
            # we want to use varlib
            # so maybe we have to clip the location
            if self.previewMathPref in ['previewVarLib']:
                clipping, self.instanceLocation = self.data_clipLocation(self.instanceLocation)
            self.previewMathPref = 'previewVarLib'
            self.ds.useVarlib = True
            self.canExtrapolate = False
        else:
            self.previewMathPref = "previewMutatorMath"
            self.ds.useVarlib = False
            self.canExtrapolate = True
        self.storage_mainPreviews.reset()
        self.storage_supportPreviews.reset()
        self.lastSupports = []
        self.glyphMutators = {}
        self.document_setNeedSave(True)
        self.document_checkFonts()                
        if self.currentGlyphName is not None:
            self.data_prepareGlyph(self.currentGlyphName, force=True)
        UpdateCurrentGlyphView()
        for window in AllGlyphWindows():
            window.editGlyphView.refresh()
        if self.textWindow:
            self.timerSchedule_textWindowRedraw(now=True)
        
    def previewPane_setMathPickerButton(self):
        # set the segmented button for the math model to the right value
        if self.previewMathPref == 'previewVarLib':
            self.pane_previewAxes.mathModelPicker.set(1)
            self.ds.useVarlib = True
        else:
            self.pane_previewAxes.mathModelPicker.set(0)
            self.ds.useVarlib = False

    def locationsList_setFilterLocationButton(self, reset=False):
        locationListButtonsInit = []
        if not reset:
            if self._listSources: 
                locationListButtonsInit.append(0)
            if self._listSupports: 
                locationListButtonsInit.append(1)
            if self._listInstances: 
                locationListButtonsInit.append(2)
            if self._listInteresting: 
                locationListButtonsInit.append(3)
            if self._listOpen: 
                locationListButtonsInit.append(4)
        self.pane_locs.filterLocations.set(locationListButtonsInit)
    
    def sourcesMenu_deleteSelectedSources(self, sender):
        # contextual menu callback for the sources list
        pass
    
    def sourcesMenu_showSelectedSourceInFinder(self, sender):
        s = self.pane_doc.l.getSelection()
        if not s: return
        s = s[0]
        path = self.fontPaths[s][0]
        if path:
            sbx.tools.showInFinder(path)
    
    # def glyphCellDrawCallback(self,  notification):
    #     # callback for glyphCellDraw notification
    #     rpos, rsize = notification['rect']
    #     small = False
    #     if rsize[1] < 30:
    #         small = True
    #     g = notification['glyph']
    #     temps = g.font.templateKeys()
    #     state = None
    #     if g is not None:
    #         glyphRelevant = g.font.path in self.openFonts
    #         if not glyphRelevant:
    #             return
    #         if g.name in temps:
    #             return
    #         state = self.data_checkGlyphWorks(g.name)
    #         if state is None:
    #             return
    #         save()
    #         scale(1.85)
    #         if state is False:
    #             # tested, does not work
    #             fill(.8,.36,.36,.8)
    #             oval(1,1,2,2)
    #             if not small:
    #                 oval(4,1,2,2)
    #                 oval(1,4,2,2)
    #                 oval(4,4,2,2)
    #                 rect(2.2,2.2,2.6,2.6)
    #         restore()
        
        
    def filterLocationsCallback(self, sender):
        # vanilla callback for the locations segmented filter button
        selection = sender.get()
        self._listSources = 0 in selection
        self._listSupports = 1 in selection
        self._listInstances = 2 in selection
        self._listInteresting = 3 in selection
        self._listOpen = 4 in selection
        self.locationsList_updateItemsCallback()
        if self.spaceWindow:
            self.spaceWindow_prepareData()
        
    def textWindowRequestsTheseGlyphsCallback(self, sender=None):
        # textwindow callback
        # update the textwindow when there is new input
        self.textWindow_updateText()
        self.timerSchedule_textWindowRedraw(now=True)

    def textWindow_selectGlyph(self, sender):
        # textWindow callback for selecting a glyph
        selectedGlyph = sender.getSelectedGlyph()
        
    def mainWindow_openTextWindow(self, sender=None):
        # textwindow callback
        if self.textWindow is None:
            self.textWindow = sbx.ui.textWindow.SkateboardTextWindow(self.mainWindow_closingTextWindowCallback,
                requestGlyphsCallback=self.textWindowRequestsTheseGlyphsCallback,
                selectGlyphCallback=self.textWindow_selectGlyph,
                previewText=self.previewTextRaw)
        else:
            self.textWindow.w.getNSWindow().makeKeyAndOrderFront_(self)
        self.textWindow_updateText()
        self.textWindow_updateGlyphs()

    def mainWindow_openSpaceWindow(self, sender=None):
        # space window callback
        if self.spaceWindow is None:
            self.spaceWindow = sbx.ui.spaceWindow.SkateboardSpaceWindow(self.mainWindow_closingSpaceWindowCallback)    #, requestSamplesCallback=self.skateboardProcessSpaceWindowSamples)
        else:
            self.spaceWindow.w.getNSWindow().makeKeyAndOrderFront_(self)

        #print("mainWindow_openSpaceWindow self.previewMathPref", self.previewMathPref)
        if self.previewMathPref == "previewMutatorMath":
            self.spaceWindow.canExtrapolate = True
        else:
            self.spaceWindow.canExtrapolate = False

        self.spaceWindow.darkMode = self._colorScheme
        # init the spacewindow
        self.publishEvent_interactionSourcesChanged()
        self.spaceWindow_update()
   
    def mainWindow_show(self, sender=None):
        # space window callback, bring skateboard window to front
        self.w.getNSWindow().makeKeyAndOrderFront_(self)
        
    def spaceWindow_update(self):
        if self.spaceWindow:
            self.spaceWindow_prepareData()
        # publishEvent("skateboardPreviewLocationChanged", 
        #     location = self.instanceLocation, 
        #     designspace = self.ds, 
        #     interactionSources = self.interactionSources, 
        #     axes = self.axisExtremes,
        #     instanceGlyph = self.instanceGlyphObject,
        #     previewGlyph = self.contouredInstanceGlyphObject,
        #     )

    def mainWindow_closingTextWindowCallback(self, sender=None):
        # callback from the text window to the main window
        # the text window is closing, so take note.
        self.textWindow = None

    def mainWindow_closingSpaceWindowCallback(self, sender=None):
        # callback from the psace window to the main window
        # the space window is closing, so take note.
        self.spaceWindow = None

    ## timers
    def timerSchedule_textWindowRedraw(self, now=False, debugTag=None):
        # timer for redraws of the text window
        # should be hesitant
        # calls timerDone_textWindow_redrawCallback_
        if self.textRedrawTimer is not None:
            self.textRedrawTimer.invalidate()
            self.textRedrawTimer = None
        if now:
            interval = self.textRedrawInterval
        else:
            interval = .01
        self.textRedrawTimer = AppKit.NSTimer.scheduledTimerWithTimeInterval_target_selector_userInfo_repeats_(
            interval,
            self,
            "timerDone_textWindowRedraw:",
            debugTag, False)
    
    def timerDone_textWindowRedraw_(self, timer):
        # called by timerSchedule_textWindowRedraw
        debugTag = timer.userInfo()
        if debugTag is not None:
            print("timerDone_textWindowRedraw_", debugTag)
        if self.textRedrawTimer is not None:
            self.textRedrawTimer.invalidate()
            self.textRedrawTimer = None
        # redraw text window
        self.textWindow_updateGlyphs()
        if self.spaceWindow:
            self.spaceWindow_update()
    
    def timerSchedule_navigatorTool_mouseUp_draggingDone(self, debugTag=None):
        # timer for redraws after mouseup / dragging stops.
        # calls timerDone_navigatorTool_mouseUpDraggingDoneCallback_
        if self.mouseUpTimer is not None:
            self.mouseUpTimer.invalidate()
            self.mouseUpTimer = None
        self.mouseUpTimer = AppKit.NSTimer.scheduledTimerWithTimeInterval_target_selector_userInfo_repeats_(
            self.mouseUpDraggingDoneTimerInterval,
            self,
            "timerDone_navigatorTool_mouseUp_draggingDone:",
            debugTag, False)

    def timerDone_navigatorTool_mouseUp_draggingDone_(self, timer):
        debugTag = timer.userInfo()
        if debugTag is not None:
            print("timerDone_navigatorTool_mouseUp_draggingDone_", debugTag)
        # redraw after a mouseup. 
        if self.mouseUpTimer is not None:
            self.mouseUpTimer.invalidate()
            self.mouseUpTimer = None
        self.drawDetails = True
        UpdateCurrentGlyphView()
        if self.textWindow:
            # one more refresh of the textwindow after dragging
            self.timerSchedule_textWindowRedraw()
        for window in AllGlyphWindows():
            window.editGlyphView.refresh()
        self.publishEvent_previewLocationChanged_dragFinished()
    
    def timerSchedule_slowAnimationTimer(self, debugTag=None):
        # timer for slow rotations through an axis
        # calls ??
        if self.slowAnimationTimer:
            self.slowAnimationTimer.invalidate()
            self.slowAnimationTimer = None
        self.slowAnimationTimer = AppKit.NSTimer.scheduledTimerWithTimeInterval_target_selector_userInfo_repeats_(
            self.slowAnimationInterval,
            self,
            "timerDone_slowAnimationTimer:",
            debugTag,
            False)
    
    def timerDone_slowAnimationTimer_(self, timer):
        debugTag = timer.userInfo()
        if debugTag is not None:
            print("timerDone_slowAnimationTimer_", debugTag)
        if self.redrawTimer is not None:
            self.redrawTimer.invalidate()
            self.redrawTimer = None
            self._redrawTimerHits = 0
        self.slowAnimationTimerTick += 1
        self.tapAnimations()
        UpdateCurrentGlyphView()
        if self.slowAnimationTimerRunning:
            self.timerSchedule_slowAnimationTimer()
        
    def timerSchedule_glyphWindowRedraw_stillDragging(self, debugTag=None):
        # timer for the detail levels after the dragging stops
        # calls timerDone_glyphWindowRedraw_stillDragging
        
        if self.redrawTimer is not None:
            self.redrawTimer.invalidate()
            self._redrawTimerHits += 1
            self.redrawTimer = None
        self.redrawTimer= AppKit.NSTimer.scheduledTimerWithTimeInterval_target_selector_userInfo_repeats_(
            self.redrawStillDraggingTimerInterval, 
            self,
            "timerDone_glyphWindowRedraw_stillDragging:", 
            #"timerDone_glyphWindowRedraw_stillDragging_profileWrapper:", 
            debugTag, False)
    
    def timerDone_glyphWindowRedraw_stillDragging_profileWrapper_(self, timer):
        pass
        #ShowProfile(self.timerDone_glyphWindowRedraw_stillDragging_, timer)

        #cProfile.runctx('self.timerDone_glyphWindowRedraw_stillDragging_(timer)', None, locals())
    
    def timerDone_glyphWindowRedraw_stillDragging_(self, timer):
        # redraw while dragging
        # finish up after timerSchedule_glyphWindowRedraw_stillDragging
        debugTag = timer.userInfo()
        if debugTag is not None:
            print("timerDone_glyphWindowRedraw_stillDragging_", debugTag)
        if self.redrawTimer is not None:
            self.redrawTimer.invalidate()
            self.redrawTimer = None
            self._redrawTimerHits = 0
        UpdateCurrentGlyphView()
        self.publishEvent_previewLocationChanged_stillDragging()
        if self.spaceWindow:
            self.spaceWindow_prepareData()
        # do not redraw the other windows here,
        # we're just slowing down, not done.
        # do not change self.currentlyDragging status here, we're still mousedown, still dragging. 
        # ooh can we afford to redraw the text window
        self.timerSchedule_textWindowRedraw()
        #if self.textWindow:
        #    self.textWindow_updateGlyphs()

    def timerSchedule_miniKBD_scrubbing(self, debugTag=None):
        # timer for the knobs
        # calls timerDone_miniKBD_scrubbingCallback
        self.drawDetails = False
        if self.miniKBDScrubbingTimer is not None:
            self.miniKBDScrubbingTimer.invalidate()
            self._miniKBDScrubbingTimerHits += 1
            self.miniKBDScrubbingTimer = None
        self.miniKBDScrubbingTimer = AppKit.NSTimer.scheduledTimerWithTimeInterval_target_selector_userInfo_repeats_(
            self.miniKBDScrubbingTimerInterval,
            self,
            "timerDone_miniKBD_scrubbingCallback:",
            debugTag, False)
    
    def timerDone_miniKBD_scrubbingCallback_(self, timer):
        # redraw when dragging is slow, catching up. Should be fast
        debugTag = timer.userInfo()
        if debugTag is not None:
            print("timerDone_miniKBD_scrubbingCallback_", debugTag)
        if self.miniKBDScrubbingTimer is not None:
            self.miniKBDScrubbingTimer.invalidate()
            self.miniKBDScrubbingTimer = None
            self._miniKBDScrubbingTimerHits = 0
        self.drawDetails = True
        UpdateCurrentGlyphView()

    ## / timers
    def data_getAxisIncrementUnit(self, axisName):
        # calculate our preferred smallest step along this axis
        axisUnit = 0
        axisValues = self.data_getAxisValues(axisName)
        if axisValues:
            mn, df, mx = axisValues
            axisUnit = (mx - mn)/self.axisSteps
        return axisUnit
    
    def tapAnimations(self):
        # give animated axes a tap
        if self.instanceLocation is None:
            self.mainWindow_stopAnimations()
            return
        animatedAxes = self.interactionSources.get('animate')
        new = self.instanceLocation.copy()
        a = math.radians(.25 * self.slowAnimationTimerTick)
        c = 0
        for axisName in animatedAxes:
            axisValues = self.data_getAxisValues(axisName)
            v = axisValues[0] + (1+math.sin(a))*.5 * (axisValues[2] - axisValues[0])
            new[axisName] = v
        self.instanceLocation = new
        self.data_instanceLocationChanged(dragging=False)
        
    def data_stepAlongAxis(self, leftRight=0, upDown=0):
        horizontalAxis = self.interactionSources.get('horizontal')
        verticalAxis = self.interactionSources.get('vertical')
        if self.instanceLocation is None:
            return
        new = self.instanceLocation.copy()
        if leftRight:
            if horizontalAxis:
                for axisName in horizontalAxis:
                    new[axisName] += leftRight * self.data_getAxisIncrementUnit(axisName)
        if upDown:
            if verticalAxis:
                for axisName in verticalAxis:
                    axisValues = self.data_getAxisValues(axisName)
                    if axisValues:
                        new[axisName] += upDown * self.data_getAxisIncrementUnit(axisName)
        self.instanceLocation = new
        
    def mainWindow_monitorLocalEventsCallback_(self, event):
        # callback for the AppKit.NSEvent
        modFlags = event.modifierFlags()
        isCommand = False
        isOption = False
        isShift = False
        isControl = False
        ourEvent = False
        
        miniKbdInput = False
        cursorKeyInput = False
        
        if modFlags & AppKit.NSCommandKeyMask:
            isCommand = True
        if modFlags & AppKit.NSAlternateKeyMask:
            isOption = True
        if modFlags & AppKit.NSShiftKeyMask:
            isShift = True
        if modFlags & AppKit.NSControlKeyMask:
            isControl = True

        main = AppKit.NSApp().mainWindow()
        if main and hasattr(main, "windowName"):
            if main.windowName() == "GlyphWindow":
                if self.drawingWithToolActivePref:
                    keyCode = event.keyCode()
                    if isShift:
                        m = 10
                    elif isOption:
                        if isCommand:
                            m = 0.01
                        else:
                            m = 0.1
                    else:
                        m = 1
                    if keyCode == 123: # cursor key left
                        self.data_stepAlongAxis(leftRight=-m)
                        ourEvent = True
                        cursorKeyInput = True
                    elif keyCode == 124: # cursor key right
                        self.data_stepAlongAxis(leftRight=m)
                        ourEvent = True
                        cursorKeyInput = True
                    elif keyCode == 126: # cursor key up
                        self.data_stepAlongAxis(upDown=m)
                        ourEvent = True
                        cursorKeyInput = True
                    elif keyCode == 125: # cursor key down
                        self.data_stepAlongAxis(upDown=-m)
                        ourEvent = True
                        cursorKeyInput = True
                                
        if ourEvent:
            if self.ds is not None:
                if self.previewMathPref in ['previewVarLib']:
                    self._clipping, self.instanceLocation = self.data_clipLocation(self.instanceLocation)
                else:
                    self._clipping = False
                if self._clipping:
                    AppKit.NSBeep()
                #self.timerSchedule_miniKBD_scrubbing()
                self.data_instanceLocationChanged(dragging=False)
                return None
        return event
    
    def getNextSkateboardMaster(self, direction, windowType):
        #print('skateboard.getNextSkateboardMaster', direction)
        # if editthatnextmaster requests something, this is where we start responding
        try:
            g = CurrentGlyph()
            if g is None:
                # no current glyph, then layer = None
                f = CurrentFont()
                if f is not None:
                    key = f.path, None
                else:
                    key = None, None
            else:
                f = g.font
                if g.layer.name in ['foreground', 'public.default']:
                    key = f.path, None
                else:
                    key = f.path, g.layer.name
            ignoreLayers = True
            if self.verbose:
                print('windowtype', windowType)
            if windowType in ["GlyphWindow"]:
                # route through all the layers
                ignoreLayers = False
            # if ignoreLayers:
            #     l = None
            # if l == "foreground":
            #     l = None
            sourceDescriptor = self.getNextSourceDescriptorFromKey(key, direction, ignoreLayers)
            if self.verbose:
                print('skateboard direction', direction)
                print('skateboard sourceDescriptor', sourceDescriptor)
            if sourceDescriptor is not None:
                candidate = self.openFonts.get(sourceDescriptor.path)
                if self.verbose:
                    print('skateboard candidate', candidate)
                if sourceDescriptor.layerName == None:
                    layerName = 'foreground'
                else:
                    layerName = sourceDescriptor.layerName
                return candidate, layerName
        except:
            print("wrapped error in getnext @@", traceback.format_exc())
        
    def mainWindow_handleControlBoardNotification(self, notification):
        pprint(notification['info'])
        speed = notification['info'].get('speed', 0.1)
        speed = speed ** 2
        #print('speed', speed)
        v = notification['info'].get('other',0)
        v = v*(speed*5)
        if notification['info']['name'] == 'enc1':
            self.data_stepAlongAxis(leftRight=v)
        elif notification['info']['name'] == 'enc2':
            self.data_stepAlongAxis(upDown=v)
        self.data_instanceLocationChanged(dragging=True)
        self.publishEvent_previewLocationChanged_dragFinished()

    def mainWindow_handleMiniKbdNotification(self, notification):
        if self.ds is None:
            return
        clipping = False
        encoderClicks = [False ,False]
        for name, encoderDataItem in notification['data'].items():
            d = encoderDataItem.speed
            n = encoderDataItem.name
            if encoderDataItem.click:
                encoderClicks[name] = True
            m = 1
            if n == 0:
                self.data_stepAlongAxis(leftRight=d*m)
            elif n == 1:
                self.data_stepAlongAxis(upDown=d*m)
            if self.previewMathPref in ['previewVarLib']:
                clipping, self.instanceLocation = self.data_clipLocation(self.instanceLocation)
            else:
                clipping = False
        self._clipping = clipping
        if clipping:
            AppKit.NSBeep()
        if encoderClicks[0]:
            print("click 1")
        if encoderClicks[1]:
            print("click 2")
        self.data_instanceLocationChanged(dragging=True)
        self.publishEvent_previewLocationChanged_dragFinished()
    
    def data_locationContext(self, location):
        # provide the axis dimensions context for this location
        if self.ds is None:
            return {}
        context = {}
        # what is the order? how do we know which one we're working with?
        for axis in self.ds.axes:
            context[axis.name] = (axis.map_forward(axis.minimum), axis.map_forward(axis.default), location[axis.name], axis.map_forward(axis.maximum))
        return context

    def data_clipLocation(self, location):
        # make sure the values in location are within the axis extrema.
        new = {}
        clipping = False
        for axis in self.ds.axes:
            v = location[axis.name]
            if v < axis.map_forward(axis.minimum):
                new[axis.name] = axis.map_forward(axis.minimum)
                clipping = True
            elif v > axis.map_forward(axis.maximum):
                new[axis.name] = axis.map_forward(axis.maximum)
                clipping = True
            else:
                new[axis.name] = v
        return clipping, Location(new)
    
    #def geometryChanged(self):
    #    # callback for when big changes happen in the designspace.
    #    # for instance, new location for master, instance, or point of interest
    #    self.data_collectSnapValues()
        
    def data_collectSnapValues(self):
        # collect all the values from all the axes
        if self.ds is None:
            self.snapValues = {}
            return
        self.snapValues = {}
        axes = {}
        units = {}
        for ad in self.ds.axes:
            self.snapValues[ad.name] = []
            units[ad.name] =  self.data_getAxisIncrementUnit(ad.name)
            axes[ad.name] = self.data_getAxisValues(ad.name)
        for sd in self.ds.sources:
            if not sd.location: continue
            for axisName, value in sd.location.items():
                t = (value-.5*units[axisName], value+.5*units[axisName])
                self.snapValues[axisName].append((t, value))
        for sd in self.ds.instances:
            if not sd.location: continue
            for axisName, value in sd.location.items():
                if type(value) == type(()):
                    for v in value:
                        t = (v-.5*units[axisName], v+.5*units[axisName])
                        self.snapValues[axisName].append((t, v))
                else:
                    t = (value-.5*units[axisName], value+.5*units[axisName])
                    self.snapValues[axisName].append((t, value))
        # ading interesting to snap values
        for location, name in self.pointsOfInterest:
            for axisName, value in location.items():
                t = (value-.5*units[axisName], value+.5*units[axisName])
                self.snapValues[axisName].append((t, value))
        # add open locations
        # add rules values
            
    def data_instanceLocationChanged(self, dragging=True):
        # respond to a new instancelocation
        # @@ mapping
        dimensions = []
        if not self.instanceLocation:
            self.previewAxesList_safelySetItems([])
            #self.pane_previewAxes.locationList.set([])
            return
        mappedLocation = None
        if self.axisMapper:
            mappedLocation = self.axisMapper.map_backward(self.instanceLocation)
        # if self.axisMapper and self.useAxisMaps:
        #     mappedLocation = self.axisMapper.map_backward(self.instanceLocation)
        #     #mappedLocation = self.axisMapper(self.instanceLocation)
        #     #mappedLocation = self.instanceLocation
        # else:
        #     # @@ mapping
        #     mappedLocation = self.axisMapper.map_backward(self.instanceLocation)
        #     #mappedLocation = self.instanceLocation
        #print('$'*30, self.useAxisMaps, mappedLocation)
        for k, v in self.instanceLocation.items():
            e = self.axisExtremes.get(k)
            if e is None:
                return
            axisMinimum, axisDefault, axisMaximum = e
            interactionNameForMenu = None
            for direction, axes in self.interactionSources.items():
                if k in axes:
                    interactionNameForMenu = direction
            if interactionNameForMenu is None:
                interactionNameForMenu = "ignore"
            snappedImage = self._emptyImage
            if self.snapped:
                if k in self.snapped:
                    snappedImage = self._snappedImage
            if k in self.mutedAxes:
                mutedAxisImage = self._ufoMutedImage
            else:
                mutedAxisImage = self._emptyImage
            if mappedLocation:
                mappedValue = ""
                if mappedLocation.get(k) != v:
                    mappedValue = "%3.3f" % mappedLocation.get(k)
            else:
                mappedValue = ""
            # normalised
            normalisedValue = sbx.tools.normalize.normalizeValue(v, (axisMinimum, axisDefault, axisMaximum), self.canExtrapolate)
            #print("+++++", normalisedValue, v, self.canExtrapolate)
            d = dict(
                controlKey = interactionNameForMenu,
                isMutedImage = mutedAxisImage,
                isSnappedImage = snappedImage,
                axisKey = k,
                valueKey = "%3.3f" % v,
                mappedValueKey = mappedValue,
                normalisedValueKey = "%3.4f" % normalisedValue,
                minimumKey = axisMinimum,
                defaultKey = axisDefault,
                maximumKey = axisMaximum,
                )
            dimensions.append(d)
        self.previewAxesList_safelySetItems(dimensions)
        if dragging:
            self.timerSchedule_glyphWindowRedraw_stillDragging()
        else:
            self.timerSchedule_navigatorTool_mouseUp_draggingDone()
    
    def data_checkGlyphDepencies(self):
        # we're done with somet dragging
        # or some sort of edit
        pass
        
    def mainWindow_openNewDocument(self, notification):
        # callback for observer skateboard.openDocument
        # sent by external scripts to open a designspace
        # this means we have to check if we have a different document open, handle saving,
        # or maybe it is open already
        pass
    
    def mainWindow_confirmSkateboardRunningCallback(self, notification):
        # callback for observer on 'skateboard.running'
        # sent by external scripts to find out if Skateboard is running
        publishEvent("skateboard.running.ping",
            location = self.instanceLocation, 
            designspace = self.ds, 
            interactionSources = self.interactionSources, 
            axes = self.axisExtremes,
            dragging=False,
            instanceGlyph = self.instanceGlyphObject,
            previewGlyph = self.contouredInstanceGlyphObject,
            )
        
    def publishEvent_previewLocationChanged_dragFinished(self):
        # callback for publishing this event (needed in several places)
        # call when the drag is finished, relatively cheap
        if self.spaceWindow:
            self.spaceWindow_prepareData()
        publishEvent("skateboard.previewLocation.changed",
            location = self.instanceLocation, 
            designspace = self.ds, 
            interactionSources = self.interactionSources, 
            axes = self.axisExtremes,
            dragging=False,
            instanceGlyph = self.instanceGlyphObject,
            previewGlyph = self.contouredInstanceGlyphObject,
            )

    def publishEvent_previewLocationChanged_stillDragging(self):
        # callback for publishing this event (needed in several places)
        # call while still dragging. Expensive. Should be used with care.
        publishEvent("skateboard.previewLocation.changed",
            location = self.instanceLocation, 
            designspace = self.ds, 
            interactionSources = self.interactionSources, 
            axes = self.axisExtremes,
            dragging=True,
            instanceGlyph = self.instanceGlyphObject,
            previewGlyph = self.contouredInstanceGlyphObject,
            )
    
    def publishEvent_interactionSourcesChanged(self):
        # callback for publishing this event (needed in several places)
        if self.ds is not None and self.spaceWindow:
            self.spaceWindow_prepareData()
        animatingAxes = self.interactionSources.get("animate", [])
        if animatingAxes:
            self.slowAnimationTimerRunning = True
            self.timerSchedule_slowAnimationTimer()
        else:
            self.mainWindow_stopAnimations()
            self.slowAnimationTimerRunning = False
            if self.slowAnimationTimer:
                self.slowAnimationTimer.invalidate()
        publishEvent("skateboard.interactionSources.changed", 
            location = self.instanceLocation, 
            designspace = self.ds, 
            interactionSources = self.interactionSources, 
            axes = self.axisExtremes,
            dragging=False,
            instanceGlyph = self.instanceGlyphObject,
            previewGlyph = self.contouredInstanceGlyphObject,
            )
        # document need save
        self.document_setNeedSave(True)

    def navigatorMenu_storeCurrentLocation(self):
        # add the current location to the series
        if not self.instanceLocation:
            return
        allPoints = [p.asTuple() for p, n in self.pointsOfInterest]
        if self.instanceLocation.asTuple() not in allPoints:
            description = "Marked %s at %s" % (self.currentGlyphName, self.instanceLocation.asString())
            edited = AskString("Label for this location:", description)
            self.pointsOfInterest.append((self.instanceLocation.copy(), edited))
        self.locationsList_updateItemsCallback()
        self.document_setNeedSave(True)
    
    def locationsList_selectLocationCallback(self, sender):
        # callback for a selection in the locationslist.
        #if self.locationsList_doNotUpdate:
        #    return
        sel = sender.getSelection()
        self.allSelectedLocations = []
        if len(sel) == 1:
            s = sel[0]
            #self.somethingSelector = s
            loc = Location(self.seriesLocations[s]['location'])
            self.allSelectedLocations.append(self.seriesLocations[s])
            self.instanceLocation = Location(self.seriesLocations[s]['location'])
            if self.verbose:
                print("locationsList_selectLocationCallback: self.instanceLocation", self.instanceLocation)
        else:
            self.allSelectedLocations = []
            for s in sel:
                loc = Location(self.seriesLocations[s]['location'])
                #print("adding", self.seriesLocations[s])
                self.allSelectedLocations.append(self.seriesLocations[s])
        self.data_instanceLocationChanged(dragging=False)
        
    def mainWindow_stopAnimations(self):
        self.slowAnimationTimerRunning = False
        self.slowAnimationTimerTick = 0
        if self.slowAnimationTimer:
            self.slowAnimationTimer.invalidate()
        
    def mainWindow_textToPasteBoard(self, text):
        pb = AppKit.NSPasteboard.generalPasteboard()
        pb.clearContents()
        pb.declareTypes_owner_([
            NSPasteboardTypeString,
        ], None)
        pb.setString_forType_(text,  AppKit.NSPasteboardTypeString)

    def document_newInstanceCallback(self, arg=None):
        # callback for the navigator menu
        # make a new instance
        # placeholder names - leave fancy editing for later
        iD = self.data_newInstanceDescriptorForCurrentLocation()
        if iD is not None:
            self.ds.addInstance(iD)
            self.document_updateCurrentDesignSpace()
            self.document_setNeedSave(True)
    
    def document_proposeNewSupportLayerName(self, location):
        # we are going to make a new layer
        if self.ds is None:
            if self.verbose:
                print("document_proposeNewSupportLayerName self.ds is None")
            return
        layerName = ['skateboard_support']
        for name in self.ds.getAxisOrder():
            layerName.append("%s_%2.2f"%(name, self.instanceLocation[name]))
        return "_".join(layerName)
    
    def document_currentLayerToSupport(self):
        # vanilla callback
        # alternative to navigatorMenu_previewToSupportCallback
        if self.currentGlyph is None:
            AppKit.NSBeep()
            return
        glyphWindow = CurrentGlyphWindow()
        if glyphWindow is None:
            AppKit.NSBeep()
            return
        previewGlyph = self.instanceGlyphObject.naked()
        thisGlyph = glyphWindow.getGlyph()
        glyphName = thisGlyph.name
        layerName = thisGlyph.layer.name
        thisFont = CurrentFont()

        #proposedLayerName = self.document_proposeNewSupportLayerName(self.instanceLocation)
        #defaultSourceDescriptor = self.ds.findDefault()
        # check if we already have this layer and font
        hasLayer = False
        for sd in self.ds.sources:
            if sd.path != thisFont.path:
                continue
            if layerName == sd.layerName:
                continue
            supportSourceDescriptor = ufoProcessor.SourceDescriptor()
            supportSourceDescriptor.path = sd.path
            supportSourceDescriptor.layerName = layerName
            supportSourceDescriptor.name = layerName
            supportSourceDescriptor.location = self.instanceLocation.copy() # use the 
            self.ds.addSource(supportSourceDescriptor)
            # rebuild sources list
            self.data_updateDocumentVerticalMetrics()
            self.ds.loadFonts(reload=True)
            self.document_checkFonts()
            self.storage_mainPreviews.reset()
            self.storage_supportPreviews.reset()
            self.document_updateCurrentDesignSpace()
            self.document_setNeedSave(True)
            return
        AppKit.NSBeep()
        return
        
    def navigatorMenu_previewToSupportCallback(self, sender=None):
        #@@
        # vanilla callback
        # new supportlayer in the default master
        # copy the outlines of the preview
        # add the new supportlayer to the designspace.
        # find out which location the supportlayer is supposed to have
        # could be it has to be a different location
        if self.currentGlyph is None:
            AppKit.NSBeep()
            return
        glyphWindow = CurrentGlyphWindow()
        if glyphWindow is None:
            AppKit.NSBeep()
            return
        previewGlyph = self.instanceGlyphObject.naked()
        glyphName = glyphWindow.getGlyph().name
        proposedLayerName = self.document_proposeNewSupportLayerName(self.instanceLocation)
        defaultSourceDescriptor = self.ds.findDefault()
        # make a dialog for the skateboard window
        # bring it to the front
        self.mainWindow_show()
        availableSources = []
        ufoNames = []
        # CCC
        for sourceDescriptor in self.ds.sources:
            if sourceDescriptor.path not in self.openFonts: continue
            # only show open fonts
            ufoNames.append(os.path.basename(sourceDescriptor.path))
            availableSources.append(sourceDescriptor.path)
        data = {
            'currentLocation': self.instanceLocation,
            'proposedLayerName': proposedLayerName,
            'glyphName': glyphName,
            'defaultSourceDescriptor': defaultSourceDescriptor,
            'availableSources': availableSources,
            'ufoNames': ufoNames
        }
        s = sbx.ui.sheets.SupportSheet(
                data,
                cancelCallback=self.navigatorToolMenu_cancelNewSupportCallback,
                applyCallback=self.navigatorToolMenu_applyNewSupportCallback,
                parentWindow=self.w)
    
    def navigatorToolMenu_cancelNewSupportCallback(self, sender):
        if self.verbose:
            print("navigatorToolMenu_cancelNewSupportCallback")

    def navigatorToolMenu_applyNewSupportCallback(self, data):
        #@@
        layerName = data['selectedName']
        selectedSource = data['selectedSource']
        targetUFO = None
        for sourceDescriptor in self.ds.sources:
            if sourceDescriptor.path == selectedSource:
                # this is the one, 
                targetUFO = self.ds.fonts.get(sourceDescriptor.name)
                break
        if targetUFO is not None:
            layerNames = []
            for layer in targetUFO.layers:
                layerNames.append(layer.name)
            if not layerName in layerNames:
                targetUFO.prepareUndo("Skateboard Preview to Support")
                layer = targetUFO.newLayer(layerName)
                layer.newGlyph(data['glyphName'])
                g = layer[data['glyphName']]
                g.clear()
                pp = g.getPointPen()
                self.instanceGlyphObject.drawPoints(pp)
                g.width = self.instanceGlyphObject.width
                targetUFO.performUndo()
                targetUFO.save()
                # check if it needs to be saved
                supportSourceDescriptor = ufoProcessor.SourceDescriptor()
                supportSourceDescriptor.path = selectedSource
                supportSourceDescriptor.layerName = layerName
                supportSourceDescriptor.name = "skateboard.support.{}".format(layerName)
                supportSourceDescriptor.location = self.instanceLocation.copy()
                self.ds.addSource(supportSourceDescriptor)
                
                if self.verbose:
                    for d in self.ds.sources:
                        print('sources d', d, d.layerName)
            # rebuild sources list
            #@@
            
            self.data_updateDocumentVerticalMetrics()
            self.ds.loadFonts(reload=True)
            self.document_checkFonts()
            self.storage_mainPreviews.reset()
            self.storage_supportPreviews.reset()
            self.document_updateCurrentDesignSpace()
            self.document_setNeedSave(True)
        
    def navigatorToolMenu_copyPreviewGlyphCallback(self, sender, target=None):
        # vanilla callback
        # previewglyph to the clipboard, include the location in the lib
        if self.currentGlyph is None:
            return
        glyphWindow = CurrentGlyphWindow()
        if glyphWindow is None:
            return
        previewGlyph = self.instanceGlyphObject.naked()
        if not self.navigatorPrefsLibKey in previewGlyph.lib:
            previewGlyph.lib[self.navigatorPrefsLibKey] = {}
        previewGlyph.lib[self.navigatorPrefsLibKey]['location'] = self.instanceLocation
        g = glyphWindow.getGlyph()
        s = writeGlyphToString(g.name, previewGlyph, self.instanceGlyphObject.drawPoints)
        if target in [None, "clipboard"]:
            self.mainWindow_textToPasteBoard(s)
        elif target == "currentLayer":
            g = CurrentGlyph()
            g.prepareUndo("Skateboard: Copy Preview To Current Layer")
            g.clear()
            g.appendGlyph(self.instanceGlyphObject)
            g.width = self.instanceGlyphObject.width
            if not self.navigatorPrefsLibKey in g.lib:
                g.lib[self.navigatorPrefsLibKey] = {}
            g.lib[self.navigatorPrefsLibKey]['location'] = self.instanceLocation
            g.update()
            g.performUndo()
        # rebuild sources list
        self.ds.loadFonts(reload=True)
        self.document_checkFonts()
        self.document_updateCurrentDesignSpace()
        self.document_setNeedSave(True)

    def data_getVariationModel(self, items, axes, bias=None, glyphName=None):
        # Return either a mutatorMath or a varlib.model object for calculating.
        try:
            if self.previewMathPref == "previewVarLib":
                # use the varlib variation model
                try:
                    #print('data_getVariationModel 1')
                    return None, ufoProcessor.VariationModelMutator(items, self.ds.axes)
                except (KeyError, AssertionError) as e:
                    msg = sys.exc_info()[1]
                    if "Base master not found" in msg.args[0]:
                        return "Skateboard can't find  master at the neutral location.", None
                    return "Skateboard can't make a varlib model for %s" % glyphName, None
            else:
                # use mutatormath model
                axesForMutator = self.ds.getMutatorAxes()
                for k, v in axesForMutator.items():
                    v['map'] = None
                    # note: @@ mapping
                    # this mutator should not do any mapping.
                    # remove the map data before building.
                    #print(v['map'])
                try:
                    bias, mut = buildMutator(items, axes=axesForMutator, bias=bias)
                except (IndexError,):
                    msg = sys.exc_info()[1]                    
                    if "list index out of range" in msg.args[0]:
                        return "Can't make a mutatorMath model for %s with incompatible masters." % glyphName, None
                    return "Can't make a mutatorMath model for %s with incompatible masters." % glyphName, None
                return None, mut
        except:
            error = traceback.format_exc()
            return "Skateboard.data_getVariationModel error: %s" % error, None
    
    def data_getAxisTags(self):
        # make a dict with axisname -> axistags
        if self.ds is None:
            return None
        tags = {}
        for ad in self.ds.axes:
            tags[ad.name] = ad.tag
        return tags

    def data_getAxisValues(self, axisName=None):
        # return the minimum / default / maximum for the axis
        # it's possible we ask for an axis that is not in the document.
        if self.ds is None:
            return None
        if axisName is None:
            # get all of them
            axes = {}
            for ad in self.ds.axes:
                axes[ad.name] = (ad.map_forward(ad.minimum), ad.map_forward(ad.default), ad.map_forward(ad.maximum))
            return axes
        for ad in self.ds.axes:
            if ad.name == axisName:
                return (ad.map_forward(ad.minimum), ad.map_forward(ad.default), ad.map_forward(ad.maximum))
        return None

    def spaceWindow_prepareAxisData(self, dd):
        axisRanges = {}
        for axisName in dd.horizontalAxes:
            axisValues = self.data_getAxisValues(axisName)
            span = []
            if axisValues:
                mn, df, mx = axisValues
                for v in [mn, df, mx]:
                    span.append(sbx.tools.normalize.normalizeValue(v, (mn, df, mx), self.canExtrapolate))
            span = list(set(span))
            span.sort()
            axisRanges['horizontal'] = tuple(span)
        for axisName in dd.verticalAxes:
            axisValues = self.data_getAxisValues(axisName)
            span = []
            if axisValues:
                mn, df, mx = axisValues
                for v in [mn, df, mx]:
                    span.append(sbx.tools.normalize.normalizeValue(v, (mn, df, mx), self.canExtrapolate))
            span = list(set(span))
            span.sort()
            axisRanges['vertical'] = tuple(span)
        dd.needsUpdate = True
        self.spaceWindowData['axisRanges'] = axisRanges
        self.spaceWindow.axisRanges = axisRanges
    
    def spaceWindow_prepareInstanceLoc(self, dd):
        # part of spaceWindow_prepareData
        instanceLoc = {}
        for axisName in dd.horizontalAxes:
            if axisName in self.instanceLocation:
                axisValues = self.data_getAxisValues(axisName)
                if axisValues:
                    mn, df, mx = axisValues
                    instanceLoc['horizontal'] = sbx.tools.normalize.normalizeValue(self.instanceLocation.get(axisName), (mn, df, mx), self.canExtrapolate)
        for axisName in dd.verticalAxes:
            if axisName in self.instanceLocation:
                axisValues = self.data_getAxisValues(axisName)
                if axisValues:
                    mn, df, mx = axisValues
                    instanceLoc['vertical'] = sbx.tools.normalize.normalizeValue(self.instanceLocation.get(axisName), (mn, df, mx), self.canExtrapolate)
        dd.needsUpdate = True
        self.spaceWindowData['instanceLoc'] = instanceLoc
        self.spaceWindow.instanceLoc = instanceLoc

    def spaceWindow_prepareSupportsData(self, dd):
        # part of spaceWindow_prepareData
        if self._listSupports:
            newSupports = []
            if self.lastSupports is not None:
                for support in self.lastSupports:
                    new = dict(horizontal=None, vertical=None)
                    for axisInSupport in support.keys():
                        if axisInSupport in dd.horizontalAxes:
                            new['horizontal'] = support[axisInSupport]
                        if axisInSupport in dd.verticalAxes:
                            new['vertical'] = support[axisInSupport]
                    newSupports.append(new)
            dd.needsUpdate = True                
            self.spaceWindowData['supports'] = newSupports
            self.spaceWindow.supports = newSupports
        else:
            self.spaceWindowData['supports'] = None
            self.spaceWindow.supports = None

    def spaceWindow_prepareMasters(self, dd):
        # part of spaceWindow_prepareData
        ignoreAxes = self.interactionSources.get('ignore', [])
        if self._listSources or self._listSupports:
            newSupports = []
            for sd in self.ds.sources:
                ignore, loc = self.ds.filterThisLocation(sd.location, self.mutedAxes)
                if ignore: continue
                t = []    #os.path.basename(sd.path)]
                key = (sd.path, sd.layerName)
                muted = False
                if key in self.mutedFonts:
                    muted = True
                
                for axisName in dd.axisOrder:
                    if axisName not in ignoreAxes:
                        t.append("%s %3.3f" % (axisName, sd.location.get(axisName, "-"), ))
                        
                t.append(os.path.basename(sd.path))
                item = dict(locationString="\n".join(t), flavor="master")
                spaceLoc = {}
                snapLoc = {}
                for axisName in dd.horizontalAxes:
                    if axisName in sd.location:
                        axisValues = self.data_getAxisValues(axisName)
                        if axisValues:
                            mn, df, mx = axisValues
                            spaceLoc['horizontal'] = sbx.tools.normalize.normalizeValue(sd.location.get(axisName), (mn, df, mx), self.canExtrapolate)
                            snapLoc['horizontal'] = axisName, sd.location.get(axisName)
                for axisName in dd.verticalAxes:
                    if axisName in sd.location:
                        axisValues = self.data_getAxisValues(axisName)
                        if axisValues:
                            mn, df, mx = axisValues
                            spaceLoc['vertical'] = sbx.tools.normalize.normalizeValue(sd.location.get(axisName), (mn, df, mx), self.canExtrapolate)
                            snapLoc['vertical'] = axisName, sd.location.get(axisName)
                item['location'] = spaceLoc
                item['snap'] = snapLoc
                item['muted'] = muted
                # @@ here we select the items for the space window.
                isSupport = False
                if sd.layerName is not None:
                    item['flavor'] = "support"
                    item['layername'] = sd.layerName
                    isSupport = True
                    if self._listSupports:
                        dd.masters.append(item)
                        continue
                if self._listSources and not isSupport:
                    dd.masters.append(item)
                dd.needsUpdate = True
            self.spaceWindowData['masters'] = dd.masters
            self.spaceWindow.masters = dd.masters
        else:
            self.spaceWindowData['masters'] = None
            self.spaceWindow.masters = None
                
    def spaceWindow_prepareInstances(self, dd):
        if self._listInstances:
            for sd in self.ds.instances:
                ignore, loc = self.ds.filterThisLocation(sd.location, self.mutedAxes)
                if ignore: continue
                t = []    #os.path.basename(sd.path)]
                for axisName in dd.axisOrder:
                    v =  sd.location.get(axisName, "-")
                    if type(v) == tuple:
                        v = v[0]
                    t.append("%s %3.3f" % (axisName, v))
                t.append(os.path.basename(sd.path))
                item = dict(locationString="\n".join(t), flavor="instance")
                spaceLoc = {}
                snapLoc = {}
                for axisName in dd.horizontalAxes:
                    if axisName in sd.location:
                        axisValues = self.data_getAxisValues(axisName)
                        if axisValues:
                            mn, df, mx = axisValues
                            spaceLoc['horizontal'] = sbx.tools.normalize.normalizeValue(sd.location.get(axisName), (mn, df, mx), self.canExtrapolate)
                            snapLoc['horizontal'] = axisName, sd.location.get(axisName)
                for axisName in dd.verticalAxes:
                    if axisName in sd.location:
                        axisValues = self.data_getAxisValues(axisName)
                        if axisValues:
                            mn, df, mx = axisValues
                            spaceLoc['vertical'] = sbx.tools.normalize.normalizeValue(sd.location.get(axisName), (mn, df, mx), self.canExtrapolate)
                            snapLoc['vertical'] = axisName, sd.location.get(axisName)
                item['location'] = spaceLoc
                item['snap'] = snapLoc
                item['muted'] = False
                dd.instances.append(item)
                dd.needsUpdate = True
            self.spaceWindowData['instances'] = dd.instances
            self.spaceWindow.instances = dd.instances
        else:
            self.spaceWindowData['instances'] = None
            self.spaceWindow.instances = None

    def spaceWindow_prepareRuleData(self, dd):
        # part of spaceWindow_prepareData
        rules = []
        ruleValues = None
        for rule in self.ds.rules:
            try:
                for conditionSetIndex, conditionSet in enumerate(rule.conditionSets):
                    cds = []
                    if ruleValues is None:
                        ruleValues = dict(horizontal=None, vertical=None, name="%s\nconditionset %d" % (rule.name, conditionSetIndex))
                    for condition in conditionSet:
                        axisName = condition['name']
                        mn, df, mx = dd.axisExtremes[axisName]
                        onlyDefault = True
                        d = None
                        if axisName in dd.horizontalAxes:
                            d = "horizontal"
                        elif axisName in dd.verticalAxes:
                            d = "vertical"
                        cMin = condition.get('minimum')
                        if not cMin:
                            cMin =  mn
                        cMax = condition.get('maximum')
                        if not cMax:
                            cMax = mx
                        ruleValues[d] = sbx.tools.normalize.normalizeValue(cMin, (mn, df, mx)), sbx.tools.normalize.normalizeValue(cMax, (mn, df, mx), self.canExtrapolate)
                    if ruleValues['horizontal'] is None:
                        if dd.horizontalAxes:
                            ruleValues['horizontal'] = dd.normalisedExtremes[dd.horizontalAxes[0]]
                    if ruleValues['vertical'] is None:
                        if dd.verticalAxes:
                            ruleValues['vertical'] = dd.normalisedExtremes[dd.verticalAxes[0]]
                    cds.append(ruleValues)
                    ruleValues = None
                    rules.append(cds)
            except:
                pass
            dd.needsUpdate = True
        self.spaceWindowData['rules'] = rules
        self.spaceWindow.rules = rules

    def spaceWindow_prepareInteresting(self, dd):
        interesting = []
        if self._listInteresting:
            for location, name in self.pointsOfInterest:
                item = dict(locationString=name, flavor="interesting")
                spaceLoc = {}
                snapLoc = {}
                for axisName in dd.horizontalAxes:
                    if axisName in location:
                        axisValues = self.data_getAxisValues(axisName)
                        if axisValues:
                            mn, df, mx = axisValues
                            spaceLoc['horizontal'] = sbx.tools.normalize.normalizeValue(location.get(axisName), (mn, df, mx), self.canExtrapolate)
                            snapLoc['horizontal'] = axisName, location.get(axisName)
                for axisName in dd.verticalAxes:
                    if axisName in location:
                        axisValues = self.data_getAxisValues(axisName)
                        if axisValues:
                            mn, df, mx = axisValues
                            spaceLoc['vertical'] = sbx.tools.normalize.normalizeValue(location.get(axisName), (mn, df, mx), self.canExtrapolate)
                            snapLoc['vertical'] = axisName, location.get(axisName)
                item['location'] = spaceLoc
                item['snap'] = snapLoc
                interesting.append(item)
            dd.needsUpdate = True
            self.spaceWindowData['interesting'] = interesting
            self.spaceWindow.interesting = interesting
        else:
            self.spaceWindowData['interesting'] = None
            self.spaceWindow.interesting = None

    def spaceWindow_prepareOpenLocations(self, dd):
        openLocations = []
        if self._listOpen:
            sourceLocations = []
            for sourceDescriptor in self.ds.sources:
                s = Location(sourceDescriptor.location)
                sourceLocations.append(s)
            self._lastOpenNodes = findOpenNodes(sourceLocations, dd.axisOrder)
            for location in self._lastOpenNodes:
                s = Location(location).asString()
                item = dict(locationString=s, flavor="open")
                spaceLoc = {}
                snapLoc = {}
                for axisName in dd.horizontalAxes:
                    if axisName in location:
                        axisValues = self.data_getAxisValues(axisName)
                        if axisValues:
                            mn, df, mx = axisValues
                            spaceLoc['horizontal'] = sbx.tools.normalize.normalizeValue(location.get(axisName), (mn, df, mx), self.canExtrapolate)
                            snapLoc['horizontal'] = axisName, location.get(axisName)
                for axisName in dd.verticalAxes:
                    if axisName in location:
                        axisValues = self.data_getAxisValues(axisName)
                        if axisValues:
                            mn, df, mx = axisValues
                            spaceLoc['vertical'] = sbx.tools.normalize.normalizeValue(location.get(axisName), (mn, df, mx), self.canExtrapolate)
                            snapLoc['vertical'] = axisName, location.get(axisName)
                item['location'] = spaceLoc
                item['snap'] = snapLoc
                dd.openLocations.append(item)
            dd.needsUpdate = True
            self.spaceWindowData['openLocations'] = dd.openLocations
            self.spaceWindow.openLocations = dd.openLocations
        else:
            self.spaceWindowData['openLocations'] = None
            self.spaceWindow.openLocations = None

    
    def spaceWindow_prepareData(self, glyphName=None, ignoreLocations=None):
        if self.ds is None:
            self.spaceWindowData = {}
            return
            
        #print("spaceWindow_prepareData self.previewMathPref", self.previewMathPref)
        if self.previewMathPref == "previewMutatorMath":
            self.spaceWindow.canExtrapolate = True
        else:
            self.spaceWindow.canExtrapolate = False
        dd = Container()
        dd.needsUpdate = False
        dd.horizontalAxes = self.interactionSources.get('horizontal', [])
        dd.verticalAxes = self.interactionSources.get('vertical', [])
        dd.axisOrder = self.ds.getAxisOrder()
        instanceLoc = {}
        axisRanges = {}
        rules = []
        dd.masters = []
        dd.instances = []
        dd.openLocations = []
        dd.trackedClusters = {}
        dd.selectedClusters = {}
        # rules
        dd.normalisedExtremes = {}
        dd.axisExtremes = {}
        for axisName in dd.axisOrder:
            axisValues = self.data_getAxisValues(axisName)
            if axisValues is None: continue
            mn, df, mx = axisValues
            dd.normalisedExtremes[axisName] = sbx.tools.normalize.normalizeValue(mn, (mn, df, mx)), sbx.tools.normalize.normalizeValue(mx, (mn, df, mx), self.canExtrapolate)
            dd.axisExtremes[axisName] = mn, df, mx
        # 
        # rules
        self.spaceWindow_prepareRuleData(dd)
        # supports
        self.spaceWindow_prepareSupportsData(dd)
        # axes
        #if not self.spaceWindowData.get("axisRanges"):
        #    self.spaceWindow_prepareAxisData(dd)
        self.spaceWindow_prepareAxisData(dd)
        # instanceloc - always do this wone
        self.spaceWindow_prepareInstanceLoc(dd)
        # sources
        self.spaceWindow_prepareMasters(dd)
        # instances
        self.spaceWindow_prepareInstances(dd)
        # interesting
        self.spaceWindow_prepareInteresting(dd)
        # open 
        self.spaceWindow_prepareOpenLocations(dd)
        self.spaceWindow.snapValues = self.snapValues
        if dd.needsUpdate:
            self.spaceWindow.update()
    
    def data_prepareGlyph(self, glyphName, force=False):
        # add this glyph to the storage
        drawingIsGood = True
        storedMutatorOk, storedItems = self.storage_mainPreviews.getStoredObject(glyphName)
        if not storedMutatorOk or force==True:
            try:
                stuff = self.data_makeGlyphMutator(glyphName)
                if stuff is not None:
                    _masters, _sourceGlyphObjects, _currentMutator, _componentMutators = stuff
                    t = self.storage_mainPreviews.storeObject(glyphName, (_masters, _sourceGlyphObjects, _currentMutator, _componentMutators))
                    for compName, compMut in _componentMutators.items():    #@XX _componentMutators can be a list!?
                        t = self.storage_mainPreviews.storeObject(compName, (_masters, _sourceGlyphObjects, compMut, {}))
            except (ValueError):
                drawingIsGood = False
        return drawingIsGood
        
    def data_makeGlyphMutator(self, glyphName=None, ignoreLocations=None, depth=0):
        # make a mutator for this glyph
        # glyphname: the name of the glyph we're making a mutator for.
        # ignorelocations: list of location dicts to ignore
        # depth: recursion depth counter
        if self.ds is None:
            return
        theseMasters = []
        theseSourceGlyphs = []
        if ignoreLocations is None:
            ignoreLocations = []        # a way to ignore certain locations
        self.sourceGlyphObjects = []    # for processing selections
        cF = CurrentFont()
        if glyphName is None:
            glyphName = self.currentGlyphName
        if glyphName is None:
            return
        glyphsForComponents = []
        
        layersForComponents = {}
        defaultSourceDescriptor = self.ds.findDefault()
        defaultFont = None
        defaults = {}
        for aD in self.ds.axes:
            # set default location, safest to go to the default. 
            defaults[aD.name] = aD.default
        #checkThisLayerForComponents = None
        for sd in self.ds.sources:
            # XXX here we go through all the sources
            # there may be glyphs in layers that are in the default master, but are not default glyphs
            # we can't rely on the font.path alone to determine the location
            thisIsDefault = defaultSourceDescriptor == sd
            # check if this master is muted
            key = (sd.path, sd.layerName)
            if key in self.mutedFonts:
                continue
            if sd.path in self.missingFonts:
                continue
            # check if this master's location is not wanted
            if ignoreLocations:
                if sd.location in ignoreLocations:
                    continue
                
            # check if this master is on a muted axis
            axes = {}
            ignoreMaster = False
            for mutedAxisName in self.mutedAxes.keys():
                # what is the default value for this axis?
                if sd.location.get(mutedAxisName, -2) != defaults.get(mutedAxisName, -1):
                    ignoreMaster = True
                    break
            if ignoreMaster:
                continue
            isCurrent = False
            if cF is not None:   
                if sd.path == cF.path:
                    isCurrent = True
            if sd.path in self.openFonts:
                # get a glyph from an open font
                f = self.openFonts.get(sd.path)
                if thisIsDefault:
                    defaultFont = f
                mg = None
                if sd.layerName is not None:
                    layerNames = []
                    for layer in f.layers:
                        layerNames.append(layer.name)
                    if sd.layerName not in layerNames:
                        continue
                    #checkThisLayerForComponents = layerObject = f.getLayer(sd.layerName)
                    layerObject = f.getLayer(sd.layerName)
                    if glyphName in layerObject:
                        # aha! opening a glyph in a layer creates that glyph
                        # but it does not contain any outlines
                        # so just opening an edit window breaks compatibility.
                        mg = layerObject[glyphName]
                        if not mg.contours and not mg.components:
                            mg = None
                            continue
                else:
                    if glyphName in f:
                        mg = f[glyphName]
                if mg is None:
                    continue
                    
                if len(mg.components)>0:
                    for comp in mg.components:
                        if not comp.baseGlyph in glyphsForComponents:
                            glyphsForComponents.append(comp.baseGlyph)
                mg.isCurrent = isCurrent
                mgNaked = mg.naked()
                self.sourceGlyphObjects.append(mg)
                theseSourceGlyphs.append(mg)
                #print('000000000 mgNaked',mgNaked, mgNaked.image)
                mmg = mg.toMathGlyph()
                mmg.guidelines = []
                theseMasters.append((Location(sd.location), mmg))
            elif sd.path in self.needFonts:
                # get a glyph from a ufo file
                # NEED TO CHECK LAYER CONTENTS HERE
                f = self.needFonts.get(sd.path)
                if sd.layerName is not None:
                    if sd.layerName in [l.name for l in f.layers]:
                        layer = None
                        for l in f.layers:
                            if l.name == sd.layerName:
                                layer = l
                                break
                        if thisIsDefault:
                            defaultFont = layer
                            #checkThisLayerForComponents = layer
                        if glyphName in layer:
                            mg = layer[glyphName]
                            mg.isCurrent = False
                            if not mg.contours and not mg.components:
                                mg = None
                                continue
                            if len(mg.components)>0:
                                for comp in mg.components:
                                    if not comp.baseGlyph in glyphsForComponents:
                                        glyphsForComponents.append(comp.baseGlyph)
                            self.sourceGlyphObjects.append(mg)
                            theseSourceGlyphs.append(mg)
                            mmg = mg.toMathGlyph()
                            mmg.guidelines = []
                            theseMasters.append((Location(sd.location), mmg))
                else:
                    if thisIsDefault:
                        defaultFont = f
                        #checkThisLayerForComponents = f
                    if glyphName in f:
                        mg = f[glyphName]    # make it a math glyph too
                        mg.isCurrent = False
                        self.sourceGlyphObjects.append(mg)
                        theseSourceGlyphs.append(mg)

                        if len(mg.components)>0:
                            for comp in mg.components:
                                if not comp.baseGlyph in glyphsForComponents:
                                    glyphsForComponents.append(comp.baseGlyph)
                        mmg = mg.toMathGlyph()
                        mmg.guidelines = []
                        theseMasters.append((Location(sd.location), mmg))
        # theseMasters contains all the geometry we have to work with
        # can we get a count of points?
        if self.data_guessDrawingLoad(theseMasters) > 1000:
            self.previewTooHeavy = True
        mut = None
        error, mut = self.data_getVariationModel(theseMasters, self.ds.axes, glyphName=glyphName)
        if error is not None:
            self.sourceGlyphObjects = []
            return theseMasters, theseSourceGlyphs, None, {}    # can't a return a list 
        addThese = []
        resultGlyphsForComponents = {}
        for n in glyphsForComponents: 
            # get the other parts here
            compMasters, compGlyphObjects, compMut, compShop = self.data_makeGlyphMutator(n, ignoreLocations=ignoreLocations, depth=depth+1)
            resultGlyphsForComponents[n] = compMut
            if compShop:
                resultGlyphsForComponents.update(compShop)
        return theseMasters, theseSourceGlyphs, mut, resultGlyphsForComponents
    
    def data_guessDrawingLoad(self, masters):
        masters = [m for l, m in masters]
        pts = 0
        for glyph in masters:
            pts += sum([len(c['points']) for c in glyph.contours])
        self.pointLoad = pts
        return pts
        
    def mainWindow_enableInterface(self, state=True):
        # $$
        if state:
            # enable everything
            #self.pane_doc.l.enable(True)
            #self.pane_previewAxes.locationList.enable(True)
            #self.pane_locs.filterLocations.enable(True)
            #self.pane_locs.locationsList.enable(True)
            self.locationsList_setFilterLocationButton()
            self.previewPane_setMathPickerButton()
            #self.pane_previewAxes.show(True)
            #self.pane_locs.show(True)
            #self.pane_doc.show(True)
            self.pane_previewAxes.mathModelPicker.enable(True)
            self.w.panes.show(True)
            self.w.file.saveButton.show(True)
            self.w.frontPage.show(False)
            self.w.file.textWindowButton.show(True)
            self.w.file.spaceWindowButton.show(True)
            self.w.file.closeButton.show(True)
            #self.w.debugButton.show(True)
            if self._remaining is not None:
                self.w.buyButton.show(False)
            #self.w.frontPage.openButton.show(False)
            self.w.file.exportUFOButton.show(True)
            self.w.file.addGlyphButton.show(True)
            if hasDesignSpaceEditor:
                self.w.file.editDSButton.show(True)
            if hasVariableFontGenerator:
                self.w.file.makeVFButton.show(True)            
        else:
            # disable everything
            #self.pane_doc.l.enable(False)
            #self.pane_previewAxes.locationList.enable(False)
            #self.pane_locs.filterLocations.enable(False)
            #self.pane_locs.locationsList.enable(False)
            self.locationsList_setFilterLocationButton(reset=True)
            self.pane_previewAxes.mathModelPicker.enable(False)
            #self.pane_previewAxes.show(False)
            #self.pane_locs.show(False)
            #self.pane_doc.show(False)
            self.w.file.saveButton.show(False)
            self.w.frontPage.show(True)
            self.mainWindow_refreshCandidateDocumentsList()
            self.w.file.textWindowButton.show(False)
            self.w.file.spaceWindowButton.show(False)
            self.w.file.closeButton.show(False)
            #self.w.frontPage.openButton.show(True)
            self.w.file.exportUFOButton.show(False)
            self.w.file.addGlyphButton.show(False)
            self.w.panes.show(False)
            #self.w.debugButton.show(False)
            if self._remaining is not None:
                self.w.buyButton.show(True)
            if hasDesignSpaceEditor:
                self.w.file.editDSButton.show(False)
            if hasVariableFontGenerator:
                self.w.file.makeVFButton.show(False)            

    def document_updateCurrentDesignSpace(self):
        if self.ds is None:
            self.pane_doc.l.set([])
            self.locationsList_safelySetItems()
            self.seriesLocations = []
            return
        self.ds.loadFonts()    # reload?
        self.document_checkFonts()
        self.document_checkAxes()
        self.locationsList_updateItemsCallback()

    def locationsList_updateItemsCallback(self):
        # update the locations list from sources, instances and poins of interest
        self.seriesLocations = [] # hmmm what about preview locations?
        if self.ds is None: return 
        sourceLocations = []
        for sourceDescriptor in self.ds.sources:
            s = Location(sourceDescriptor.location)
            sourceLocations.append(s)
        if self._listInstances:
            for instanceDescriptor in self.ds.instances:
                #s = Location(instanceDescriptor.location).asString()
                s = "%s %s" %(instanceDescriptor.familyName, instanceDescriptor.styleName)
                self.seriesLocations.append(dict(location=instanceDescriptor.location, description=s, flavor="Instance", obj=instanceDescriptor))
        if self._listSources:
            for sourceDescriptor in self.ds.sources:
                s = Location(sourceDescriptor.location).asString()
                name = os.path.basename(sourceDescriptor.path)
                if not sourceDescriptor.layerName:
                    self.seriesLocations.append(dict(location=sourceDescriptor.location, description=name, flavor="Source"))
        if self._listSupports:
            for sourceDescriptor in self.ds.sources:
                s = Location(sourceDescriptor.location)
                name = os.path.basename(sourceDescriptor.path)
                d = "%s layer in %s" %(sourceDescriptor.layerName, name)
                if sourceDescriptor.layerName:
                    self.seriesLocations.append(dict(location=sourceDescriptor.location, description=d, flavor="Support"))
        if self._listOpen:
            self._lastOpenNodes = findOpenNodes(sourceLocations, self.ds.getAxisOrder())
            for loc in self._lastOpenNodes:
                s = Location(loc).asString()
                self.seriesLocations.append(dict(location=loc, description=s, flavor="Open"))
        if self._listInteresting:
            for location, name in self.pointsOfInterest:
                if not name:
                    name = Location(location).asString()
                self.seriesLocations.append(dict(location=location, description=name, flavor="Interesting"))
        self.pane_locs.locationsList.set(self.seriesLocations)
        #self.locationsList_safelySetItems(self.seriesLocations)
        
    def document_checkAxes(self):
        # get the axes data from the designspace,
        # provide the axes
        newLocation = None
        newAxisNames = []
        newAxisExtremes = {}
        loc = {}
        for aD in self.ds.axes:
            # set default location, safest to go to the default. 
            loc[aD.name] = aD.default
            newAxisNames.append(aD.name)
            # @@ mapping
            newAxisExtremes[aD.name] = (aD.map_forward(aD.minimum), aD.map_forward(aD.default), aD.map_forward(aD.maximum))
        geometryChanged = False
        if self.axisNames != newAxisNames:
            self.axisNames = newAxisNames
            geometryChanged = True
        if self.axisExtremes != newAxisExtremes:
            self.axisExtremes = newAxisExtremes
            geometryChanged = True
        if geometryChanged:
            self.instanceLocation = Location(loc)
        self.axisMapper = ufoProcessor.varModels.AxisMapper(self.ds.axes)
        self.data_collectSnapValues()
    
    def getAllFonts(self):
        # return a list of all masters
        return list(self.needFonts.items()) + list(self.openFonts.items())
            
    def document_checkFonts(self):
        # find out which fonts are needed by the designspace
        # and then which fonts we have open in RoboFont.
        self.fontPaths = []
        self.openFonts = {}    # fonts, actually open. key:fontpath, value: fontobject
        self.needFonts = {}    # fonts, needed, but closed
        self.missingFonts = {}
        show = []
        if self.ds is None:
            self.pane_doc.l.set([])
            return
        fp = set()
        sourceDescriptors = {}
        for sd in self.ds.sources:
            # here we have access to the layer name
            k = (sd.path, sd.layerName)
            if k not in self.fontPaths:
                self.fontPaths.append(k)
                sourceDescriptors[k] = sd
            if sd.path is None:
                self.missingFonts[sd.path] = True
            elif not os.path.exists(sd.path):
                self.missingFonts[sd.path] = True
        af = AllFonts()
        currentFont = []
        if af:
            currentFont = af[0]
        #hasNeutral = self.ds.getNeutralFont()
        hasNeutral = self.ds.findDefault()
        defaultLayerName = None
        defaultPath = None
        ufoStatus = None
        if hasNeutral is not None:
            defaultLayerName = hasNeutral.layerName
            defaultPath = hasNeutral.path
        for p, layerName in self.fontPaths:
            ufoIsNeutral = False
            fontIsMuted = False
            layerNameString = layerName
            if layerName is None:
                layerNameString = ""
            fontObj = None
            glyphObj = None
            fontIsNeutralImage = self._emptyImage
            if os.path.exists(p):
                
                fontIsOpen = self._ufoAvailableImage
                ufoStatus = "available"
            else:
                fontIsOpen = self._ufoMissingImage
                ufoStatus = "missing"
            if p == defaultPath and layerName == defaultLayerName:
                fontIsNeutralImage = self._ufoNeutralImage
                ufoIsNeutral = True
            currentFontDone = False
            for f in AllFonts():
                if f.path == p and p is not None:
                    if f == currentFont:
                        fontIsOpen = self._ufoCurrentImage
                        ufoStatus = "current"
                    else:
                        fontIsOpen = self._ufoOpenImage
                        ufoStatus = "open"
                    fontObj = f
                    self.openFonts[p] = fontObj
                    break
            if fontObj is None:
                # the font is not open
                # so: open it
                if os.path.exists(p):
                    try:
                        fontObj = RFont(p, showInterface=False)
                        self.needFonts[p] = fontObj
                    except:
                        # faulty thing?
                        print("Skateboard: Trouble reading ufo from", p)
            if fontObj:
                #@@ check vertical metrics
                self.data_updateDocumentVerticalMetrics(fontObj)
                if self.currentGlyph is None:
                    glyphObj = "-"
                else:
                    if self.currentGlyph.name in fontObj:
                        glyphObj = fontObj[self.currentGlyph.name].__class__.__name__ + " " + self.currentGlyph.name
                    else:
                        glyphObj = "-"
            names = []
            key = (p,layerName)
            if key in self.mutedFonts:
                fontIsMutedImage = self._ufoMutedImage
            else:
                fontIsMutedImage = self._emptyImage
            show.append(dict(
                isNeutralImage=fontIsNeutralImage,
                isOpenImage=fontIsOpen,
                isMutedImage = fontIsMutedImage,
                name=os.path.basename(p ),
                _fontObject=str(fontObj.__class__.__name__),
                _glyphObject=glyphObj,
                _fontid = "%X"%id(fontObj),
                layer=layerNameString,
                #function="",
                factor="",
                _sourceDescriptorKey = key,    # used by getSourceDescriptorFromKey
                _ufoStatus = ufoStatus,
                _ufoIsNeutral = ufoIsNeutral
                ))
        self.pane_doc.l.set(show)
    
    def getNextSourceDescriptorFromKey(self, key, direction=0, ignoreLayers=True):
        # return the sourcedescriptor that matches this path + layerName
        verbose = False
        if verbose:
            print('skateboard getNextSourceDescriptorFromKey direction', direction)
            print('skateboard getNextSourceDescriptorFromKey ignoreLayers', ignoreLayers)
            print('skateboard getNextSourceDescriptorFromKey key', key)
        if self.ds is None:
            if verbose:
                print('skateboard getNextSourceDescriptorFromKey 1')
            return
        #print([os.path.basename(sd.path) for sd in self.ds.sources])
        # make a list of sources we cycle through
        # include layernames or not
        try:
            sourceOrder = []
            sources = []
            if ignoreLayers:
                # font window, space center
                for sd in self.ds.sources:
                    if sd.path not in self.openFonts: continue
                    if not sourceOrder:
                        sourceOrder.append(sd.path)
                        sources.append(sd)
                        continue
                    if sd.path not in sourceOrder:
                        sourceOrder.append(sd.path)
                        sources.append(sd)
            else:
                # glyph window
                for sd in self.ds.sources:
                    if sd.path not in self.openFonts: continue
                    sources.append(sd)
                    
            if self.verbose:
                #print('Routing: (ignoreLayers=', ignoreLayers)     
                for i, s in enumerate(sources):
                    print("\t",  i, os.path.basename(s.path), s.layerName)
            for i, sd in enumerate(sources):
                newKey = (sd.path, sd.layerName)
                if key == newKey:
                    if verbose:
                        print('\t key matched', i, key)
                    if direction == 0:
                        if verbose:
                            print('\ta', i)
                        return sd
                    elif direction == -1:
                        if verbose:
                            print('\tb', i-1)
                        return sources[i-1]
                    elif direction == 1:
                        if verbose:
                            print('@@', i, i == len(sources)-1)
                        if i == len(sources)-1:
                            ii = 0
                        else:
                            ii = i+ 1
                        if self.verbose:
                            print('\tc', ii )
                        return sources[ii]
        except:
            print("Skateboard: Error routing.", traceback.format_exc())
        finally:
            return None

    def getSourceDescriptorFromKey(self, key):
        # return the sourcedescriptor that matches this path + layerName
        for sd in self.ds.sources:
            if key == (sd.path, sd.layerName):
                return sd
        return None

    def debug_makeObjectBrowserCallback(self, sender):
        try:
            methodNames = []
            m = [getattr(self, method) for method in dir(self)]
            for item in m:
                if type(item) == MethodType:
                    methodNames.append(item.__name__)
            methodNames.sort()
            t = time.strftime("%a, %d %b %Y %H:%M:%S", time.localtime())
            browserWindow = vanilla.Window((1000, 600), "Skateboard Debug %s" %(t), minSize=(400,500))
            data = {}
            if self.instanceLocation is not None:
                data['instanceLocation'] = Location(self.instanceLocation).asString()
            data['openFonts'] = [os.path.basename(p) for p in list(self.openFonts.keys())]
            data['needFonts'] = [os.path.basename(p) for p in list(self.needFonts.keys())]
            data['designspace document'] = self.ds
            if self.ds is not None:
                data['fonts'] = list([(k, v.path) for k, v in self.ds.fonts.items()])
                data['designspace path'] = self.ds.path
                data['document.sources'] = self.ds.sources
                data['document.instances'] = self.ds.instances
                data['document.rules'] = self.ds.rules
            data['math model'] = self.previewMathPref
            data['show sparks pref'] = self.showSparksPref
            data['rounding'] = self.previewRoundPref
            data['textWindow'] = self.textWindow
            data['spaceWindow'] = self.spaceWindow
            data['skateboardBundle.version'] = skateboardBundle.version
            data['previewGlyphName'] = self.currentGlyphName
            data['defcon glyphobservers'] = [str(g) for g in self._defconObservers_glyphs.values()]
            if self.checker:
                data['designspaceProblems'] = [str(p) for p in self.checker.problems]
            browserWindow.browser = vanilla.ObjectBrowser((0,0,0,0), data)
            browserWindow.open()
            
        except:
            print("Skateboard: Ironically, can't open the debug window because of a problem.", traceback.format_exc())
    
    def mainWindow_becameMain(self, sender):
        # vanilla callback
        # window became main - look for document changes
        if not self.w.isVisible():
            return
        hasEdits = self.document_hasExternalModifications()
        if hasEdits:
            if self.verbose:
                print("Skateboard: designspace has external edits")
            self.mainWindow_openDocumentWithPath(self.ds.path)
            self.document_setNeedSave()
        else:
            self.mainWindow_refreshCandidateDocumentsList()
            self.document_updateCurrentDesignSpace()
    
    def mainWindow_checkSaveStateBeforeClosing(self, sender):
        if self.documentNeedSave:
            # ask if we need to save this
            result = mojo_AskYesNoCancel("Save Designspace?",
                title="Skateboard",
                informativeText="The current Designspace has unsaved changes.")
        
    def mainWindow_shouldClose(self, sender=None):
        # vanilla callback
        # close the docuent, then close the window
        # see also document_closeDesignSpaceCallback
        def callback(result):
            #print('2', result)
            if result == 1:
                # save first
                #print("4 save first")
                self.document_saveDesignSpaceCallback()
                self.document_setNeedSave(False)
                self.mainWindow_presentCloseDocumentDialog_followUp()
                self.w.close()
                return True
            if result == 0:
                #print('3 closing')
                self.document_setNeedSave(False)
                self.mainWindow_presentCloseDocumentDialog_followUp()
                self.w.close()
                return True
            else:
                return False
        
        # are there unsaved changes?
        if self.documentNeedSave:
            #print('5 need save')
            vanilla.dialogs.askYesNoCancel(messageText="Save this designspace?", informativeText="", parentWindow=self.w, resultCallback=callback)
            return False
        else:
            #print('6 no save ')
            self.mainWindow_presentCloseDocumentDialog_followUp()
            self.w.close()
            return True

        #self.mainWindow_presentCloseDocumentDialog(self.mainWindow_shouldClose_followUp)
    
    def mainWindow_shouldClose_followUp(self, result):
        if result == -1:
            return
        self.mainWindow_stopAnimations()
        self.cb = None
        self.ds = None
        self.mainWindow_removeObservers()

        self.instanceFont = None
        if self.localEventMonitor is not None:
            AppKit.NSEvent.removeMonitor_(self.localEventMonitor)
        self.localEventMonitor = None
        if self.verbose:
            print("remove monitor removed")
        if self.textWindow is not None:
            self.textWindow.mainWindowIsClosing()
        if self.spaceWindow is not None:
            self.spaceWindow.mainWindowIsClosing()
        if self.navigatorTool is not None:
            self.navigatorTool.RFPreviewCallback = None
            if self.verbose:
                print("uninstalling navigator")
            self.navigatorTool.clear()
            uninstallTool(self.navigatorTool)
            self.navigatorTool = None
        self.miniKBDScrubbingTimer = None
        publishEvent("skateboard.window.closed",
            location=None,
            designspace=None,
            interactionSources=None,
            axes=None,
            instanceGlyph = self.instanceGlyphObject,
            previewGlyph = self.contouredInstanceGlyphObject,
            )
                
    def data_pairsFromGlyphNames(self, names):
        # return a list of unique glyphnames and the kernpairs
        pairs = []
        for i, name in enumerate(names):
            if i < len(names)-1:
                p = (name, names[i+1])
                if p not in pairs:
                    pairs.append(p)
        return pairs
    
    def data_prepareNewDesignSpace(self,
        igoreMuting=False,
        save=False,
        ignoreAnisotropic=True,
        clipLocations=True,
        instances=None,
        tagName="temp",
        remapAxes=False,    # true if we want the axis extremes + default to be in designspace coords
        ):
        # more generic support for export_generateUFOCallback
        # make a new designspace based on the current one
        # but with muting supported
        # @@ mapping
        # 
        newDs = SkateboardDesignSpaceProcessor()
        #newDs.update(self.ds)
        newDs.findDefault()
        newDs.loadFonts()
        newDs.useVarlib = self.ds.useVarlib

        # axes
        for ad in self.ds.axes:
            if ad.name in self.mutedAxes:
                continue
            newAd = AxisDescriptor()
            newAd.name = ad.name
            newAd.tag = ad.tag
            if remapAxes:
                newAd.minimum = ad.map_forward(ad.minimum)
                newAd.maximum = ad.map_forward(ad.maximum)
                newAd.default = ad.map_forward(ad.default)
                newAd.map = []    #ad.map[:]
            else:
                newAd.minimum = ad.minimum
                newAd.maximum = ad.maximum
                newAd.default = ad.default
                newAd.map = ad.map[:]
            newAd.labelNames.update(ad.labelNames)
            newDs.addAxis(newAd)
        # sources
        defaults = {}
        for aD in self.ds.axes:
            # set default location, safest to go to the default. 
            defaults[aD.name] = aD.default
        for sd in self.ds.sources:
            key = sd.path, sd.layerName
            if not igoreMuting:
                if key in self.mutedFonts:
                    continue
            ignoreMaster = False
            for mutedAxisName in self.mutedAxes.keys():
                # what is the default value for this axis?
                if mutedAxisName in sd.location:
                    if sd.location[mutedAxisName] != defaults.get(mutedAxisName):
                        ignoreMaster = True
                        break
            if ignoreMaster:
                continue
            newSd = SourceDescriptor()
            newSd.name = sd.name
            newSd.path = sd.path
            newSd.location = {}
            # don't clip for mutatormath
            if self.previewMathPref in ['previewVarLib']:
                clipping, loc = self.data_clipLocation(sd.location)
                newSd.location = loc.asDict()
            else:
                newSd.location.update(sd.location)

            for mutedAxisName in self.mutedAxes.keys():
                if mutedAxisName in newSd.location:
                    del newSd.location[mutedAxisName]
            newSd.layerName = sd.layerName
            newSd.copyLib = sd.copyLib
            newSd.copyInfo = sd.copyInfo
            newSd.copyGroups = sd.copyGroups
            newSd.copyFeatures = sd.copyFeatures
            newSd.muteKerning = sd.muteKerning
            newSd.muteInfo = sd.muteInfo
            newSd.mutedGlyphNames = sd.mutedGlyphNames[:]
            newSd.familyName = sd.familyName
            newSd.styleName = sd.styleName
            newDs.addSource(newSd)
        # instances
        if instances is None:
            instances = self.ds.instances[:]
        for sd in instances:
            ignore, loc = self.ds.filterThisLocation(sd.location, self.mutedAxes)
            if igoreMuting:
                if ignore: continue
            iD = InstanceDescriptor()
            if ignoreAnisotropic:
                # don't add anisotropic instances 
                d = {}
                for axisName, value in sd.location.items():
                    if type(value)== tuple:
                        d[axisName] = value[0]
                    else:
                        d[axisName] = value
                iD.location = d
            else:
                iD.location = loc
            clipping, loc = self.data_clipLocation(iD.location)
            if clipping and clipLocations:
                # don't add clipped instances
                continue
            iD.location = loc
            iD.familyName = sd.familyName
            iD.styleName = sd.styleName
            iD.path = sd.path
            newDs.addInstance(iD)
        # the Rules, Gromit, we've forgotten the rules!
        # do we need to process them?
        for rD in self.ds.rules:
            newDs.addRule(rD)
        
        # debug save
        dsDebugPath = None
        if save:
            dsDebugPath = self.ds.path.replace('.designspace', "_{}.designspace".format(tagName))
            newDs.write(dsDebugPath)
        return dsDebugPath, newDs

    def data_newInstanceDescriptorForCurrentLocation(self):
        # make a new instance descriptor for the current location
        if self.instanceLocation is None:
            return
        iD = InstanceDescriptor()
        if self.verbose:
            print("data_newInstanceDescriptorForCurrentLocation", self.instanceLocation)
        iD.location = self.instanceLocation.asDict()
        iD.familyName = os.path.basename(self.ds.path).replace(".designspace", "")
        iD.kerning = True
        styleNameParts = []
        for name in self.ds.getAxisOrder():
            styleNameParts.append("%s_%2.2f"%(name, self.instanceLocation[name]))
        iD.styleName = "_".join(styleNameParts)
        filename = "%s_%s.ufo" % (iD.familyName, iD.styleName)
        iD.path = os.path.join(os.path.dirname(self.ds.path), "instances", filename)
        return iD
    
    def mainWindow_newGlyphsCallback(self, item):
        # add a new glyph to the designspace.
        #
        selection = []
        if type(item) == dict:
            selection = item.get("newGlyphs")
        sbx.ui.sheets.AddGlyphsSheet(
            data=dict(proposedNames=selection),
            cancelCallback=self.sheet_cancel_genericCallback,
            applyCallback=self.mainWindow_applyNewGlyphsCallback,
            parentWindow=self.w,
            )
    
    def mainWindow_openAllFonts(self):
        # open all the fonts we have
        for f in self.needFonts.keys():
            f = OpenFont(f, showInterface=True)
            w = CurrentFontWindow().window()
            w.show() # ? bizar

    def mainWindow_applyNewGlyphsCallback(self, data=None):
        if not data:
            return
        added = []
        self.mainWindow_openAllFonts()
        for f in self.openFonts.values():
            print("mainWindow_applyNewGlyphsCallback f", f.__class__)
            for name, uniValue in data['proposedNames']:
                if not name in f:
                    f.newGlyph(name)
                    g = f[name]
                    g.width = 500
                    #if version >= "3.0":
                        # RF 3.0
                        #g.markColor = (0, 0.95, 0.95, .05)
                    g.unicode = uniValue
                    g.update()
                    added.append((name, f.path))
        # do all the updates - 
        for f in self.openFonts.values():
            f.update()
        for name, path in added:
            self.data_prepareGlyph(name, force=True)
        self.document_updateCurrentDesignSpace()
        #@@ 
    
    def mainWindow_copyCurrentGlyphToAllMasters(self):
        # copy the current glyph to all masters in this designspace.
        # response to
        g = CurrentGlyph()
        self.mainWindow_openAllFonts()
        for f in self.openFonts.values():
            if f == g.font:
                continue
            if not g.name in f:
                f.newGlyph(g.name)
            target = f[g.name]
            target.clear()    # really?
            if self.verbose:
                print("appending", g, g.font, "to", target, target.font)
            target.appendGlyph(g)
            target.width = g.width
            target.update()
        self.data_prepareGlyph(g.name)
        g.update()
        OpenGlyphWindow(g)
        
    def mainWindow_openDSEditorCallback(self, sender):
        # if we have designspaceeditor installed, this is a button that should:
        # - ask if this document needs to be saved / and is ok to close?
        # - open this document in designspace editor
        # - close this window
        def callback(result):
            #print(result)
            if result == 1:
                thisPath = self.ds.path
                self.document_saveDesignSpaceCallback()
                self.document_closeDesignSpaceCallback()
                openDocs = self.getOpenDesignspacePaths()
                if thisPath in openDocs:
                    editor = self.getDesignspaceWindowFromPath(thisPath)
                    editor.w.getNSWindow().makeKeyAndOrderFront_(editor)
                else:
                    DesignSpaceEditor(thisPath)

        vanilla_AskYesNoCancel("Open this document in DesignSpaceEditor?",
            informativeText="Skateboard will save and close.",
            resultCallback=callback,
            parentWindow=self.w,
            )
    
    def document_muteMissingSourcesAndLayers(self):
        # mute all the layers and sources that are missing
        shouldMute = []
        for sD in self.ds.sources:
            usedLayers = []
            for sd in self.ds.sources:
                key = sd.path, sd.layerName
                if not sd.name in self.ds.fonts:
                    self.mutedFonts[key] = True
                    shouldMute.append(key)
                    continue
                if sd.layerName:
                    usedLayers.append(sd.layerName)
                    if sd.name in self.ds.fonts:
                        thisFont = self.ds.fonts[sd.name]
                        if not sd.layerName in thisFont.layerOrder:
                            # missing layer!
                            shouldMute.append(sd)
                            key = sd.path, sd.layerName
                            if not key in self.mutedFonts:
                                self.mutedFonts[key] = True
                                shouldMute.append(key)
        return shouldMute
            
    def export_makeVariableFontCallback(self, sender):
        # check if we need to save first
        if self.documentNeedSave:
            if self.verbose:
                print("need to save before generating?")
            self.document_saveDesignSpaceCallback()
        bar = ProgressBar()
        bar.label("Generating variable font…")
        bar.tick()

        muted = self.document_muteMissingSourcesAndLayers()
        generatingDesignSpace = self.ds
        # save the ds as a 
        tempPath, generatingDesignSpace = self.data_prepareNewDesignSpace(
            igoreMuting = False,
            ignoreAnisotropic = True,
            save=True,
            clipLocations=True,
            tagName="VFExport",
            remapAxes = False
            )
        try:
            p = BatchDesignSpaceProcessor(generatingDesignSpace.path)
            p.useVarlib = True
            orgBatchDebugPref = getDefault("Batch.Debug")
            setDefault("Batch.Debug",True)
            dstPath = generatingDesignSpace.path.replace(".designspace", "_%s.ttf" % time.strftime("%Y%m%d.%H%M", time.gmtime()))
            # chck if all the fonts and layers are present
            p.generateVariationFont(dstPath, autohint=False, fitToExtremes=False, releaseMode=True, glyphOrder=None)
        except:
            self.reportError("Error generating Variable font..", traceback.format_exc())
        finally:
            setDefault("Batch.Debug", orgBatchDebugPref)
            if os.path.exists(dstPath):
                sbx.tools.showInFinder(dstPath)
            bar.close()            
            
    def export_generateUFOCallback(self, sender=None):
        # vanilla callback for exportUFO button
        # make a UFO for the current location as a quick preview
        # this is not the "Add Instance to designspace"
        #@@ compare with generate UFO
        if not self.instanceLocation:
            return
        bar = ProgressBar()
        orgMathPref = self.ds.useVarlib
        if self.previewMathPref == "previewMutatorMath":
            bar.label("Generating preview with MutatorMath…")
            ufoNameMathTag = "MM"
        else:
            bar.label("Generating preview with varlib…")
            ufoNameMathTag = "VL"
        bar.tick()
        neutral = self.ds.getNeutralFont()
        iD = self.data_newInstanceDescriptorForCurrentLocation()
        self.ds._glyphMutators = {}
        self.ds.roundGeometry = self.previewRoundPref
        exportFont = None
        try:
            self.ds.mutedAxisNames = self.mutedAxes
            # set mathmodel
            if self.previewMathPref == "previewMutatorMath":
                self.ds.useVarlib = False
            else:
                self.ds.useVarlib = True
            
            #$$$
            exportFont = self.ds.makeInstance(iD, doRules=True, bend=False)    #self.useAxisMaps)
            self.ds.mutedAxisNames = None
        except:
            print(traceback.format_exc())
            self.reportError("Export UFO failed to generate..", traceback.format_exc())
        finally:
            self.ds.useVarlib = orgMathPref
            bar.close()
        if exportFont is not None:
            if neutral is not None:
                exportFont.lib.update(neutral.lib)
            ufoName = "Preview_%s-%s_%s.ufo" %(iD.familyName, iD.styleName, ufoNameMathTag)
            exportFolder = os.path.join(os.path.dirname(self.ds.path), self.previewFolderName)
            if not os.path.exists(exportFolder):
                os.makedirs(exportFolder)
            exportPath = os.path.join(exportFolder, ufoName)
            # make sure the preview is not open
            for f in AllFonts():
                if f.path == exportPath:
                    # close it
                    f.close()
            exportFont.save(exportPath)
            new = OpenFont(exportPath, showInterface=True)
        else:
            if self.verbose:
                print('nothing exported')
    
    def reportError(self, text, tb):
        if tb.find("Traceback") == 0:
            tb = "Something went wrong. At this time it may not be possible to assign blame and point fingers. So, pardon the traceback posted below, it may be helpful in a report. Thanks!\n\n%s" % tb
        vanilla.dialogs.message("%s" %text, tb, parentWindow=self.w)

    def data_getOpticalSizeValue(self):
        # return a number if we have an actual optical size axis
        # return None otherwise
        optical = None
        for axisRecord in self.ds.axes:
            if axisRecord.tag == "opsz":
                if axisRecord.name not in self.mutedAxes:
                    return self.instanceLocation.get(axisRecord.name)
        return None
    
    def documentHasRulesProblems(self):
        if self.checker:
            for problem in self.checker.problems:
                if problem.category == 7:
                    return True
        return False
        
    def textWindow_updateText(self):
        # only call when there is new input
        if not self.textWindow:
            return
        rawText = self.textWindow.getRawText()
        if rawText is not None:
            if rawText != self.previewTextRaw:
                self.document_setNeedSave(True)
            self.previewTextRaw = rawText
        self.textGlyphNames = sbx.tools.reader.splitText(rawText)
        rawNames = sbx.tools.reader.splitText(rawText)
        if not rawNames:
            return
        new = []
        for n in rawNames:
            if n == "?":
                # replace with current glyph
                new.append(self.currentGlyphName)
            else:
                new.append(n)
        self.textRawNames = new
        if not self.documentHasRulesProblems():
            try:
                self.textGlyphNames = processRules(self.ds.rules, self.instanceLocation, self.textRawNames )
            except:
                self.reportError("Problem while processing rules", traceback.format_exc())    # issue_41 2
        # schedule a rebuild of the kerning here?
        #self.instanceInfoMutator = self.ds.getInfoMutator()
            
        # kerning
        pairs = self.data_pairsFromGlyphNames(self.textGlyphNames)
        if pairs:
            if pairs != self.lastKerningPairs:
                self.lastKerning = None
                self.lastKerningLocation = None
                self.instanceKerningMutator = self.ds.getKerningMutator(pairs=pairs)
                self.lastKerningPairs = pairs
        else:
            self.instanceKerningMutator = None
        
    def textWindow_updateGlyphs(self):
        # the location
        if not self.textWindow:
            return
        if self.ds is None:
            return
        startTextRedraw = time.time()
        opticalSize = self.data_getOpticalSizeValue()
        if self.textWindowMode == "glyph":
            glyphsForPreview = []
            # what happens if we just skip the update
            # if it is not the current glyph
            if not self.currentGlyphName == self.instanceGlyphObject.name:
                return
            try:
                gt = self.contouredInstanceGlyphObject
                # this really needs a layer
                gt.layer = RLayer()
                gt.name = "preview_glyph"
                gt.clearComponents()
                glyphsForPreview.append(gt)
                for s in self.sourceGlyphObjects:
                    if s.contours or s.components:
                        glyphsForPreview.append(s)
                try:
                    self.textWindow.setGlyphs(glyphsForPreview, opticalSize,
                        self.documentUnitsPerEm,
                        self.documentAscender,
                        self.documentDescender,
                        )
                except TypeError:
                    print("textWindow_updateGlyphs error @@ 2", traceback.format_exc())
            except:
                if self.verbose:
                    print("textWindow_updateGlyphs error @@", traceback.format_exc())
            
            return
        #@@@ here we calculate a line of glyphs
        
        kernPairs = []    #self.data_pairsFromGlyphNames(self.textGlyphNames)
        neutral = self.ds.getNeutralFont()
        if neutral is None:
            return
        previewFont = self.data_getPreviewFont(None)    #self.instanceLocation)
        # issue_41
        if not self.documentHasRulesProblems():
            try:
                self.textGlyphNames = processRules(self.ds.rules, self.instanceLocation, self.textRawNames )
            except:
                self.reportError("Problem while processing rules", traceback.format_exc())
        if self.currentlyDragging:
            if not self.textRedrawComps:
                compedNames = self.textRedrawComps = sbx.tools.followComponents(neutral, self.textGlyphNames)
                random.shuffle(compedNames)
            else:
                self.textRedrawComps = self.textRedrawComps[2:] + self.textRedrawComps[:2]
                compedNames = self.textRedrawComps
        else:
            compedNames = self.textRedrawComps = sbx.tools.followComponents(neutral, self.textGlyphNames)
        allGlyphsDone = False
        if self.currentlyDragging:
            randomSubset = []
            if len(compedNames) < 4:
                randomSubset = compedNames
            else:
                randomSubset = compedNames[:2]
        else:
            randomSubset = compedNames
            self.textRedrawGlyphHistory = {}
            allGlyphsDone = True
        for name in randomSubset:
            # here we should ask for new mutators
            self.data_prepareGlyph(name)
        generatingStart = time.time()
        for name in randomSubset:
            storedMutatorOk, storedItems = self.storage_mainPreviews.getStoredObject(name)
            if storedMutatorOk:
                _masters, _sourceGlyphObjects, _currentMutator, _componentMutators = storedItems
                if not name in previewFont:
                    previewFont.newGlyph(name)
                previewFont[name].clear()
                if _currentMutator is not None:
                    try:
                        thisInstance = _currentMutator.makeInstance(self.instanceLocation, bend=self.useAxisMaps)
                        previewFont[name].fromMathGlyph(thisInstance)
                    except (IndexError, AssertionError):
                        previewFont[name].appendGlyph(self.placeholderGlyph)
                        previewFont[name].width = self.placeholderGlyph.width
                else:
                    previewFont[name].appendGlyph(self.placeholderGlyph)                        
        generatingDuration = time.time() - generatingStart

        if not allGlyphsDone:
            self.timerSchedule_textWindowRedraw()
        else:
            self.textRedrawComps = []
            self.textRedrawOrder = []
            # do the kerning now?
            if self.lastKerningLocation != self.instanceLocation:
                if self.instanceKerningMutator:
                    try:
                        thisKerning = self.instanceKerningMutator.makeInstance(self.instanceLocation, bend=self.useAxisMaps)
                        thisKerning.extractKerning(previewFont)
                    except:
                        self.reportError("Problem with kerning mutator", traceback.format_exc())
                if self.instanceInfoMutator:
                    try:
                        instanceInfo = self.instanceInfoMutator.makeInstance(self.instanceLocation, bend=self.useAxisMaps)
                        if instanceInfo:
                            previewFont.info.fromMathInfo(instanceInfo, guidelines=False)
                    except:
                        self.reportError("Problem with info mutator", traceback.format_exc())

        # add preview fonts
        glyphsForPreview = []
        for n in self.textGlyphNames:
            if n is None:
                continue
            if n in previewFont:
                glyphsForPreview.append(previewFont[n])
        #previewFont.save("previewFont.ufo")
        self.textWindow.w.mline.setFont(previewFont)
        self.textWindow.w.mline.set(glyphsForPreview)
        if opticalSize is not None:
            self.textWindow.setOpticalPointSize(opticalSize)
        
    def data_calculateGlyphsForLine(self, font, location, names):
        #print('data_calculateGlyphsForLine', font, names)
        #@@@ here we calculate a line of glyphs
        if self.ds is None:
            return
        kernPairs = []    #self.data_pairsFromGlyphNames(self.textGlyphNames)
        for name in names:
            # here we should ask for new mutators
            #print('data_calculateGlyphsForLine preparing', name)
            self.data_prepareGlyph(name)
        for name in names:
            storedMutatorOk, storedItems = self.storage_mainPreviews.getStoredObject(name)
            #print("data_calculateGlyphsForLine storedMutatorOk", storedMutatorOk, name)
            if storedMutatorOk:
                _masters, _sourceGlyphObjects, _currentMutator, _componentMutators = storedItems
                if not name in font:
                    font.newGlyph(name)
                font[name].clear()
                #print("data_calculateGlyphsForLine _currentMutator", _currentMutator)
                if _currentMutator is not None:
                    try:
                        thisInstance = _currentMutator.makeInstance(location, bend=self.useAxisMaps)
                        font[name].clear()
                        font[name].fromMathGlyph(thisInstance)
                        font[name].width = thisInstance.width
                        #print("data_calculateGlyphsForLine instance", font[name])
                    except (IndexError, AssertionError):
                        font[name].appendGlyph(self.placeholderGlyph)
                        font[name].width = self.placeholderGlyph.width
                else:
                    font[name].appendGlyph(self.placeholderGlyph)                        

        # if self.instanceKerningMutator:
        #     try:
        #         thisKerning = self.instanceKerningMutator.makeInstance(self.instanceLocation, bend=self.useAxisMaps)
        #         thisKerning.extractKerning(font)
        #     except:
        #         self.reportError("Problem with kerning mutator", traceback.format_exc())
        if self.instanceInfoMutator:
            try:
                instanceInfo = self.instanceInfoMutator.makeInstance(self.instanceLocation, bend=self.useAxisMaps)
                if instanceInfo:
                    font.info.fromMathInfo(instanceInfo, guidelines=False)
            except:
                self.reportError("Problem with info mutator", traceback.format_exc())
        #print("data_calculateGlyphsForLine", id(font), font.keys())
        return font

    def glyphWindow_drawPreviewCallbackForNavigatorTool(self, thisScale):
        # weird callback - this does the drawing for the navigatorTool.drawPreview
        self.data_instanceLocationChanged(dragging=False)
        g = CurrentGlyph()
        b = g.bounds
        glyphName = g.name
        storedMutatorOk, storedItems = self.storage_mainPreviews.getStoredObject(glyphName)
        if not storedMutatorOk:
            self.data_prepareGlyph(glyphName)
            storedMutatorOk, storedItems = self.storage_mainPreviews.getStoredObject(glyphName)
        if not storedMutatorOk:
            return
        _masters, _sourceGlyphObjects, _currentMutator, _componentMutators = storedItems
        try:
            thisInstance = _currentMutator.makeInstance(self.instanceLocation, bend=self.useAxisMaps)
        except (IndexError, ):
            # something is failing in the interpolation
            # can't draw
            return
        ctx.save()
        yMin = 0
        if b:
            xMin, yMin, xMax, yMax = b
            ctx.fill(*self._glyphViewPreviewBackgroundColor)
            ctx.stroke(None)
            ctx.rect(xMin-100, yMin-100, xMax-xMin+200, yMax-yMin+200)
        yMin = -200    #min(yMin, -200)
        ctx.newPath()
        p = sbx.tools.pens.PreviewPen({})
        ap = AreaPen({})
        centerI = thisInstance.width * .5
        dx = g.width * .5 - thisInstance.width * .5
        ctx.translate(dx, 0)
        ctx.fill(*self._glyphViewPreviewFillColor)
        thisInstance.draw(p)
        thisInstance.draw(ap)
        self.instanceGlyphObject.fromMathGlyph(thisInstance)
        self.instanceGlyphObject.width = thisInstance.width
        self.instanceGlyphObject.name = thisInstance.name
        self.contouredInstanceGlyphObject.appendGlyph(self.instanceGlyphObject)
        self.contouredInstanceGlyphObject.width = thisInstance.width
        if self.previewRoundPref:
            self.instanceGlyphObject.round()

        def drawComps(self, pen, areaPen, baseGlyph, transformation, _componentMutators):
            # draw nested components
            compMut = _componentMutators[baseGlyph]
            if compMut is None:
                return
            compInstance = compMut.makeInstance(self.instanceLocation, bend=self.useAxisMaps)
            tp = TransformPen(outPen=p, transformation=transformation)
            tpa = TransformPen(outPen=ap, transformation=transformation)
            compInstance.draw(tp)
            compInstance.draw(tpa)
            for compInstanceComp in compInstance.components:
                if type(compInstanceComp) is dict:
                        compInstanceCompbaseGlyph = compInstanceComp['baseGlyph']
                        compInstanceComptransformation = compInstanceComp['transformation']
                else:
                        compInstanceCompbaseGlyph = compInstance.baseGlyph
                        compInstanceComptransformation = compInstanceComp.transformation
                t1 = Transform(*transformation)
                t2 = Transform(*compInstanceComptransformation)
                t3 = t1.transform(t2)
                drawComps(self, p, ap, compInstanceCompbaseGlyph, t3, _componentMutators)
        for comp in self.instanceGlyphObject.components:
            drawComps(self, p, ap, comp.baseGlyph, comp.transformation, _componentMutators)
        
        ctx.fill(0)
        ctx.drawPath()
        
        if thisScale > 0:
            if self.instanceGlyphObject is not None:                
                # thisScale can be 0 when zoomed out
                ts = self._captionTextSize*thisScale
                ctx.font(self._captionTextFont)
                ctx.fontSize(ts)
                widthInfo = marginInfo = ""
                try:
                    if self.instanceGlyphObject.leftMargin is not None and self.instanceGlyphObject.rightMargin is not None:
                        marginInfo = "\n|< %3.3f %3.3f >|" %(self.instanceGlyphObject.leftMargin, self.instanceGlyphObject.rightMargin)
                except TypeError:
                    pass
                if self.instanceGlyphObject.width is not None:
                    widthInfo = "\nwidth %3.3f" % (self.instanceGlyphObject.width)
                if g.width != 0:
                    widthInfo += " change: %3.3f" % (self.instanceGlyphObject.width / g.width)
                captionText = "location: %s\n%s\n%s\narea: %3.3f" % (Location(self.instanceLocation).asString(), os.path.basename(self.ds.path), self.previewMathPref, ap.value)
                captionText += marginInfo
                captionText += widthInfo
                captionPos = (0*thisScale, yMin*thisScale)
                tw, th = ctx.textSize(captionText)
                ctx.fill(1,1,1, 0.6)
                ctx.rect(captionPos[0]-2, captionPos[1]-2, tw+4, th+4)
                ctx.fill(0)
                ctx.text(captionText,captionPos)
                ctx.fill(None)
                ctx.stroke(0)
                ctx.strokeWidth(1*thisScale)
                ctx.line((0, 0), (0, -self.marginIndicatorLineHeight*thisScale))
                ctx.line((self.instanceGlyphObject.width, 0), (self.instanceGlyphObject.width, -self.marginIndicatorLineHeight*thisScale))
                ctx.stroke(None)
        p = sbx.tools.pens.PreviewPen(g.font)
        ctx.translate(thisInstance.width, 0)
        g.draw(p)
        ctx.fill(0)
        ctx.drawPath()        
        ctx.restore()
        
    def spaceWindow_dragEventCallback(self, notification):
        # callback for observer for skateboardSpaceWindowEvent
        if self.instanceLocation is None:
            return
        newLoc = notification['data']['location']
        #print("----+", newLoc, self.canExtrapolate)
        mouse = notification['data'].get("mouse")
        self.snapped = notification.get('snapped', [])
        if mouse == "mouseDragged" or mouse == "mouseDown":
            self.drawDetails = False
            for axisName, value in newLoc.items():
                self.instanceLocation[axisName] = value
            self.data_instanceLocationChanged(dragging=True)
        elif mouse == "mouseUp":
            self.drawDetails = True
            self.data_instanceLocationChanged(dragging=False)
    
    def textWindow_changedSettingsCallback(self, notification):
        # text window settings change observer callback
        self.textWindowMode = notification['mode']
        self.timerSchedule_textWindowRedraw(now=True)
        
    def navigatorTool_clickCallback(self, notification):
        # navigator click callback
        # callback for navigatorToolEvent.mouseup
        modFlags = AppKit.NSEvent.modifierFlags()
        if notification['notificationName'] == 'navigatorToolEvent.mouseup':
            self.currentlyDragging = False
        elif notification['notificationName'] == 'navigatorToolEvent.mousedown':
            self.currentlyDragging = True
            self.locationsList_clearSelection()
        
    def navigatorTool_activeCallback(self, notification):
        # navigator active callback
        self.drawingWithToolActivePref = True
        # we should also set the previewcallback here
        if self.navigatorTool is not None:
            self.navigatorTool.RFPreviewCallback = self.glyphWindow_drawPreviewCallbackForNavigatorTool

    def navigatorTool_inactiveCallback(self, notification):
        # navigator inactive callback
        self.drawingWithToolActivePref = False

    def mainWindow_proposePDFName(self):
        pdfName = [self.currentGlyph.name]
        folder = os.path.dirname(self.ds.path)
        for name in self.ds.getAxisOrder():
            pdfName.append("%s_%2.2f"%(name, self.instanceLocation[name]))
        return os.path.join(folder, "_".join(pdfName) + ".pdf")
        
    def navigatorTool_menuCallback(self, notification):
        # navigator menu callback
        refresh = False
        textRefresh = False
        cmd = notification['item']
        if cmd == "document_currentLayerToSupport":
            self.document_currentLayerToSupport()
            refresh = True
        if cmd == "newInstance":
            self.document_newInstanceCallback()
        if cmd == "storeLocation":
            self.navigatorMenu_storeCurrentLocation()
        #elif cmd == "makeUFO":
        #    self.makeUFOCallback(None)
        elif cmd == "savePDF":
            path = self.mainWindow_proposePDFName()
            path = PutFile(message="Save this PDF to:", fileName=os.path.basename(path))
            GlyphWindowToPDF(path)
        elif cmd == "newSupport":
            #@@
            self.navigatorMenu_previewToSupportCallback(None)
        elif cmd == "copyPreview":
            self.navigatorToolMenu_copyPreviewGlyphCallback(None, target="clipboard")
        elif cmd == "copyPreviewToLayer":
            self.navigatorToolMenu_copyPreviewGlyphCallback(None, target="currentLayer")
        #elif cmd == "trackSelection":
        #    self.navigatorToolMenu_trackSelectionCallback(None)
        elif cmd == "openText":
            self.mainWindow_openTextWindow()
        elif cmd == "openSpace":
            self.mainWindow_openSpaceWindow()
        elif cmd == "showSkateboard":
            self.mainWindow_show()
        elif cmd == "showSparks":
            self.showSparksPref = not self.showSparksPref
            refresh = True
        elif cmd in ['previewLeft', 'previewOrigin', 'previewCentered', 'previewRight']:
            if cmd != self.previewLocationPref:
                refresh = True
            self.previewLocationPref = cmd
        elif cmd in ['previewSolid', 'previewWireframe']:
            refresh = False
            if cmd != self.previewDrawPref:
                refresh = True
                textRefresh = True
            self.previewDrawPref = cmd
        elif cmd in ["previewVarLib", "previewMutatorMath"]:
            refresh = True
            if cmd != self.previewMathPref:
                refresh = True
                textRefresh = True
                # also clear cached mutators
                self.storage_mainPreviews.reset()
                self.storage_supportPreviews.reset()
                self.lastSupports = []
            self.previewMathPref = cmd
        #if cmd in ['useAxisMaps']:
        #    self.useAxisMaps = not self.useAxisMaps
        #    refresh = True
        if cmd in ['previewRound']:
            self.previewRoundPref = not self.previewRoundPref
            self.ds.roundGeometry = self.previewRoundPref
            refresh = True
            textRefresh = True
        if cmd == "previewElsewhere":
            self._drawPreviewsElsewhere = not self._drawPreviewsElsewhere
        if cmd == "previewWithNavigator":
            self._drawPreviewWithNavigator = not self._drawPreviewWithNavigator
            refresh = True
        if cmd == "copyCurrentToAllMasters":
            self.mainWindow_copyCurrentGlyphToAllMasters()
        if refresh:
            self.data_prepareGlyph(self.currentGlyphName, force=True)
            UpdateCurrentGlyphView()
            for window in AllGlyphWindows():
                window.editGlyphView.refresh()
        if textRefresh:
            if self.textWindow:
                self.timerSchedule_textWindowRedraw(now=True)
        
    def navigatorTool_data_locationContextCallback(self, notification):
        # navigator callback callbackForSkateboard
        callbackFromNavigatorTool = notification.get("callback")
        data = notification.get('data')
        if data is not None:
            # the navigator menu can send this request without the data
            # if it needs an update for the state of the menus. 
            self.snapped = notification.get("snapped")
            draggedLocation = notification['data']
            if draggedLocation is not None and self.instanceLocation is not None:
                self.instanceLocation.update(draggedLocation)
            self.data_instanceLocationChanged(dragging=True)
        if callbackFromNavigatorTool:
            callbackFromNavigatorTool(dict(
                location=self.instanceLocation,
                context=self.data_locationContext(self.instanceLocation),
                interactionSources = self.interactionSources,
                previewDrawPref = self.previewDrawPref,
                previewLocationPref = self.previewLocationPref,
                previewMathPref = self.previewMathPref,
                previewRoundPref = self.previewRoundPref,
                previewWithNavigator = self._drawPreviewWithNavigator,
                previewElsewhere = self._drawPreviewsElsewhere,
                useAxisMaps = self.useAxisMaps,
                previewCallback = self.glyphWindow_drawPreviewCallbackForNavigatorTool,
                showSparksPref = self.showSparksPref,
                snapValues = self.snapValues
                ))
        
    def navigatorTool_mouseDraggedCallback(self, notification):
        self.drawDetails = False

    def navigatorTool_mouseUpCallback(self, notification):
        self.timerSchedule_navigatorTool_mouseUp_draggingDone()
    
    # handle glyph window menu for selected points
    # check selection
    # store selection
    
    def glyphWindowMenuCallback_trackSelection(self, sender):
        self.trackedSelection = True
        # vanilla callback
        # store the indexes of the current selection
        selectedSingles = None
        selectedPairs = None
        if self.currentGlyph is None:
            return
        if self.ds is None:
            self.currentSelection = None
            self.currentSelectedClusters = {}
            self.trackedClusters = {}
            return
        key = (self.ds.path, self.currentGlyph.name)
        # how to handle a new selection or de-selection
        if key in self.currentSelectedClusters:
            selectedSingles, selectedPairs = self.currentSelectedClusters[key]
            self.trackedClusters[key] = selectedSingles, selectedPairs

    def glyphWindowMenuCallback_clearSelection(self, sender):
        self.trackedSelection = False
        if self.currentGlyph is None:
            return
        if self.ds is None:
            self.currentSelection = None
            self.currentSelectedClusters = {}
            self.trackedClusters = {}
            return
        key = (self.ds.path, self.currentGlyph.name)
        if key in self.trackedClusters:
            if key in self.trackedClusters:
                del self.trackedClusters[key]
                # request redraw
                self.timerSchedule_navigatorTool_mouseUp_draggingDone()
        
    def glyphWindow_glyphAdditionContextualMenuItems(self, notification):
        if self.currentGlyph is None:
            return
        if self.ds is None:
            return
        items = []
        key = (self.ds.path, self.currentGlyph.name)
        if key in self.currentSelectedClusters:
            items.append(dict(title="Track Selection in %s" % self.currentGlyph.name, callback=self.glyphWindowMenuCallback_trackSelection))
        if key in self.trackedClusters:
            items.append(dict(title="Clear Selection in %s" % self.currentGlyph.name, callback=self.glyphWindowMenuCallback_clearSelection))
        if items:
            myItem = dict(title="Skateboard", items=items)
            notification['additionContextualMenuItems'].append(myItem)
        
    def glyphWindow_drawBackgroundWrapper(self, notification):
        # observer callback
        # wrapper
        ctx.save()
        try:
            self.glyphWindow_drawBackgroundHandler(notification)
        finally:
            ctx.restore()
    
    def glyphWindow_drawForegroundWrapper(self, notification):
        if not self._lastDrawingValues:
            return
        self._lastDrawingValues = None
        # # 5.8 selected markers
        # if self.drawingWithToolActivePref and not dd.drawingIsHeavy:
        #     save()
        #     try:
        #         self.glyphWindow_presentSelectedMarkers(dd)
        #     finally:
        #         restore()
        # # 5.9 colinear points
        # if self._trackSelection and dd.selectedSegments:
        #     save()
        #     try:
        #         self.glyphWindow_presentSelectedSegments(dd)
        #     finally:
        #         restore()
    
    def glyphWindow_prepareDrawingState(self):
        # move all drawing prep here
        dd = Container()
        dd.glyph = None
        dd.scale = 1
        dd.view = None
        dd.shouldDrawPoint = None
        dd.shouldDrawRect = None
        dd.notDrawn = 0
        dd.componentMutators = []
        dd.eventTime = time.time()
        dd.drawingIsGood = True
        dd.drawingIsHeavy = False
        dd.thisInstance = None
        dd.instanceGlyphObject_noSupport = None
        dd.offsetForInstanceGlyph = 0
        dd.offsetForInstanceGlyph_noSupport = 0
        dd.currentMutator = None
        dd.currentSupportMutator = None    # mutator for this glyph, but without the current support master
        dd.factors = []
        dd.collectors = []
        dd.trackedClusters = {}
        dd.selectedClusters = {}

        self.instanceGlyphObject.clear()     # Somehow the masters and the instanceglyphObject go out of sync
        self.contouredInstanceGlyphObject.clear()

        dd.isSupportGlyph, dd.supportLocation = self.currentGlyphIsSupport()
        if self.pointLoad > self._maxPoints:
            dd.drawingIsHeavy = True
        dd.selectedClusters = {}, {}
        if not dd.drawingIsHeavy:
            if self.currentGlyph is not None:
                key = (self.ds.path, self.currentGlyph.name)
                if key in self.trackedClusters:
                    dd.trackedClusters = self.trackedClusters[key]
                elif key in self.currentSelectedClusters:
                    dd.selectedClusters = self.currentSelectedClusters[key]
        return dd

    def glyphWindow_drawBackgroundHandler(self, notification):
        # observer callback
        # here we do all the drawing in the glyph background.
        viewKey = id(notification['view'])
        # 0 // return as quick as we can:
        # no document
        if self.ds is None:
            return 
        # no glyph
        if notification['glyph'] is None:
            return
        if notification['glyph'] is not None:
            # not a glyph of this designspace
            if not notification['glyph'].font.path in self.openFonts:
                self._lastDrawingValues = None
                return
        # preference about drawing now?
        if self.drawingWithToolActivePref:
            if not self._drawPreviewWithNavigator:
                self._lastDrawingValues = None
                return
        else:
            if not self._drawPreviewsElsewhere:
                self._lastDrawingValues = None
                return
        # 1 // start collecting some values that we need in a generic object.
        dd = self.glyphWindow_prepareDrawingState()
        dd.glyph = notification['glyph']  # would be useful to know the glyph earlier
        dd.scale = notification['scale']
        dd.view = notification['view']    # once we start drawing there will be a view
        dd.selectedSources = self.pane_doc.l.getSelection()
        dd.anchorMarkerSize = 4 * dd.scale
        dd.shouldDrawPoint = dd.view.shouldDrawPoint
        dd.shouldDrawRect = dd.view.shouldDrawRect
        self.sourceGlyphObjects = []

        # 2 // get mutators for the instance and the components
        storedMutatorOk, storedItems = self.storage_mainPreviews.getStoredObject(dd.glyph.name)
        if not storedMutatorOk:
            # it should be ready here
            dd.drawingIsGood = False
        else:
            _masters, self.sourceGlyphObjects, dd.currentMutator, dd.componentMutators = storedItems
        if dd.currentMutator is None:
            dd.drawingIsGood = False
        if dd.drawingIsGood:
            if dd.isSupportGlyph and dd.supportLocation is not None:
                storedSupportOk, storedSupportItems = self.storage_supportPreviews.getStoredObject(dd.glyph.name)
                if not storedSupportOk:
                    try:
                        _m, _sg, dd.currentSupportMutator, _cM = self.data_makeGlyphMutator(dd.glyph.name, ignoreLocations=[dd.supportLocation])
                        t = self.storage_supportPreviews.storeObject(dd.glyph.name, (_m, _sg, dd.currentSupportMutator, _cM))
                    except (ValueError):
                        self.reportError("Error processing masters for support mutator", traceback.format_exc())
                else:
                    _m, _sg, dd.currentSupportMutator, _cM = storedSupportItems
        dd.scaledMarkerSize = self._markerSize * dd.scale
        dd.scaledErrorReportMarkerSize = 2 * dd.scaledMarkerSize
        dd.thisStrokeWidth = 0.4 * dd.scale
        dd.ascender = dd.glyph.font.info.ascender
        dd.descender = dd.glyph.font.info.descender
        # 3 // make instances for the components
        dd.componentStorage = {}
        for compName in dd.componentMutators:
            cmut = dd.componentMutators[compName]
            if cmut:
                try:
                    cg = cmut.makeInstance(self.instanceLocation, bend=self.useAxisMaps)
                except IndexError:
                    continue
                if self.previewRoundPref:
                    cg = cg.round()
                dd.componentStorage[compName] = cg
                pp = RGlyph().fromMathGlyph(cg)
                self.contouredInstanceGlyphObject.appendGlyph(pp)
                
        # 4 // collect centered points from the masters
        dd.debugCollectorNames = []
        for i, m in enumerate(self.sourceGlyphObjects):
            dx = 0
            if self.previewLocationPref == "previewCentered":
                dx = .5*dd.glyph.width - .5 * m.width
            elif self.previewLocationPref == "previewRight":            
                dx = dd.glyph.width - m.width
            collector = sbx.tools.pens.CollectorPen({})
            collector.tag = i
            collector.setOffset(dx,0)
            m.draw(collector)
            dd.collectors.append(collector)
            dd.debugCollectorNames.append(m.name)

        # 4.1 // collect the centered points from the instance
        if dd.currentMutator:
            collector = sbx.tools.pens.CollectorPen({})
            dd.factors = dd.currentMutator.getFactors(self.instanceLocation)
            #if self.previewMathPref == "previewMutatorMath":
            #    self._lastFactors = [a[0] for a in dd.factors]
            #else:
            #    self._lastFactors = dd.factors
            dd.thisInstance = None
            try:
                dd.thisInstance = dd.currentMutator.makeInstance(self.instanceLocation, bend=self.useAxisMaps)
            except (IndexError, AssertionError):
                dd.drawingIsGood = False
        
        if hasattr(dd.currentMutator, 'getSupports'):
            self.lastSupports = dd.currentMutator.getSupports()
        else:
            self.lastSupports = []
            
        # 4.2 // can't draw?
        if not dd.drawingIsGood:
            ctx.save()
            try:
                if self._problemResponseScheme == "show" or self._problemResponseScheme is None:
                    self.glyphWindow_presentDrawingProblem(dd, orientation="bottom")
                self.data_logFailingGlyph(dd.glyph.name)
            finally:
                ctx.restore()
            # save()
            # try:
            #     self.glyphWindow_presentDrawingProblem(dd, orientation="left")
            #     self.data_logFailingGlyph(dd.glyph.name)
            # finally:
            #     restore()
            self._lastDrawingValues = None
            return

        # 4.3 // make an instance for the ignore-support if relevant
        self.instanceGlyphObject_noSupport.clear()
        self.hasNoSupportGlyph = False
        if dd.currentSupportMutator is not None:
            mg = dd.currentSupportMutator.makeInstance(self.instanceLocation, bend=self.useAxisMaps)
            if mg is not None:
                self.instanceGlyphObject_noSupport.fromMathGlyph(mg)
                self.instanceGlyphObject_noSupport.width = mg.width
                self.instanceGlyphObject_noSupport.name = dd.glyph.name
                if self.previewRoundPref:
                    self.instanceGlyphObject_noSupport.round()
                self.hasNoSupportGlyph = True
            
        # 5 // process the instance
        self.instanceGlyphObject.fromMathGlyph(dd.thisInstance)
        self.instanceGlyphObject.width = dd.thisInstance.width
        self.instanceGlyphObject.name = dd.glyph.name
        self.contouredInstanceGlyphObject.appendGlyph(self.instanceGlyphObject)
        self.contouredInstanceGlyphObject.width = dd.thisInstance.width
        if self.previewRoundPref:
            self.instanceGlyphObject.round()
        dx = 0
        if self.previewLocationPref == "previewCentered":
            dd.offsetForInstanceGlyph = dx = .5*dd.glyph.width - .5 * self.instanceGlyphObject.width
            if self.instanceGlyphObject_noSupport is not None:
                dd.offsetForInstanceGlyph_noSupport = .5 * self.instanceGlyphObject_noSupport.width
        elif self.previewLocationPref == "previewRight":
            dd.offsetForInstanceGlyph = dx = dd.glyph.width - dd.thisInstance.width
            if self.instanceGlyphObject_noSupport is not None:
                dd.offsetForInstanceGlyph_noSupport = dd.glyph.width - self.instanceGlyphObject_noSupport.width
        # 5.1 // save # s1
        ctx.save()
        # we're centered
        ctx.translate(dx,0)

        # 5.5 the instance
        ctx.save()
        try:
            self.glyphWindow_presentInstance(dd)
        finally:
            ctx.restore()

        # 5.2 measurements
        if not dd.drawingIsHeavy:
            ctx.save()
            try:
                self.glyphWindow_presentMeasurements(dd)
            finally:
                ctx.restore()
        # 5.3 metrics info
        if self.drawingWithToolActivePref:
            ctx.save()
            try:
                self.glyphWindow_presentMarginInfo(dd)
            finally:
                ctx.restore()
        # 5.4 anchors
        if not dd.drawingIsHeavy:
            ctx.save()
            try:
                self.glyphWindow_presentAnchors(dd)     
            finally:
                ctx.restore()

        # 5.5.1 the instance-nosupport
        if self.instanceGlyphObject_noSupport is not None and self.hasNoSupportGlyph:
            ctx.save()
            try:
                self.glyphWindow_glyphWindow_presentDeltas_noSupport(dd)
            finally:
                ctx.restore()

        # 5.5.2 colinear points
        if self._trackSelection and (dd.selectedClusters or dd.trackedClusters):
            ctx.save()
            try:
                self.glyphWindow_presentSelectedClusters(dd)
            finally:
                ctx.restore()
        
        # 5.6 // restore # s1
        ctx.restore()
        
        # 5.7 the sparks
        # these need to be drawn at the normal (0,0), not relative to dx.
        if self.showSparksPref and self.drawDetails and self.drawingWithToolActivePref and not dd.drawingIsHeavy:
            ctx.save()
            try:
                self.glyphWindow_presentSparks(dd)
            except:
                if self.verbose:
                    print("Skateboard: drawing error for presentSparks", traceback.format_exc())
            finally:
                ctx.restore()
        # 5.8 selected markers
        if self.drawingWithToolActivePref and not dd.drawingIsHeavy and not self.currentlyDragging and self.showSparksPref:
            ctx.save()
            try:
                self.glyphWindow_presentSelectedMarkers(dd)
            finally:
                ctx.restore()
        # if self._trackSelection and dd.selectedSegments:
        #     save()
        #     try:
        #         self.glyphWindow_presentSelectedSegments(dd)
        #     finally:
        #         restore()
        # 6.0 store that this glyph data_checkGlyphWorks
        # .. what about the components?
        self.data_logWorkingGlyph(dd.glyph.name)
        # 6.0 return the collected values
        self._lastDrawingValues = dd
        return dd
    
    def _draw_bPoint(self, bpt, dd):
        # draw a marker on a bpoint
        ms = 2
        _in, _ankr, _out = bpt.bcpIn, bpt.anchor, bpt.bcpOut
        if _in != (0,0):
            ctx.oval(_ankr[0]+_in[0]-ms*dd.scaledMarkerSize , _ankr[1]+_in[1]-ms*dd.scaledMarkerSize , 2*ms*dd.scaledMarkerSize, 2*ms*dd.scaledMarkerSize)
        if _out != (0,0):
            ctx.oval(_ankr[0]+_out[0]-ms*dd.scaledMarkerSize , _ankr[1]+_out[1]-ms*dd.scaledMarkerSize , 2*ms*dd.scaledMarkerSize, 2*ms*dd.scaledMarkerSize)
        ctx.oval(_ankr[0]-ms*dd.scaledMarkerSize , _ankr[1]-ms*dd.scaledMarkerSize , 2*ms*dd.scaledMarkerSize, 2*ms*dd.scaledMarkerSize)
    
    def _pointsToColor(self, pts, alpha=1):
        aa = sbx.tools.plot(pts)
        if aa is None:
            return (0,1,.5,alpha)
        poly = abs(aa)
        isCol = poly < self.colinearTolerance
        factor = 1
        if not isCol:
            if poly > 0:
                factor = self.colinearTolerance / poly
            else:
                factor = 0
        if not isCol:
            col = (1-factor,factor,0, alpha)
        else:
            col = (0,1,0, alpha)
        return factor, col

    def glyphWindow_presentSelectedClusters(self, dd):
        tracked = False
        if dd.trackedClusters:
            singles, pairs = dd.trackedClusters
            tracked = True
        else:
            singles, pairs = dd.selectedClusters
        sw = 2 * dd.scale
        if tracked:
            sw *= 2
        ctx.save()
        ctx.strokeWidth(sw)
        glyphKinkScore = 0
        for ci, s in singles.items():
            try:
                bpts = self.instanceGlyphObject.contours[ci].bPoints
            except IndexError:
                # some sort of compatibility issue
                continue
            for bpt in s:
                # draw singles here
                if not bpt.index < len(bpts):
                    continue
                ia = bpts[bpt.index]
                iaInAbs = ia.anchor[0]+ia.bcpIn[0], ia.anchor[1]+ia.bcpIn[1]
                iaOutAbs = ia.anchor[0]+ia.bcpOut[0], ia.anchor[1]+ia.bcpOut[1]
                factor, col = self._pointsToColor([ia.anchor, iaInAbs, iaOutAbs])
                ctx.stroke(None)
                ctx.fill(*col)
                self._draw_bPoint(ia, dd)
                ctx.stroke(*col)
                ctx.fill(None)
                ctx.line((ia.anchor[0], ia.anchor[1]), (ia.anchor[0]+ia.bcpIn[0], ia.anchor[1]+ia.bcpIn[1]))
                ctx.line((ia.anchor[0], ia.anchor[1]), (ia.anchor[0]+ia.bcpOut[0], ia.anchor[1]+ia.bcpOut[1]))
        ctx.restore()

        ctx.save()
        ctx.strokeWidth(sw)
        for ci, pairs in pairs.items():
            try:
                bpts = self.instanceGlyphObject.contours[ci].bPoints
            except IndexError:
                # some sort of compatibility issue
                print("Skateboard: compatibility issue", dd.thisInstance.name)
                continue
            for a, b in pairs:
                if a.index > len(bpts):
                    continue
                if b.index > len(bpts):
                    continue
                try:
                    ia = bpts[a.index]
                    ib = bpts[b.index]
                except IndexError:
                    # some sort of compatibility issue
                    print("Skateboard: compatibility issue", dd.thisInstance.name)
                    continue
                iaInAbs = ia.anchor[0]+ia.bcpIn[0], ia.anchor[1]+ia.bcpIn[1]
                ibOutAbs = ib.anchor[0]+ib.bcpOut[0], ib.anchor[1]+ib.bcpOut[1]
                factor, col = self._pointsToColor([iaInAbs, ia.anchor, ib.anchor, ibOutAbs])
                ctx.stroke(None)
                ctx.fill(*col)
                self._draw_bPoint(ia, dd)
                self._draw_bPoint(ib, dd)
                ctx.stroke(*col)
                ctx.fill(None)
                ctx.line((ia.anchor[0], ia.anchor[1]), (ia.anchor[0]+ia.bcpIn[0], ia.anchor[1]+ia.bcpIn[1]))
                ctx.line((ib.anchor[0], ib.anchor[1]), (ib.anchor[0]+ib.bcpOut[0], ib.anchor[1]+ib.bcpOut[1]))
                ctx.line((ia.anchor[0], ia.anchor[1]), (ib.anchor[0], ib.anchor[1]))
                ctx.stroke(None)
        ctx.restore()
        
    def glyphWindow_presentSelectedMarkers(self, dd):
        # draw markers for the currently selected masters
        for i, c in enumerate(dd.collectors):
            thisMarkerSize = dd.scaledMarkerSize * 1.5
            if c.tag in dd.selectedSources:
                # index matches one of the selected sources in the list
                # make it bigger
                thisMarkerSize = self._selectedMarkerSizeScale * dd.scaledMarkerSize
            for p in c.onCurves:
                if i == 0:
                    ctx.fill(0,.5,1)
                else:
                    if c.tag in dd.selectedSources:
                        ctx.fill(*self._offCurveColor)
                    else:
                        ctx.fill(*self._sparkStrokeColor)
                        ctx.stroke(None)
                ctx.oval(p[0]-.5*thisMarkerSize , p[1]-.5*thisMarkerSize , thisMarkerSize, thisMarkerSize)

    def glyphWindow_presentSparks(self, dd):
        # draw the points we want to see when nothing is selected
        ctx.stroke(None)
        ctx.fill(None)
        for i, c in enumerate(dd.collectors[:-1]):
            done = []
            thisMarkerSize = dd.scaledMarkerSize
            fromSelectedSource = False
            dd.thisStrokeWidth = 0.4 * dd.scale
            if i in dd.selectedSources:
                # index matches one of the selected sources in the list
                # make it bigger
                fromSelectedSource = True
            sw = 0.4 * dd.scale
            
            for j, p in enumerate(c.onCurves):
                ctx.strokeWidth(dd.thisStrokeWidth)
                drawAsNeedle = False
                if p in done: continue
                if i > len(dd.factors)-1:
                    continue
                if dd.factors[i] == 0:
                    # draw lines differently if the factor is 0
                    drawAsNeedle = False
                try:
                    end = dd.collectors[-1].onCurves[j]
                except IndexError:
                    print("glyphWindow_presentSparks error", dd.glyph.name, len(dd.collectors), dd.collectors, j)
                    pprint(debugCollectorNames)
                    print("#"*30)
                    print("c.onCurves", len(c.onCurves))
                    pprint(c.onCurves)
                    print("#"*30)
                    print("dd.collectors[-1].onCurves", len(dd.collectors[-1].onCurves))
                    pprint(dd.collectors[-1].onCurves)
                    print("#"*30)
                if i == 0:
                    ctx.stroke(0,.4,.8)
                else:
                    ctx.stroke(*self._sparkStrokeColor)
                    #ctx.stroke(0,0,0,0.25)
                if fromSelectedSource:
                    ctx.lineDash(None)               
                else:       
                    ctx.lineDash(3*dd.scale,3*dd.scale)
                if dd.shouldDrawPoint(p) or dd.shouldDrawPoint(end):
                    if drawAsNeedle:
                        needle(p, end, 2*sw , .5*sw, 4)
                    else:
                        ctx.line(p, end)
                else:
                    dd.notDrawn += 1
                done.append(p)
            done = []
            for j, p in enumerate(dd.collectors[-1].offCurves):
                if p in done: continue
                end = dd.collectors[-1].offCurves[j]
                ctx.stroke(0,0,1)
                if dd.shouldDrawPoint(p) or dd.shouldDrawPoint(end):
                    if drawAsNeedle:
                        needle(p, end, 2*sw , .5*sw, 4)
                    else:
                        ctx.line(p, end)
                else:
                    dd.notDrawn += 1
                done.append(p)
        ctx.stroke(None)

    def glyphWindow_presentInstance(self, dd):
        # draw the instance and its components
        # called from drawBackground
        ctx.newPath()
        p = sbx.tools.pens.PreviewPen(dd.componentStorage)
        self.instanceGlyphObject.draw(p )
        if self.previewDrawPref == "previewWireframe":
            ctx.strokeWidth(0.4 * dd.scale)
            ctx.stroke(*self._strokePreview_filledStrokeColor)
            ctx.fill(None)
        elif self.previewDrawPref == "previewSolid":
            ctx.strokeWidth(0.4 * dd.scale)
            if self.previewMathPref in ["previewVarLib"]:
                ctx.fill(*self._fillPreview_varLibColor)
            else:
                ctx.fill(*self._fillPreview_mutMathColor)
            ctx.stroke(*self._strokePreview_filledStrokeColor)
        ctx.drawPath()

        if self.showSparksPref and self.drawDetails:
            collector = sbx.tools.pens.CollectorPen({})
            collector.setOffset(dd.offsetForInstanceGlyph,0)
            self.instanceGlyphObject.draw(collector )
            # XXX sometimes this is not the same glyph as the ones already in the dd.collectors.
            dd.debugCollectorNames.append("instanceGlyphObject")
            dd.collectors.append(collector)
            # draw offcurves for the instance
            # ugh, they're relative
            ctx.translate(-dd.offsetForInstanceGlyph,0)
            if not dd.drawingIsHeavy:
                # only draw offcurves for the instance glyph
                # so, not for the masters
                for p in collector.offCurves:
                    bx = (p[0]-.5*dd.scaledMarkerSize , p[1]-.5*dd.scaledMarkerSize , dd.scaledMarkerSize, dd.scaledMarkerSize)
                    if dd.shouldDrawRect(bx):
                        ctx.fill(*self._offCurveColor)
                        ctx.oval(*bx)
            
    def glyphWindow_presentAnchors(self, dd):
        # draw anchors
        # called from drawBackground
        ctx.fill(*self._offCurveColor)
        ctx.stroke(None)
        ctx.fontSize(10*dd.scale)
        h = 10*dd.scale
        ctx.lineHeight(h)
        ctx.font(self._captionTextFont)
        for instanceAnchor in self.instanceGlyphObject.anchors:
            if not dd.shouldDrawPoint((instanceAnchor.x, instanceAnchor.y)):
                dd.notDrawn += 1
                continue
            pos = instanceAnchor.x, instanceAnchor.y + h
            ctx.oval(instanceAnchor.x-.5*dd.anchorMarkerSize, instanceAnchor.y-.5*dd.anchorMarkerSize, dd.anchorMarkerSize, dd.anchorMarkerSize)
            dd.view.drawTextAtPoint(instanceAnchor.name, pos=pos, attributes=self._attrMargins, centerX=True)

    def glyphWindow_presentMarginInfo(self, dd):
        # draw margin and width info
        # called from drawBackground
        ctx.strokeWidth(0.4 * dd.scale)
        ctx.stroke(0,0,0,0.85)
        a, b = (0,0), (0, -self.marginIndicatorLineHeight*dd.scale)
        c, d = (self.instanceGlyphObject.width,0), (self.instanceGlyphObject.width, -self.marginIndicatorLineHeight*dd.scale)
        if dd.shouldDrawPoint(a) or dd.shouldDrawPoint(b):
            ctx.line(a, b)
        if dd.shouldDrawPoint(c) or dd.shouldDrawPoint(d):
            ctx.line(c, d)
        ctx.fill(0)
        ctx.stroke(None)
        fs = 10*dd.scale
        ctx.fontSize(fs)
        lh = 10*dd.scale
        ctx.lineHeight(lh)
        ctx.font(self._captionTextFont)
        v = 30
        try:
            rightText = "r %3.3f\nw %3.3f" % (self.instanceGlyphObject.rightMargin, self.instanceGlyphObject.width)
            rightPos = (d[0]-1*dd.scale, d[1]-2*lh)
            leftText = "l %3.3f" % (self.instanceGlyphObject.leftMargin)
            leftPos =  (b[0]-1*dd.scale, d[1]-lh)
            if dd.shouldDrawPoint(leftPos):
                dd.view.drawTextAtPoint(leftText, pos=leftPos, attributes=self._attrMargins, centerX=True)
            else:
                dd.notDrawn += 1
            if dd.shouldDrawPoint(rightPos):
                dd.view.drawTextAtPoint(rightText, pos=rightPos, attributes=self._attrMargins, centerX=True, backgroundYadd=-10)
            else:
                dd.notDrawn += 1
        except TypeError:
            pass

    def glyphWindow_presentMeasurements(self, dd):
        # calculate measurements
        # called from drawBackground
        mms = 4 * dd.scale
        hmms = .5 * mms
        for m in dd.glyph.naked().measurements:
            # Yay Frederik!
            ctx.fill(0)
            if m.startPoint is not None and m.endPoint is not None:
                x1, y1 = m.startPoint
                x2, y2 = m.endPoint
            
                measureLineAngle = math.atan2(y1-y2, x1-x2) - .5*math.pi
                r = self.contouredInstanceGlyphObject.naked().getRepresentation("doodle.Beam", beam=((x1-dd.offsetForInstanceGlyph,y1),(x2-dd.offsetForInstanceGlyph,y2)), canHaveComponent=False, italicAngle=dd.glyph.font.info.italicAngle)
                r.intersects.sort()
                for i, mp in enumerate(r.intersects):
                    if i == len(r.intersects)-1:
                        break
                    mp1 = r.intersects[i]
                    mp2 = r.intersects[(i+1)]
                    if not (dd.shouldDrawPoint(mp1) or dd.shouldDrawPoint(mp1)):
                        dd.notDrawn += 1
                        continue

                    dmp = math.hypot(mp1[1]-mp2[1], mp1[0]-mp2[0])
                    needlex = math.cos(measureLineAngle) * self._measureLineCurveOffset * dd.scale
                    needley = math.sin(measureLineAngle) * self._measureLineCurveOffset * dd.scale
                    textPos = .5*(mp1[0]+mp2[0])+needlex, .5*(mp1[1]+mp2[1])+needley
                    dd.view.drawTextAtPoint("%3.3f" % dmp, pos=textPos, attributes=self._attrMeasurements, centerX=True)
                    a,b = mp1[0] + needlex, mp1[1] + needley
                    c,d = mp2[0] + needlex, mp2[1] + needley

                    ctx.fill(*self._measureLineColor)
                    ctx.stroke(None)
                    #ctx.oval(a, b, mms, mms)
                    #ctx.oval(c, d, mms, mms)

                    ctx.newPath()
                    ctx.moveTo(mp1)
                    ctx.curveTo((a,b), (c,d), mp2)
                    ctx.fill(None)
                    ctx.strokeWidth(1 * dd.scale)
                    ctx.stroke(*self._measureLineColor)
                    ctx.drawPath()
                for pt in r.intersects:
                    if dd.shouldDrawPoint(pt):
                        ctx.fill(*self._measureLineColor)
                        ctx.stroke(None)
                        ctx.oval(pt[0]-hmms, pt[1]-hmms, mms, mms)
                    else:
                        dd.notDrawn += 1
        ctx.lineDash(None)
        
    def glyphWindow_presentDrawingProblem(self, dd, orientation="bottom"):
        # here we have a drawing problem 
        # show things that will help resolve the drawing problem.
        # called from drawBackground
        # draw all the master outlines
        margin = 250
        verticalMargin = 100
        horizontalMargin = 40
        steps = len(self.sourceGlyphObjects)
        if orientation == "bottom":
            totalSourceWidth = sum([g.width for g in self.sourceGlyphObjects])
            if totalSourceWidth is None or totalSourceWidth == 0:
                totalSourceWidth = 1000
            viewingWidth = dd.glyph.width + 2 * margin
            thisScale = viewingWidth / totalSourceWidth
            ctx.translate(-margin, -margin)
            ctx.scale(thisScale)
            for i, m in enumerate(self.sourceGlyphObjects):
                p = sbx.tools.pens.ContourMarkingPreviewPen(m.font, markerSize=dd.scaledErrorReportMarkerSize, reversedScale=dd.scale/thisScale, showMarkers=True)
                m.draw(p )
                p.paint()
                ctx.translate(m.width,0)
        elif orientation == "left":
            totalSourceHeight = sum([g.bounds[3]-g.bounds[1]+verticalMargin for g in self.sourceGlyphObjects]) + 2 * verticalMargin
            verticalOffset = totalSourceHeight / steps
            thisScale = (dd.ascender-dd.descender) / totalSourceHeight
            ctx.translate(-horizontalMargin, dd.ascender+verticalMargin)
            ctx.scale(thisScale)
            for i, m in enumerate(self.sourceGlyphObjects):
                ctx.translate(0,-m.bounds[3]-verticalMargin)
                ctx.translate(-m.width+m.rightMargin,0)
                p = sbx.tools.pens.ContourMarkingPreviewPen(m.font, markerSize=dd.scaledErrorReportMarkerSize, reversedScale=dd.scale/thisScale, showMarkers=True)
                m.draw(p )
                p.paint()
                ctx.translate(m.width-m.rightMargin,0)

    def glyphWindow_glyphWindow_presentDeltas_noSupport(self, dd):
        # drawing callback when we have support / no support glyphs
        # draw outline for reference (can go later)
        noSupportCollector = sbx.tools.pens.CollectorPen({})
        noSupportCollector.tag = "no_support"
        self.instanceGlyphObject_noSupport.draw(noSupportCollector)
        instanceCollector = sbx.tools.pens.CollectorPen({})
        instanceCollector.tag = "support"
        self.instanceGlyphObject.draw(instanceCollector)
        ctx.stroke(None)
        ctx.fill(None)
        drawAsNeedle = True
        thisMarkerSize = 2 * dd.scaledMarkerSize
        matchedMarkerSize = 5  * dd.scaledMarkerSize
        sw = 1 * dd.scale
        #deltaColor = (.05,.25, .85, 0.8)
        deltaColor = self._offCurveColor
        matchColor = (0,.85, .25, 0.3)
        ignoreThreshold = 2
        ctx.stroke(*deltaColor)
        ctx.strokeWidth(sw)
        
        ctx.lineDash(3*dd.scale,3*dd.scale)
        distances = {}
        try:
            if instanceCollector.onCurves and noSupportCollector.onCurves:
                for j, start in enumerate(instanceCollector.onCurves):
                    end = noSupportCollector.onCurves[j]
                    if not (start,end) in distances:
                        distances[(start,end)] = math.hypot(start[1]-end[1], start[0]-end[0])
                    if distances[(start,end)] < ignoreThreshold:
                        continue
                    ctx.line(end, start)
            if instanceCollector.offCurves and noSupportCollector.offCurves:
                for j, start in enumerate(instanceCollector.offCurves):
                    end = noSupportCollector.offCurves[j]
                    if not (start,end) in distances:
                        distances[(start,end)] = math.hypot(start[1]-end[1], start[0]-end[0])
                    if distances[(start,end)] < ignoreThreshold:
                        continue
                    ctx.line(end, start)
            ctx.stroke(None)
            ctx.lineDash(None)
            if instanceCollector.offCurves and noSupportCollector.offCurves:
                for j, start in enumerate(instanceCollector.offCurves):
                    end = noSupportCollector.offCurves[j]
                    if not (start,end) in distances:
                        distances[(start,end)] = math.hypot(start[1]-end[1], start[0]-end[0])
                    if distances[(start,end)] < ignoreThreshold:
                        continue
                    if dd.shouldDrawPoint(start) or dd.shouldDrawPoint(end):
                        ctx.fill(*deltaColor)
                        ctx.oval(end[0]-.5*thisMarkerSize, end[1]-.5*thisMarkerSize, thisMarkerSize, thisMarkerSize)
                        ctx.oval(start[0]-.5*thisMarkerSize, start[1]-.5*thisMarkerSize, thisMarkerSize, thisMarkerSize)
                    else:
                        dd.notDrawn += 1
            if instanceCollector.onCurves and noSupportCollector.onCurves:
                for j, start in enumerate(instanceCollector.onCurves):
                    end = noSupportCollector.onCurves[j]
                    if not (start,end) in distances:
                        distances[(start,end)] = math.hypot(start[1]-end[1], start[0]-end[0])
                    if distances[(start,end)] < ignoreThreshold:
                        if dd.shouldDrawPoint(start) or dd.shouldDrawPoint(end):
                            ctx.fill(*matchColor)
                            ctx.oval(end[0]-.5*matchedMarkerSize, end[1]-.5*matchedMarkerSize, matchedMarkerSize, matchedMarkerSize)
                        else:
                            dd.notDrawn += 1
                        continue
                    if dd.shouldDrawPoint(start) or dd.shouldDrawPoint(end):
                        ctx.fill(*deltaColor)
                        #ctx.fill(*self._offCurveColor)
                        ctx.oval(end[0]-.5*thisMarkerSize, end[1]-.5*thisMarkerSize, thisMarkerSize, thisMarkerSize)
                        ctx.oval(start[0]-.5*thisMarkerSize, start[1]-.5*thisMarkerSize, thisMarkerSize, thisMarkerSize)
                    else:
                        dd.notDrawn += 1
        except IndexError:
            print("Skateboard: Incompatible outlines in Support")
            
    def observed_glyphWindowWillOpenCallback(self, notification):
        # observer callback            
        g = notification['glyph']
        if g is None: return
        glyphRelevant = self.data_checkGlyphRelevant(g)
        viewKey = id(notification['view'])
        if not glyphRelevant:
            if viewKey in self.currentGlyphViews:
                del self.currentGlyphViews[viewKey]
            return
        self.currentGlyph = notification['glyph']
        self.currentGlyphName = self.currentGlyph.name
        self.currentGlyphViews[viewKey] = notification['glyph']
        # alright, add self.currentGlyph.name to the shoppinglist
        self.data_prepareGlyph(notification['glyph'].name)
        
    def observed_glyphWindowWillCloseCallback(self, notification):
        # observer callback
        viewKey = id(notification['view'])
        if viewKey in self.currentGlyphViews:
            del self.currentGlyphViews[viewKey]
        self.document_checkFonts()
    
    def locationsList_clearSelection(self):
        self.pane_locs.locationsList.setSelection([])
        
    def locationsList_safelySetItems(self, items=None):
        self.locationsList_doNotUpdate = True
        if items is None:
            items = []
        self.pane_locs.locationsList.set(items)
        self.locationsList_doNotUpdate = False
        
    def previewAxesList_safelySetItems(self, items=None):
        # set the values of the location list, but make sure it doesn't call an update
        self.previewAxesList_doNotUpdate = True
        if items is None:
            items = []
        self.pane_previewAxes.locationList.set(items)
        self.previewAxesList_doNotUpdate = False
            
    def previewAxesList_handleEditsCallback(self, sender):
        # vanilla callback for all changes in the previewlocation and interaction source popup
        if self.currentlyDragging:
            # whenever the coordinates are written to the preview location list
            # this method is called to process the new input.
            # But since we know we're dragging we don't have to do anything else
            # the redraw will happen anyway.
            return
        if self.previewAxesList_doNotUpdate:
            return
        newSources = {}
        editedLocation = {}
        validLocation = True
        didEditLocation = False
        didChangeInteractions = False
        for things in sender.get():
            thisAxisName = things['axisKey']
            selectedDirection = things['controlKey']
            try:
                editedLocation[things['axisKey']] = float(things['valueKey'])
                if editedLocation[thisAxisName] != self.instanceLocation.get(thisAxisName):
                    didEditLocation = True
            except (ValueError, KeyError):
                validLocation = False
                AppKit.NSBeep()
            if not selectedDirection in newSources:
                newSources[selectedDirection] = []
            newSources[selectedDirection].append(thisAxisName)
        if newSources != self.interactionSources:
            didChangeInteractions = True
            self.interactionSources = newSources
        if didChangeInteractions:
            self.publishEvent_interactionSourcesChanged()
        if didEditLocation:
            if validLocation:
                self.instanceLocation = Location(editedLocation)
                self.locationsList_clearSelection()
                if not self.currentlyDragging:
                    self.data_instanceLocationChanged()
                    self.publishEvent_previewLocationChanged_dragFinished()
    
    def document_saveDesignSpaceCallback_checkForSaveAs(self, sender=None):
        # vanilla callback
        # make sure we want to do a save or a save as
        modFlags = AppKit.NSEvent.modifierFlags()
        if AppKit.NSAlternateKeyMask & modFlags:
            self.document_saveAsDesignSpaceCallback(sender=sender)
        else:
            self.document_saveDesignSpaceCallback(sender=sender)
        
    def document_saveAsDesignSpaceCallback(self, sender=None):
        currentName = os.path.basename(self.ds.path)
        newPathName = PutFile(message="Save %s As:" % currentName, fileName=currentName)
        if newPathName is not None:
            self.ds.path = newPathName
            self.document_saveDesignSpaceCallback()

            self.w.getNSWindow().setRepresentedURL_(AppKit.NSURL.fileURLWithPath_(self.ds.path))
            self.w.setTitle(str(os.path.basename(self.ds.path)))
            
            #self.document_modificationTime = self.document_getModificationTime()

    def document_saveDesignSpaceCallback(self, sender=None):
        # vanilla callback
        # save the current designsace.
        try:
            interestingItems = [(Location(a).asDict(), b) for a, b in self.pointsOfInterest]
            if interestingItems:
                self.ds.lib[self.interestingLocationLibKey] = interestingItems
            # save the prefs
            if not self.interactionSourcesLibKey in self.ds.lib:
                self.ds.lib[self.interactionSourcesLibKey] = {}
            self.ds.lib[self.interactionSourcesLibKey] = self.interactionSources
            # muted fonts
            mutedFontsForSave = []
            for path, layerName in self.mutedFonts.keys():
                fileName = os.path.basename(path)
                if layerName is None:
                    layerName = "foreground"
                mutedFontsForSave.append((fileName, layerName))
            if mutedFontsForSave:
                self.ds.lib[self.mutedSourcesLibKey] = mutedFontsForSave
            else:
                # remove the key from the lib, otherwise old data remains
                if self.mutedSourcesLibKey in self.ds.lib:
                    del self.ds.lib[self.mutedSourcesLibKey]
            # save math model preference
            self.ds.lib[self.previewMathPrefKey] = self.previewMathPref
            # muted axes
            if self.mutedAxes:
                self.ds.lib[self.mutedAxesLibKey] = list(self.mutedAxes.keys())
            else:
                # remove the key from the lib, otherwise old data remains
                if self.mutedAxesLibKey in self.ds.lib:
                    del self.ds.lib[self.mutedAxesLibKey]
            self.ds.lib[self.previewLocationLibKey] = Location(self.instanceLocation).asDict()
            self.ds.lib[self.previewTextLibKey] = self.previewTextRaw
            self.ds.write(self.ds.path)
            self.document_setNeedSave(False)
            self.document_modificationTime = self.document_getModificationTime()
        except:
            print("Skateboard: Problem while saving..", traceback.format_exc())
    
    def mainWindow_openDocumentWithPath(self, path):
        # open the document at this path
        ext = os.path.splitext(path)[-1]
        if ext == ".designspace":
            self._openDesignspaceDocument(path)
        elif ext == ".sp3":
            self._convertSuperpolatorDocument(path)
        self.mainWindow_confirmSkateboardRunningCallback(notification=None)    #$$

    
    def _convertSuperpolatorDocument(self, path):
        # this a superpolator path and needs to be converted first
        newPathName = path.replace(".sp3", "_converted.designspace")
        newPathName = PutFile(message="Save this Superpolator document to designspace first:", fileName=os.path.basename(newPathName))
        if newPathName:
            sbx.tools.sp3.sp3_to_designspace(path, newPathName)
            self._openDesignspaceDocument(newPathName)
    
    def _openDesignspaceDocument(self, path):
        # first, make sure we can reasonably ready this file
        shouldOpen = True
        msg = "This designspace has structural problems."
        if hasDesignspaceProblems:
            # we should have the module
            checker = designspaceProblems.DesignSpaceChecker(path)
            checker.checkDesignSpaceGeometry()
            checker.checkSources()
            checker.checkInstances()
            if (0,0) in checker.problems:
                # we really can't continue
                self.reportError(msg, "The file can't be parsed.")
                return
            elif (2, 10) in checker.problems:
                # we really can't continue
                self.reportError(msg, "A source has an anisotropic location.")
                return
            elif (1, 9) in checker.problems or (1, 9) in checker.problems:
                # we really can't continue
                self.reportError(msg, "Sorry, the axis values don't make sense.")
                return
            elif (2, 7) in checker.problems:
                # we really can't continue
                self.reportError(msg, "No source defined for default location.")
                return
                
        else:
            print("Don't have the designspaceProblems package")

        # this is a designspace path
        try:
            openDocs = self.getOpenDesignspacePaths()
            if path in openDocs:
                editor = self.getDesignspaceWindowFromPath(path)
                #editor.save(None)
                editor.close()
                #editor.w.getNSWindow().makeKeyAndOrderFront_(editor)
            

            self.ds = SkateboardDesignSpaceProcessor()
            self.ds.useVarlib = True
            # XX Wrap in try / except
            self.ds.read(path)
            
            self.ds.findDefault()
            self.data_updateDocumentVerticalMetrics()
            self.ds.loadFonts()    # reload
            self.checker = designspaceProblems.DesignSpaceChecker(self.ds)
            #self.checker.checkEverything()
            self.checker.checkDesignSpaceGeometry()
            self.checker.checkSources()
            self.checker.checkInstances()
            self.ds.update = self.document_updateCurrentDesignSpace # add skateboard update callback
            self.document_modificationTime = self.document_getModificationTime()
            self.axisNames = []
            self.document_checkFonts()
            self.document_checkAxes()
            self.spaceWindowData = {}
            self.interactionSources = {"vertical":[], "horizontal":[], "ignore":[]}
            try:
                self.document_readSuperpolatorPrefs()
            except:
                # some error reading superpolator prefs
                print("Skateboard: Failed to read superpolator preferences")
                print(traceback.format_exc())
            self.document_readPrefs()
            self.w.file.saveButton.enable(True)
            #self.w.frontPage.openButton.setTitle("Close")
            self.document_setNeedSave(False)
            self.mainWindow_enableInterface(True)
            if self.interestingLocationLibKey in self.ds.lib:
                pointData = []
                try:
                    pointData = [(self.data_validateLocation(a), b) for a, b in self.ds.lib[self.interestingLocationLibKey]]
                finally:
                    self.pointsOfInterest = pointData
            self.w.getNSWindow().setRepresentedURL_(AppKit.NSURL.fileURLWithPath_(self.ds.path))
            self.w.setTitle(str(os.path.basename(self.ds.path)))

            for window in AllGlyphWindows():
                key = id(window.editGlyphView)
                self.currentGlyphViews[key] = window.getGlyph()
                    
            if self.previewLocationLibKey in self.ds.lib:
                prefLoc = self.data_validateLocation(self.ds.lib[self.previewLocationLibKey])
                self.instanceLocation = prefLoc
            else:
                self.instanceLocation = Location(self.ds.newDefaultLocation())
            # set defaults for interaction sources
            tags = self.data_getAxisTags()
            self.axisTagToInteractionSourceDefaults = dict(wdth='horizontal', wght='vertical', opsz="vertical", slnt='horizontal' )
            if not self.hasInteractionSources():
                axesNames = self.instanceLocation.keys()
                for axisName in axesNames:
                    if self.hasInteractionSources(axisName):
                        continue
                    t = tags.get(axisName)
                    preferredDirectionForAxis = self.axisTagToInteractionSourceDefaults.get(t, "ignore")
                    if not axisName in self.interactionSources[preferredDirectionForAxis]:
                        self.interactionSources[preferredDirectionForAxis].append(axisName)
            publishEvent("skateboard.document.opened",
                location = self.instanceLocation, 
                designspace = self.ds, 
                interactionSources = self.interactionSources, 
                axes = self.axisExtremes,
                instanceGlyph = self.instanceGlyphObject,
                previewGlyph = self.contouredInstanceGlyphObject,
                )
            self.document_updateCurrentDesignSpace()
            self.data_instanceLocationChanged()
            self.navigatorTool = navigatorTool.NavigatorTool()
            installTool(self.navigatorTool)
            self.navigatorTool.RFPreviewCallback = self.glyphWindow_drawPreviewCallbackForNavigatorTool
            app = AppKit.NSApp()
            app.getNextSkateboardMasterCallback = self.getNextSkateboardMaster
        except:
            print(traceback.format_exc())
        
        #print('opening self.ds.useVarlib', self.ds.useVarlib)
    
    def data_validateLocation(self, loc):
        # check the location against the defined axes.
        # remove things we don't know
        # add missing axes.
        validLoc = self.ds.newDefaultLocation()
        for axisName, value in loc.items():
            if axisName in validLoc:
                validLoc[axisName] = value
        return Location(validLoc)
        
    def document_getModificationTime(self):
        if self.ds is None:
            return None
        return os.path.getmtime(self.ds.path)
        
    def document_hasExternalModifications(self):
        if self.ds is None:
            self.document_modificationTime = None
            return False
        currentModificationTime = self.document_getModificationTime()
        if self.document_modificationTime != currentModificationTime:
            self.document_modificationTime = currentModificationTime
            return True
        return False
        
    def mainWindow_presentCloseDocumentDialog(self, followupCallback):
        if self.documentNeedSave:
            # handle saving dialog
            result = vanilla_AskYesNoCancel(
                "Save Designspace?",
                #title="Skateboard",
                informativeText="The current Designspace has unsaved changes.",
                parentWindow=self.w,
                resultCallback=followupCallback    #self.mainWindow_closeDocument
                )
    
    def mainWindow_closeDocument(self, result):
        # close the dcument, leave the window open
        if result == 1:
            # save dialog
            self.document_saveDesignSpaceCallback()
            self.document_setNeedSave(False)
            self.mainWindow_presentCloseDocumentDialog_followUp()

        setDefault(self.showSparksPrefKey, self.showSparksPref)
        setDefault(self.previewMathPrefKey, self.previewMathPref)
        
    def mainWindow_presentCloseDocumentDialog_followUp(self):
        # do the actual document closing
        #print("mainWindow_presentCloseDocumentDialog_followUp")
        if self._defconObservers_glyphs:
            for i, g in self._defconObservers_glyphs.items():
                #%% defcon.glyph observer
                g.removeObserver(self, "Glyph.Changed")
                g.removeObserver(self, "Glyph.SelectionChanged")
        self._defconObservers_glyphs = {}
        if self.ds is not None:
            self.ds.update = None
        self.ds = None
        #self.w.frontPage.openButton.setTitle("Open")
        self.w.file.saveButton.enable(False)
        self.mainWindow_enableInterface(False)
        self.instanceLocation = None
        self.previewFonts = {}
        self.document_updateCurrentDesignSpace()
        self.data_instanceLocationChanged()
        self.pointsOfInterest = []
        if self.textWindow is not None:
            self.textWindow.mainWindowIsClosing()
        if self.spaceWindow is not None:
            self.spaceWindow.mainWindowIsClosing()
        if self.navigatorTool is not None:
            self.navigatorTool.RFPreviewCallback = None
            uninstallTool(self.navigatorTool)
            self.navigatorTool = None
        publishEvent("skateboard.document.closed",
            location=None,
            designspace=None,
            interactionSources=None,
            axes=None,
            instanceGlyph = self.instanceGlyphObject,
            previewGlyph = self.contouredInstanceGlyphObject,
            )
        self.storage_mainPreviews = sbx.tools.storageObject.StorageContainer()
        self.storage_supportPreviews = sbx.tools.storageObject.StorageContainer()
        self.storage_kerning = sbx.tools.storageObject.StorageContainer()
        self.storage_fontInfo = sbx.tools.storageObject.StorageContainer()

        self.w.getNSWindow().setRepresentedURL_(None)
        self.w.setTitle(self.toolNameForTitle)
        self.data_updateDocumentVerticalMetrics()
        self.currentGlyphViews = {}
        self.document_modificationTime = None

        app = AppKit.NSApp()
        app.getNextSkateboardMasterCallback = None

    def document_selectCandidateDesignspace(self, sender):
        # vanilla callback for popup list with candidate documents
        # self.w.frontPage.openDesignSpaces
        sel = sender.get()
        if sel is not None:
            path = self.candidateDocuments[sel]
            self.mainWindow_openDocumentWithPath(path)            
        
    def mainWindow_refreshCandidateDocumentsList(self):
        self.candidateDocuments = self.getOpenDesignspacePaths()
        if self.candidateDocuments:
            self.w.frontPage.openDesignSpaces.setItems([os.path.basename(p) for p in self.candidateDocuments])
            self.w.frontPage.openDesignSpaces.show(True)
        else:
            self.w.frontPage.openDesignSpaces.show(False)
    
    def getOpenDesignspacePaths(self):
        # make a dict with all open designspace editor files
        wrappers = getDesignSpaceDocuments()
        docs = []
        for obj in wrappers:
            if obj.designSpacePath is not None:
                docs.append(obj.designSpacePath)
        return docs
    
    def getDesignspaceWindowFromPath(self, path):
        # find the designspace window wrapper that belongs with this path
        wrappers = getDesignSpaceDocuments()
        docs = []
        for obj in wrappers:
            if obj.designSpacePath == path:
                return obj

    def document_openDesignSpaceCallback(self, sender):
        # clicking the "Open" button
        resultPath = GetFile(message="Open a designspace or sp3 file",
            title="Open", directory=None, fileName=None,
            allowsMultipleSelection=False, fileTypes=["designspace", 'sp3'])
        if resultPath is None:
            if self.ds is not None:
                # we had one open, went to the open dialog and changed our mind. 
                return
            self.ds = None
        else:
            self.mainWindow_openDocumentWithPath(resultPath)
        
    def document_closeDesignSpaceCallback(self, sender=None):
        # vanilla callback
        # clicking the "Close" button
        # similar to mainWindow_shouldClose
        #print("1 document_closeDesignSpaceCallback")
        def callback(result):
            #print(result)
            if result == 1:
                # save first
                #print("3 save first")
                self.document_saveDesignSpaceCallback()
                self.document_setNeedSave(False)
                self.mainWindow_presentCloseDocumentDialog_followUp()
            if result == 0:
                #print("2 just close")
                self.mainWindow_presentCloseDocumentDialog_followUp()
                self.document_setNeedSave(False)
            else:
                return False
        
        # are there unsaved changes?
        if self.documentNeedSave:
            vanilla.dialogs.askYesNoCancel(messageText="Save this designspace?", informativeText="", parentWindow=self.w, resultCallback=callback)
        else:
            self.mainWindow_presentCloseDocumentDialog_followUp()
        return False
        
    def document_selectAndOpenDesignSpaceCallback(self, sender):
        # vanilla callback, open document, close document
        # open a file dialog to get a designspace
        if "Open" in sender.getTitle():
            # clicking the "Open" button
            resultPath = GetFile(message="Open a designspace file",
                title="Open", directory=None, fileName=None,
                allowsMultipleSelection=False, fileTypes=["designspace"])
            if resultPath is None:
                if self.ds is not None:
                    # we had one open, went to the open dialog and changed our mind. 
                    return
                self.ds = None
            else:
                self.mainWindow_openDocumentWithPath(resultPath)
        elif "Close" in sender.getTitle():
            # clicking the "Close" button
            self.mainWindow_presentCloseDocumentDialog(self.mainWindow_closeDocument)
                        
    def document_setNeedSave(self, state=True):
        self.documentNeedSave = state
        if state:
            self.w.file.saveButton.enable(True)
            self.w.getNSWindow().setDocumentEdited_(True)
        else:
            self.w.file.saveButton.enable(False)
            self.w.getNSWindow().setDocumentEdited_(False)
    
    def document_readPrefs(self):
        # read the document skateboard prefs
        # interaction sources
        docInteractionSources = self.ds.lib.get(self.interactionSourcesLibKey)
        orderedAxisNames = [a.name for a in self.ds.axes]
        validInteractionSources = {'horizontal': [], 'vertical': [], 'ignore': []}
        if docInteractionSources:
            for k, axisNames in docInteractionSources.items():
                if k in validInteractionSources:
                    for axisName in axisNames:
                        if axisName in orderedAxisNames:
                            validInteractionSources[k].append(axisName)
        self.interactionSources = validInteractionSources
        # mathmodel
        if self.previewMathPrefKey in self.ds.lib:
            self.previewMathPref = self.ds.lib[self.previewMathPrefKey]
        
        # previewtext
        rawText = self.ds.lib.get(self.previewTextLibKey)
        if rawText is not None:
            self.previewTextRaw = rawText
        # muted fonts
        mutedSourcesFromDoc = self.ds.lib.get(self.mutedSourcesLibKey)
        if mutedSourcesFromDoc is not None:
            mutedSources = {}
            for ufoName, layerName in mutedSourcesFromDoc:
                for sD in self.ds.sources:
                    if os.path.basename(sD.path) == ufoName:
                        if layerName in ["foreground", 'public.default']:
                            layerName = None
                        if layerName == sD.layerName:
                            sourceDescriptorKey = (sD.path, sD.layerName)
                            mutedSources[sourceDescriptorKey] = sD
            self.mutedFonts = mutedSources                            
        # muted axes
        mutedAxesFromDoc = self.ds.lib.get(self.mutedAxesLibKey)
        orderedAxisNames = [a.name for a in self.ds.axes]
        if mutedAxesFromDoc is not None:
            mutedAxes = {}
            for axisName in mutedAxesFromDoc:
                if axisName not in orderedAxisNames:
                    continue
                mutedAxes[axisName] = True
            self.mutedAxes = mutedAxes
        
    def document_readSuperpolatorPrefs(self):
        # read the document, or superpolator prefs
        superpolatorPrefs = self.ds.lib.get(self.superpolatorDataLibKey)
        orderedAxisNames = [a.name for a in self.ds.axes]
        if not superpolatorPrefs:
            if len(orderedAxisNames)>0:
                self.interactionSources["horizontal"] = [orderedAxisNames[0]]
            if len(orderedAxisNames)>1:
                self.interactionSources["vertical"] = [orderedAxisNames[1]]
            return
        if 'verticalPreviewAxis' in superpolatorPrefs:
            wantAxisName = superpolatorPrefs['verticalPreviewAxis']
            if wantAxisName in orderedAxisNames:
                self.interactionSources["vertical"] = [wantAxisName]
        else:
            if len(orderedAxisNames)>0:
                self.interactionSources["vertical"] = [orderedAxisNames[0]]
        if 'horizontalPreviewAxis' in superpolatorPrefs:
            wantAxisName = superpolatorPrefs['horizontalPreviewAxis']
            if wantAxisName in orderedAxisNames:
                self.interactionSources["horizontal"] = [wantAxisName]
        else:
            self.interactionSources["horizontal"] = [orderedAxisNames[1]]
        #self.previewTextRaw = superpolatorPrefs.get('previewtext', None)
    
    def sourcesList_selectionChangedCallback(self, sender):
        # vanilla callback
        # called when the selection changed in sources list
        self.timerSchedule_navigatorTool_mouseUp_draggingDone()
        #for window in AllGlyphWindows():
        #    window.editGlyphView.refresh()
    
    def closeFontCallback(self, sender=None):
        for i in self.pane_doc.l.getSelection():
            key = self.fontPaths[i]
            #if key[0] in self.openFonts:
                #print("closeFontCallback", key)
    
    def sourcesList_selectFontCallback(self, sender=None):
        # vanilla callback
        # also called from the sources menu
        openThese = []

    
        bar = ProgressBar()
        bar.label("Opening UFO…")
        bar.tick()
        for i in self.pane_doc.l.getSelection():
            path, layerName = self.fontPaths[i]
            if not os.path.exists(path):
                # alert the font does not exist
                AppKit.NSBeep()
                continue
            f = OpenFont(path, showInterface=True)
            f.document().showWindows()
            if self.currentGlyph is not None:
                # if we have a current glyph - open the edit window
                if self.currentGlyph.name in f:
                    if self.verbose:
                        print("open edit window for %s"%self.currentGlyph.name)
                    OpenGlyphWindow(f[self.currentGlyph.name])
                else:
                    if self.verbose:
                        print("this glyph not in this font")
            if layerName is not None:
                SetCurrentLayerByName(layerName)
        bar.close()

    def observed_startObservingKerning(self):
        pass
    def observed_stopObservingKerning(self):
        pass
            
    def observed_startObservingCurrentGlyph(self):
        """
        """
        key = id(self.currentGlyph)
        
        if key not in self._defconObservers_glyphs:
            self._defconObservers_glyphs[key] = self.currentGlyph
            try:
                #%% defcon.glyph observer
                self.currentGlyph.addObserver(self, "observed_currentGlyphEditedCallback", "Glyph.Changed")
                self.currentGlyph.addObserver(self, "observed_currentGlyphSelectionChangedCallback", "Glyph.SelectionChanged")
            except AssertionError:
                pass
        else:
            if self.verbose:
                print("observed_startObservingCurrentGlyph skipping observer for", key)
    
    def observed_stopObservingCurrentGlyph(self):
        # stop observing the current glyph 
        if not self._defconObservers_glyphs:
            if self.verbose:
                print("observed_stopObservingCurrentGlyph nothing left")
            return
        if id(self.currentGlyph) in self._defconObservers_glyphs:
            g = self._defconObservers_glyphs[id(self.currentGlyph)]
            #%% defcon.glyph observer
            g.removeObserver(self, "Glyph.Changed")
            g.removeObserver(self, "Glyph.SelectionChanged")
            del self._defconObservers_glyphs[id(self.currentGlyph)]
    
    def observed_currentGlyphSelectionChangedCallback(self, defconNotification):
        # defcon observer callback for changes inside the glyph
        key = (self.ds.path, self.currentGlyph.name)
        selectedSingles, selectedPairs = sbx.tools.findCluster.findCluster(self.currentGlyph)
        if not selectedSingles and not selectedPairs:
            if key in self.currentSelectedClusters:
                del self.currentSelectedClusters[key]
        else:
            self.currentSelectedClusters[key] = selectedSingles, selectedPairs
        
    def observed_currentGlyphEditedCallback(self, defconNotification):
        # defcon observer callback for changes inside the glyph
        # destroy the mutators for the current glyph so that we will make a new one when we're drawing.
        if self.currentGlyphName is not None:
            self.storage_mainPreviews.notifyAboutChange(self.currentGlyphName)
            self.storage_supportPreviews.notifyAboutChange(self.currentGlyphName)
            self.data_prepareGlyph(self.currentGlyphName)
            if self.ds is not None:
                k = (self.ds.path, self.currentGlyphName)
                self.ds._glyphMutators.clear()
                self.timerSchedule_navigatorTool_mouseUp_draggingDone()
                if self.textWindow:
                    self.timerSchedule_textWindowRedraw()

    def data_checkGlyphRelevant(self, glyph):
        # hm this might not be enough
        # it could be a layer that is not used
        if glyph is None:
            return False
        if glyph.font is None:
            return False
        return glyph.font.path in self.openFonts

    def data_checkFontRelevant(self, font):
        if font is None:
            return False
        return font.path in self.openFonts
        
    def observed_currentGlyphChangedCallback(self, notification):
        # robofont observer callback for a different current glyph
        # do we need to check glyph relevance here?

        if self.ds is None:
            # we should not be here.
            # there is no document, this callback should have been removed
            return

        glyphRelevant = self.data_checkGlyphRelevant(notification['glyph'])
        if not glyphRelevant:
            return
        viewKey = id(notification['view'])
        lastGlyphName = self.currentGlyphName
        if self.currentGlyphName is not None:
            if self.ds is not None:
                k = (self.ds.path, self.currentGlyphName)
        # start observing the new glyph
        self.currentGlyphName = None
        if notification['glyph'] is None:
            if self.currentGlyph is not None:
                self.observed_stopObservingCurrentGlyph()
            self.currentGlyphName = None
            if viewKey in self.currentGlyphViews:
                del self.currentGlyphViews[viewKey]
        else:
            if self.currentGlyph is not None:
                self.observed_stopObservingCurrentGlyph()
            self.currentGlyph = notification['glyph']
            self.currentGlyphViews[viewKey] = notification['glyph']
            self.observed_startObservingCurrentGlyph()
            self.currentGlyphName = notification['glyph'].name
            self.currentFontPath = notification['glyph'].font.path
            parentPath = notification['glyph'].font.path
            if parentPath != self.currentFontPath:
                self.currentFontPath = parentPath
            self.document_updateCurrentDesignSpace()
        self.data_prepareGlyph(notification['glyph'].name)            
        if self.textWindow:
            #
            if lastGlyphName != self.currentGlyphName:
                # update the text window
                self.textWindow_updateText()
            self.timerSchedule_textWindowRedraw()

    def glyphIsSupport(self, thisGlyph=None):
        # return True if this glyph (ufo + layername) is a support
        # Note: check what the frontlayer name is.
        if thisGlyph is None:
            return False, None
        #isEmpty = checkGlyphIsEmpty(thisGlyph, allowWhiteSpace=True)
        isEmpty = thisGlyph.getRepresentation("Skateboard.isGlyphEmpty")
        sourceKey = (self.currentFontPath, thisGlyph.layerName)
        locationForKey = None
        for sd in self.ds.sources:
            if sourceKey == (sd.path, sd.layerName):
                locationForKey = sd.location
        return (sourceKey in self.fontPaths) and (not isEmpty), locationForKey
    
    def currentGlyphIsSupport(self):
        # return True if the current glyph (ufo + layername) is a support
        # Note: check what the frontlayer name is.
        if self.currentGlyph is None:
            return False, None
        isEmpty = self.currentGlyph.getRepresentation("Skateboard.isGlyphEmpty")
        sourceKey = (self.currentFontPath, self.currentGlyph.layerName)
        locationForKey = None
        for sd in self.ds.sources:
            if sourceKey == (sd.path, sd.layerName):
                locationForKey = sd.location
        return (sourceKey in self.fontPaths) and (not isEmpty), locationForKey
        
    def observed_fontDidOpenCallback(self, notification):
        # observer callback
        font = notification['font']
        self.document_checkFonts()

    def observed_fontDidCloseCallback(self, notification):
        # observer callback
        self.document_checkFonts()

    
if __name__ == "__main__":
    import os
    o = OpenWindow(SkateBoard)
    
